﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UIPopup = Windows.UI.Xaml.Controls.Primitives.Popup;

using pdftron.PDF;
using pdftron.Common;
using pdftron.PDF.Annots;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using Windows.UI.Input;
using pdftron.PDF.Tools.Controls;
using pdftron.SDF;
using pdftron.PDF.OCG;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    public class AnnotEditRectGroup: AnnotEdit
    {
        protected Dictionary<IAnnot, int> mSelectedAnnotations = new Dictionary<IAnnot, int>();
        protected SolidColorBrush mFillPaint = new SolidColorBrush();
        protected PDFRect mSelectedRectInPage = null;
        protected double mOpacity = 1;
        protected bool mCreated = false;
        protected bool mIsFillColor = false;

        public ToolType GetToolMode()
        {
            return mToolMode;
        }       
               
        public AnnotEditRectGroup(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_annot_edit_rect_group;
            mToolMode = ToolType.e_annot_edit_rect_group;
            mSelectedColor = new Color();

            mPt1 = new Point();
            mPt2 = new Point();

            // appearance of the selection widget
            mSelectionBackgroundColor = new SolidColorBrush(Color.FromArgb(128, 128, 128, 125));
            mCtrlPointFillBrush = new SolidColorBrush(Color.FromArgb(255, 150, 180, 255));
            mCtrlPointBorderBrush = new SolidColorBrush(Color.FromArgb(255, 200, 200, 200));
            mCtrlPointBorderThickness = 2;
            mCtrlPointRadius = 7.5;

            // register for the keyboard callback
            mKeyboardHelper = new KeyboardHelper();
            mKeyboardHelper.SubscribeToKeyboard(true);            
        }

        /// <summary>
        /// Select the group which the annnotation passed is part of
        /// </summary>
        /// <param name="annotation">Annotation part of a group</param>
        /// <param name="pageNumber">Page number where the annotation is located</param>
        internal void SelectAnnotationGroup(IAnnot annotation, int pageNumber)
        {
            bool isRtReply = false;
            try
            {
                isRtReply = UtilityFunctions.IsReplyTypeReply(annotation);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            mAnnot = annotation;

            if ((mAnnot != null) && (!isRtReply))
            {
                KeyValuePair<ToolType, List<IAnnot>> pair = CanSelectGroupAnnotation(mPDFView, annotation, pageNumber);

                if (pair.Key == ToolType.e_none)
                {
                    mToolManager.CreateDefaultTool();
                    return;
                }

                mGroupAnnotations = pair.Value;
                
                try
                {
                    mPDFView.DocLockRead();

                    PopulateMenu();

                    mSelectedRectInPage = FindSelectedAnnotationsAndRect();
                    mAnnotPageNum = pageNumber;
                    mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);

                    mJustSwitchedFromAnotherTool = false;
                    CreateAppearance();
                    EnableScrolling();

                    mCreated = true;
                }
                catch (Exception) { }
                finally
                {
                    mPDFView.DocUnlockRead();
                }
                this.IsHitTestVisible = false;
                mToolManager.RaiseSelectionGroupChangedEvent(mSelectedAnnotations);                
            }
        }

        internal override void OnCreate()
        {            
            try
            {
                mPDFView.DocLockRead();
                mGroupAnnotations = mToolManager.CurrentTool.GroupAnnotations;
                mSelectedRectInPage = FindSelectedAnnotationsAndRect();
                mAnnotPageNum = mToolManager.CurrentTool.CurrentAnnotPageNumber;
                mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                
                PopulateMenu();
                InitializeCanvasAndGeometry();
                mCreated = true;
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlockRead();
            }
            this.IsHitTestVisible = false;

            mToolManager.RaiseSelectionGroupChangedEvent(mSelectedAnnotations);
        }

        protected override void SetPosition() 
        {
            // Place the tool's canvas on top of the page
            mPageCropOnClient = BuildPageBoundBoxOnClient(mAnnotPageNum);
            if (mPageCropOnClient == null)
                return;
            
            this.SetValue(Canvas.LeftProperty, mPageCropOnClient.x1);
            this.SetValue(Canvas.TopProperty, mPageCropOnClient.y1);
            this.Width = mPageCropOnClient.x2 - mPageCropOnClient.x1;
            this.Height = mPageCropOnClient.y2 - mPageCropOnClient.y1;

            // If no annotations selected, it will return null
            if (mSelectedRectInPage == null)
                return;

            // Convert to screen space
            PDFDouble x1 = new PDFDouble(mSelectedRectInPage.x1);
            PDFDouble y1 = new PDFDouble(mSelectedRectInPage.y1);
            PDFDouble x2 = new PDFDouble(mSelectedRectInPage.x2);
            PDFDouble y2 = new PDFDouble(mSelectedRectInPage.y2);

            // Get coordinates of two opposite points in screen space
            mPDFView.ConvPagePtToScreenPt(x1, y1, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(x2, y2, mAnnotPageNum);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            mLeft = mOldLeft = Math.Min(x1.Value, x2.Value);
            mRight = mOldRight = Math.Max(x1.Value, x2.Value);
            mTop = mOldTop = Math.Min(y1.Value, y2.Value);
            mBottom = mOldBottom = Math.Max(y1.Value, y2.Value);
            mHorizontalMiddle = (mLeft + mRight) / 2.0;
            mVerticalMiddle = (mTop + mBottom) / 2.0;

            CreateSelectionAppearance();

            if (mIsShowingCommandMenu)
            {
                HideContextMenu(mContextMenu);
                PositionContextMenu(mContextMenu);
                ShowContextMenu(mContextMenu);
            }
        }

        /// <summary>
        /// Adds in the menu options according to the type of annotation
        /// </summary>
        protected override void PopulateMenu()
        {
            if (mAnnot != null)
            {
                mMenuTitles = null;
                try
                {
                    mMenuTitles = new Dictionary<string, string>();
                    //locks the document first as accessing annotation/doc information isn't thread safe.
                    mPDFView.DocLockRead();

                    mMenuTitles.Add("done", ResourceHandler.GetString("EditOption_Done"));
                    mMenuTitles.Add("ungroup", ResourceHandler.GetString("EditOption_Ungroup"));
                    mMenuTitles.Add("flatten", ResourceHandler.GetString("EditOption_Flatten"));
                    mMenuTitles.Add("opacity", ResourceHandler.GetString("EditOption_Opacity"));
                    mMenuTitles.Add("fill color", ResourceHandler.GetString("EditOption_FillColor"));
                    mMenuTitles.Add("color", ResourceHandler.GetString("EditOption_Color"));
                    mMenuTitles.Add("delete", ResourceHandler.GetString("EditOption_Delete"));
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlockRead();
                }

                if (mMenuTitles != null && !mAnnotIsSticky) // if sticky, we don't create menu until it's needed
                {
                    mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                    mContextMenu.ItemClicked += OnCommandMenuClicked;
                }
            }
        }

        /// <summary>
        /// Go over and check if pointer location is over the selected group rectangle
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <returns></returns>
        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (mAnnot == null)
            {
                SetMAnnot();
            }
            if (mSelectedRectInPage != null)
            {
                //if point pressed is outside the selected rect in page space then
                //exit multi-select tool.
                PointerPoint downPoint = e.GetCurrentPoint(mViewerCanvas);
                PDFDouble x = new PDFDouble(downPoint.Position.X);
                PDFDouble y = new PDFDouble(downPoint.Position.Y);
                mPDFView.ConvAnnotationCanvasPtToScreenPt(x, y);
                mPDFView.ConvScreenPtToPagePt(x, y, mAnnotPageNum);

                bool groupRectFound = false;

                if (mSelectedRectInPage.Contains(x.Value, y.Value))
                {
                    groupRectFound = true;
                }
                else
                {
                    groupRectFound = IsControlPointPressedAt(downPoint.Position.X, downPoint.Position.Y);                    
                }

                if (!groupRectFound)
                {
                    BackToPan();
                    return false;
                }                
            }

            base.PointerPressedHandler(sender, e);
            return false;
        }        

        /// <summary>
        /// Make sure to Check if ctrl points got clicked considering their radius/thickness
        /// </summary>
        /// <param name="x"> X Pressed position </param>
        /// <param name="y"> Y Pressed position </param>
        /// <returns> True if control got pressed otherwise False </returns>
        private bool IsControlPointPressedAt(double x, double y)
        {
            // Always consider control point location + radius/thickness
            var pnt = mCtrlPointLocations;
            var rds = mCtrlPointRadius;
            PDFRect rect;

            foreach (UIPoint pt in pnt)
            {
                rect = new PDFRect(pt.X - rds, pt.Y - rds, pt.X + rds, pt.Y + rds);
                if (rect.Contains(x, y))
                {
                    return true;
                }
            }

            return false;
        }        

        internal override bool PointerMovedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (!mIsManipulated || mToolManager.ReadOnlyMode)
            {
                return false;
            }
            mPt2.x = e.GetCurrentPoint(mViewerCanvas).Position.X;
            mPt2.y = e.GetCurrentPoint(mViewerCanvas).Position.Y;
            if (!mStartMoving) // ensure we actually move a little before we start with updates and stuff
            {
                if ((Math.Abs(mPt2.x - mPt1.x) > START_MOVING_THRESHOLD)
                    || (Math.Abs(mPt2.y - mPt1.y) > START_MOVING_THRESHOLD))
                {
                    mStartMoving = true;
                }
                else
                {
                    return false;
                }
            }

            if ((mToolMode != ToolType.e_annot_edit_rect_group) 
                && (mToolMode != ToolType.e_annot_edit_group_create)
                && (mToolMode != ToolType.e_lasso_create))
            {
                return false;
            }
            
            CheckBoundsOnPointerMoved();
            return true;
        }

        protected void UpdateAnnotationsOnMoveAndResize()
        {
            PDFDouble x1 = new PDFDouble(mLeft);
            PDFDouble y1 = new PDFDouble(mTop);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            mPDFView.ConvScreenPtToPagePt(x1, y1, mAnnotPageNum);

            PDFDouble x2 = new PDFDouble(mRight);
            PDFDouble y2 = new PDFDouble(mBottom);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
            mPDFView.ConvScreenPtToPagePt(x2, y2, mAnnotPageNum);

            PDFDouble old_x1 = new PDFDouble(mOldLeft);
            PDFDouble old_y1 = new PDFDouble(mOldTop);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(old_x1, old_y1);
            mPDFView.ConvScreenPtToPagePt(old_x1, old_y1, mAnnotPageNum);

            PDFDouble old_x2 = new PDFDouble(mOldRight);
            PDFDouble old_y2 = new PDFDouble(mOldBottom);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(old_x2, old_y2);
            mPDFView.ConvScreenPtToPagePt(old_x2, old_y2, mAnnotPageNum);

            double xDisplacement = (x1.Value - old_x1.Value);
            double yDisplacement = (y1.Value - old_y1.Value);

            // current control point location
            PDFDouble controlPointX1 = new PDFDouble(mCtrlPointLocations[e_ul].X);
            PDFDouble controlPointY1 = new PDFDouble(mCtrlPointLocations[e_ul].Y);

            mPDFView.ConvAnnotationCanvasPtToScreenPt(controlPointX1, controlPointY1);
            mPDFView.ConvScreenPtToPagePt(controlPointX1, controlPointY1, mAnnotPageNum);

            PDFDouble controlPointX2 = new PDFDouble(mCtrlPointLocations[e_lr].X);
            PDFDouble controlPointY2 = new PDFDouble(mCtrlPointLocations[e_lr].Y);

            mPDFView.ConvAnnotationCanvasPtToScreenPt(controlPointX2, controlPointY2);
            mPDFView.ConvScreenPtToPagePt(controlPointX2, controlPointY2, mAnnotPageNum);

            PDFRect selectedAreaInPage = new PDFRect(mSelectedRectInPage.x1, mSelectedRectInPage.y1, mSelectedRectInPage.x2, mSelectedRectInPage.y2);
            selectedAreaInPage.Normalize();

            double offsetX = Math.Min(controlPointX1.Value, controlPointX2.Value);
            double offsetY = Math.Max(controlPointY2.Value, controlPointY1.Value);
            double scaleX = Math.Abs((controlPointX2.Value - controlPointX1.Value) / (selectedAreaInPage.x2 - selectedAreaInPage.x1));
            double scaleY = Math.Abs((controlPointY2.Value - controlPointY1.Value) / (selectedAreaInPage.y2 - selectedAreaInPage.y1));

            foreach (var annotationItem in mSelectedAnnotations)
            {
                PDFRect annotRect = annotationItem.Key.GetRect();
                annotRect.Normalize();

                // check bounds
                PDFRect newAnnotRect = null;
                switch (mEffectiveCtrlPoint)
                {
                    case MOVE_WIDGET_CODE: // regular move
                    case e_ll:
                    case e_lm:
                    case e_lr:
                    case e_mr:
                    case e_ur:
                    case e_um:
                    case e_ul:
                    case e_ml:

                        if (!DoesAnnotationSupportResize(annotationItem.Key))
                        {
                            continue;
                        }

                        annotRect = new PDFRect(annotRect.x1 - selectedAreaInPage.x1, annotRect.y1 - selectedAreaInPage.y2,
                            annotRect.x2 - selectedAreaInPage.x1, annotRect.y2 - selectedAreaInPage.y2);

                        newAnnotRect = new PDFRect(annotRect.x1 * scaleX + offsetX,
                            annotRect.y1 * scaleY + offsetY,
                            annotRect.x2 * scaleX + offsetX,
                            annotRect.y2 * scaleY + offsetY);
                        newAnnotRect.Normalize();

                        // It is possible during viewing that GetRect does not return the most accurate bounding box
                        // of what is actually rendered, to obtain the correct behavior when resizing/moving, we
                        // need to call refreshAppearance before resize
                        if ((annotationItem.Key.GetAnnotType() != AnnotType.e_Stamp)
                            && (annotationItem.Key.GetAnnotType() != AnnotType.e_Text))
                        {
                            annotationItem.Key.RefreshAppearance();
                        }

                        annotationItem.Key.SetDateToNow();

                        annotationItem.Key.Resize(newAnnotRect);
                        break;
                }

                AnnotType annotationItemType = annotationItem.Key.GetAnnotType();
                if (annotationItemType == AnnotType.e_Line || annotationItemType == AnnotType.e_Circle || annotationItemType == AnnotType.e_Square || annotationItemType == AnnotType.e_Polyline
                    || annotationItemType == AnnotType.e_Polygon || annotationItemType == AnnotType.e_Ink || annotationItemType == AnnotType.e_FreeText)
                {
                    annotationItem.Key.RefreshAppearance();
                }

                mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
                mSelectedRectInPage.Set(x1.Value, y1.Value, x2.Value, y2.Value);
            }
        }        

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mCreated)
            {
                if (mIsManipulated)
                {
                    if (mStartMoving)
                    {
                        try
                        {
                            mPDFView.DocLock(true);
                            UpdateAnnotationsOnMoveAndResize();

                            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
                            {
                                return false;
                            }
                            mPointerID = -1;
                            mEffectiveCtrlPoint = -1;
                            mViewerCanvas.ReleasePointerCaptures();

                            // points in Screen space
                            PDFDouble x1 = new PDFDouble(mLeft);
                            PDFDouble y1 = new PDFDouble(mTop);
                            PDFDouble x2 = new PDFDouble(mRight);
                            PDFDouble y2 = new PDFDouble(mBottom);

                            mLeft = mOldLeft = Math.Min(x1.Value, x2.Value);
                            mRight = mOldRight = Math.Max(x1.Value, x2.Value);
                            mTop = mOldTop = Math.Min(y1.Value, y2.Value);
                            mBottom = mOldBottom = Math.Max(y1.Value, y2.Value);
                            mHorizontalMiddle = (mLeft + mRight) / 2.0;
                            mVerticalMiddle = (mTop + mBottom) / 2.0;

                            CreateSelectionAppearance();

                            // Update the viewer at the location of the old rectangle
                            x1 = new PDFDouble(mOldLeft);
                            y1 = new PDFDouble(mOldTop);
                            x2 = new PDFDouble(mOldRight);
                            y2 = new PDFDouble(mOldBottom);
                            mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                            mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);

                            Rect updateRect = new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
                            mPDFView.Update(updateRect);

                            mToolManager.RaiseAnnotationGroupEditedEvent(mSelectedAnnotations);
                        }
                        catch (Exception ex)
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                        }
                        finally
                        {
                            mPDFView.DocUnlock();
                        }
                    }

                    // Our new values now become our old values for the next manipulation
                    mOldLeft = mLeft;
                    mOldRight = mRight;
                    mOldTop = mTop;
                    mOldBottom = mBottom;
                    if (mContextMenu == null)
                    {
                        mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                        mContextMenu.ItemClicked += OnCommandMenuClicked;
                    }
                    mIsShowingCommandMenu = true;
                    PositionContextMenu(mContextMenu);
                    ShowContextMenu(mContextMenu);
                }
                // reset values concerned with moving the widget
                mStartMoving = false;
                mIsManipulated = false;
                EnableScrolling();
            }

            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                return false;
            }
            mPointerID = -1;
            mEffectiveCtrlPoint = -1;
            mViewerCanvas.ReleasePointerCaptures();

            return false;
        }

        /// <summary>
        /// Find all selected annotations rectangles and return a group rectangle
        /// </summary>
        /// <returns> Annotation Group Rectangle that combines all selected annotations </returns>
        protected virtual PDFRect FindSelectedAnnotationsAndRect()
        {
            if (mGroupAnnotations == null)
            {
                return new PDFRect();
            }

            mSelectedAnnotations.Clear();
            PDFRect selectedRectInPage = null;

            mAnnotIsTextMarkup = false;
            foreach (IAnnot annot in mGroupAnnotations)
            {
                PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot, mAnnotPageNum);
                annotRect.Normalize();

                if (!DoesAnnotationSupportResize(annot))
                {
                    mAnnotIsTextMarkup = true;
                }
                mSelectedAnnotations.Add(annot, mAnnotPageNum);

                if (selectedRectInPage == null)
                {
                    selectedRectInPage = new PDFRect(annotRect.x1, annotRect.y1, annotRect.x2, annotRect.y2);
                }
                else
                {
                    if (annotRect.x1 < selectedRectInPage.x1)
                    {
                        selectedRectInPage.x1 = annotRect.x1;
                    }

                    if (annotRect.y1 < selectedRectInPage.y1)
                    {
                        selectedRectInPage.y1 = annotRect.y1;
                    }

                    if (annotRect.x2 > selectedRectInPage.x2)
                    {
                        selectedRectInPage.x2 = annotRect.x2;
                    }

                    if (annotRect.y2 > selectedRectInPage.y2)
                    {
                        selectedRectInPage.y2 = annotRect.y2;
                    }
                }
            }

            SetMAnnot();
            return selectedRectInPage;
        }

        protected void SetMAnnot()
        {
            if ((mGroupAnnotations != null) && (mGroupAnnotations.Count > 0))
            {
                mAnnot = mGroupAnnotations[mGroupAnnotations.Count - 1];
            }
        }
                        
        protected void SetAnnotOptions(IAnnot annotation)
        {
            mType = annotation.GetAnnotType();

            if (annotation.IsMarkup())
            {
                mMarkup = (IMarkup)annotation;
            }

            mAnnotIsSticky = (mType == AnnotType.e_Text);
            mAnnotIsTextMarkup = (mType == AnnotType.e_Highlight ||
                                    mType == AnnotType.e_Underline ||
                                    mType == AnnotType.e_StrikeOut ||
                                    mType == AnnotType.e_Squiggly);

            if ((annotation.IsMarkup() && mType != AnnotType.e_FreeText)
                || mType == AnnotType.e_Text)
            {
                if (annotation.GetAnnotType() == AnnotType.e_Stamp)
                {
                    pdftron.SDF.Obj sigObj = annotation.GetSDFObj();
                    pdftron.SDF.Obj stampObj = annotation.GetSDFObj();

                    sigObj = sigObj.FindObj(ToolManager.SignatureAnnotationIdentifyingString);
                    stampObj = stampObj.FindObj(ToolManager.StamperAnnotationIdentifyingString);

                    if (sigObj != null)
                    {
                        mAnnotIsSignature = true;
                    }
                    else if (stampObj != null)
                    {
                        mAnnotIsStamper = true;
                    }
                }
            }
        }        

        protected override void DeleteAnnotationOnKeyDown()
        {
            try
            {
                mPDFView.DocLock(true);

                foreach (var annotationItem in mSelectedAnnotations)
                {
                    mToolManager.RaiseAnnotationPreRemovedEvent(annotationItem.Key, mAnnotPageNum);

                    mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                    mPage.AnnotRemove(annotationItem.Key);
                    mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
                }

                if (mToolManager.UndoRedoAction != null)
                {
                    mToolManager.UndoRedoAction.TakeSnapshot("Multi-annotation delete");
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
                mAnnot = null;
            }
        }

        #region Handlers For Popup Menu

        internal override void OnCommandMenuClicked(string title)
        {
            if (mAnnot == null)
            {
                SetMAnnot();
            }

            mIsShowingCommandMenu = false;
            HideContextMenu(mContextMenu);

            if (title.Equals("done", StringComparison.OrdinalIgnoreCase))
            {
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                DeleteGroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("ungroup", StringComparison.OrdinalIgnoreCase))
            {
                UngroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("fill color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = true;
                if (mAnnot.GetAnnotType() == AnnotType.e_FreeText)
                {
                    mEffectiveBrush = e_text_fill;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_fill;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("opacity", StringComparison.OrdinalIgnoreCase))
            {
                HandleOpacity();
            }
            else if (title.Equals("color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = false;
                AnnotType aType = mAnnot.GetAnnotType();
                if (aType == AnnotType.e_Line || aType == AnnotType.e_Ink || aType == AnnotType.e_Polyline)
                {
                    mEffectiveBrush = e_stroke;
                    ShowColorPicker();
                }
                else if (aType == AnnotType.e_Highlight)
                {
                    mEffectiveBrush = e_highlight;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_text_markup;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("flatten", StringComparison.OrdinalIgnoreCase))
            {
                HandleFlatten();
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }
        }

        /// <summary>
        /// Sets the selected color of the current annotation
        /// </summary>
        protected override void SetCurrentAnnotColor()
        {
            try
            {
                mPDFView.DocLock(true);
                ColorPt color = new ColorPt(mSelectedColor.R / 255.0, mSelectedColor.G / 255.0, mSelectedColor.B / 255.0);
                foreach (var annotationItem in mSelectedAnnotations)
                {
                    AnnotType aType = annotationItem.Key.GetAnnotType();
                    if (mIsFillColor)
                    {
                        if (aType == AnnotType.e_FreeText)
                        {
                            mEffectiveBrush = e_text_fill;
                        }
                        else
                        {
                            mEffectiveBrush = e_fill;
                        }
                    }
                    else
                    {
                        if (aType == AnnotType.e_Line || aType == AnnotType.e_Ink || aType == AnnotType.e_Polyline)
                        {
                            mEffectiveBrush = e_stroke;
                        }
                        else if (aType == AnnotType.e_Highlight)
                        {
                            mEffectiveBrush = e_highlight;
                        }
                        else
                        {
                            mEffectiveBrush = e_text_markup;
                        }
                    }

                    SetAnnotOptions(annotationItem.Key);
                    switch (annotationItem.Key.GetAnnotType())
                    {
                        case AnnotType.e_Circle:
                        case AnnotType.e_Square:
                        case AnnotType.e_Line:
                        case AnnotType.e_Ink:
                        case AnnotType.e_Highlight:
                        case AnnotType.e_Underline:
                        case AnnotType.e_Squiggly:
                        case AnnotType.e_StrikeOut:
                        case AnnotType.e_Polygon:
                        case AnnotType.e_Polyline:
                            if (mEffectiveBrush == e_stroke || mEffectiveBrush == e_highlight || mEffectiveBrush == e_text_markup)
                            {
                                if (mUseColor)
                                {
                                    mMarkup.SetColor(color, 3);
                                }
                                else
                                {
                                    ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                    mMarkup.SetColor(emptyColor, 0); // using 0 color points makes it transparent
                                }
                            }
                            else if (mEffectiveBrush == e_fill)
                            {
                                if (mUseColor)
                                {
                                    mMarkup.SetInteriorColor(color, 3);
                                }
                                else
                                {
                                    ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                    mMarkup.SetInteriorColor(emptyColor, 0);
                                }
                            }
                            mMarkup.RefreshAppearance();
                            mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
                            break;
                        case AnnotType.e_FreeText:
                            
                            pdftron.PDF.Annots.FreeText ft = (FreeText)annotationItem.Key;
                            if (mEffectiveBrush == e_text_stroke)
                            {
                                if (mUseColor)
                                {
                                    ft.SetTextColor(color, 3);
                                }
                                else
                                {
                                    ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                    ft.SetTextColor(emptyColor, 0);
                                }
                            }
                            else if (mEffectiveBrush == e_text_markup)
                            {
                                if (mUseColor)
                                {
                                    ft.SetLineColor(color, 3);
                                }
                                else
                                {
                                    ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                    ft.SetLineColor(emptyColor, 0);
                                }
                            }
                            else if (mEffectiveBrush == e_text_fill)
                            {
                                if (mUseColor)
                                {
                                    ft.SetColor(color, 3);
                                }
                                else
                                {
                                    ColorPt emptyColor = new ColorPt(0, 0, 0, 0);
                                    ft.SetColor(emptyColor, 0);
                                }
                            }
                            ft.RefreshAppearance();
                            mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
                            break;
                    }
                }
            }
            catch (Exception) { }
            finally
            {
                mPDFView.DocUnlock();
                if (mSelectedAnnotations.Count > 0)
                {
                    if (mToolManager.UndoRedoAction != null)
                    {
                        if (mIsFillColor)
                        {
                            mToolManager.UndoRedoAction.TakeSnapshot("Multi-annotation fill color change");
                        }
                        else
                        {
                            mToolManager.UndoRedoAction.TakeSnapshot("Multi-annotation color change");
                        }
                    }
                }
            }
        }

        #region Opacity        

        protected override void UpdateOpacitySettingsAndSetOpacity(double opacity)
        {
            foreach (var annotationItem in mSelectedAnnotations)
            {
                // Update settings
                AnnotType type = annotationItem.Key.GetAnnotType();

                if (type == AnnotType.e_Underline || type == AnnotType.e_StrikeOut || type == AnnotType.e_Squiggly)
                {
                    Settings.TextMarkupOpacity = opacity;
                }
                else if (type == AnnotType.e_Highlight)
                {
                    Settings.HightlightOpacity = opacity;
                    //roamingSettings.Values["Tools:HighlightOpacity"] = opacity;
                }
                else if (annotationItem.Key.IsMarkup())
                {
                    Settings.MarkupOpacity = opacity;
                }

                SetGroupOpacity(opacity, annotationItem.Key);
            }

            if (mSelectedAnnotations.Count > 0)
            {
                if (mToolManager.UndoRedoAction != null)
                {
                    mToolManager.UndoRedoAction.TakeSnapshot("Multi-annotation Opacity change");
                }
            }
        }

        protected void SetGroupOpacity(double opacity, IAnnot annotation)
        {
            try
            {
                mPDFView.DocLock(true);
                SetAnnotOptions(annotation);
                mMarkup.SetOpacity(opacity);
                mMarkup.RefreshAppearance();
                mPDFView.UpdateWithAnnot(annotation, mAnnotPageNum);
            }
            catch (Exception)
            {

            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #endregion Opacity

        #region AnnotationProperties

        protected override void FlattenAnnotation(PDFPage page)
        {
            foreach (var annotationItem in mSelectedAnnotations)
            {
                annotationItem.Key.Flatten(page);
                mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
            }

            if ((mToolManager != null) && (mToolManager.UndoRedoAction != null))
            {
                mToolManager.UndoRedoAction.TakeSnapshot("Flatten Annotations");
            }
        }

        #endregion AnnotationProperties

        #endregion Handlers For Popup Menu

        protected void DeleteGroupAnnotations()
        {
            mViewerCanvas.Children.Clear();

            try
            {
                mPDFView.DocLock(true);
                foreach (var annotationItem in mSelectedAnnotations)
                {
                    mToolManager.RaiseAnnotationPreRemovedEvent(annotationItem.Key, mAnnotPageNum);
                    mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                    mPage.AnnotRemove(annotationItem.Key);
                    mPDFView.UpdateWithAnnot(annotationItem.Key, mAnnotPageNum);
                }
            }
            catch (Exception e)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
            }
            finally
            {
                mPDFView.DocUnlock();
                mAnnot = null;

                mToolManager.RaiseAnnotationGroupRemovedEvent(mSelectedAnnotations);

            }
        }

        protected bool DoesAnnotationSupportResize(IAnnot annotation)
        {
            try
            {
                return ((annotation.GetAnnotType() != AnnotType.e_Highlight)
                        && (annotation.GetAnnotType() != AnnotType.e_StrikeOut)
                        && (annotation.GetAnnotType() != AnnotType.e_Underline)
                        && (annotation.GetAnnotType() != AnnotType.e_Squiggly)
                        && (annotation.GetAnnotType() != AnnotType.e_Text)
                        && (annotation.GetAnnotType() != AnnotType.e_Stamp));
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return false;
        }

        protected void BackToPan()
        {
            mSelectedRectInPage = new PDFRect();
            mSelectedAnnotations.Clear();
            mEffectiveCtrlPoint = -1;
            mAnnot = null;
            mGroupAnnotations = null;
            mNextToolMode = ToolType.e_pan;
            EndCurrentTool(ToolType.e_pan);
        }

        protected void UngroupAnnotations()
        {
            List<IAnnot> annotations = mSelectedAnnotations.Keys.ToList();
            if ((annotations == null) || (annotations.Count == 0))
            {
                return;
            }

            try
            {
                mToolManager.RaiseAnnotationGroupPreEditedEvent(mSelectedAnnotations);
                UtilityFunctions.UngroupAnnotations(mPDFView, annotations);
                mToolManager.RaiseAnnotationGroupEditedEvent(mSelectedAnnotations);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            BackToPan();

        }
    }
}
