﻿using pdftron.PDF.Tools;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System.IO;
using Windows.ApplicationModel.Resources;
using Windows.Storage;
using Windows.UI.ViewManagement;

namespace pdftron.PDF.Tools.Collections
{
    public class PinnedItem : ViewModelBase
    {
        private bool _isSelected;
        private PinnedType _pinnedItemType;
        private bool _needsBorder;
        private string _thumbnailLocation;
        private bool _thumbLoaded;
        private bool _showNameInCoverMode = true;
        private double _opacity = 1.0;
        private string _documentName = string.Empty;
        private bool _hasAdditionalIcon;
        private string _additionalIconLocation;

        public PinnedItem(string documentName, string token, string documentPath)
        {
            DocumentName = documentName;
            Token = token;
            DocumentPath = documentPath;
            CheckAdditionalIcon();
        }

        public PinnedItem(IStorageItem file)
        {
            File = file;
            DocumentName = file.Name;
            DocumentPath = file.Path;
            if (file is StorageFile)
            {
                PinnedItemType = PinnedType.File;
                CheckAdditionalIcon();
            }
            else
            {
                SetToFolder();
            }
        }

        public bool IsSelected
        {
            get => _isSelected;
            set => Set(ref _isSelected, value);
        }

        public PinnedType PinnedItemType
        {
            get => _pinnedItemType;
            set
            {
                if (Set(ref _pinnedItemType, value))
                    RaisePropertyChanged(nameof(IsFolder));
            }
        }

        public bool IsFolder => _pinnedItemType == PinnedType.Folder;

        public bool NeedsBorder
        {
            get => _needsBorder;
            set => Set(ref _needsBorder, value);
        }

        public string ThumbnailLocation
        {
            get => _thumbnailLocation;
            set => Set(ref _thumbnailLocation, value);
        }

        public bool ThumbLoaded
        {
            get => _thumbLoaded;
            set => Set(ref _thumbLoaded, value);
        }

        public bool ShowNameInCoverMode
        {
            get => _showNameInCoverMode;
            set => Set(ref _showNameInCoverMode, value);
        }

        public double Opacity
        {
            get => _opacity;
            set => Set(ref _opacity, value);
        }

        public string DocumentName
        {
            get => _documentName;
            set => Set(ref _documentName, value);
        }

        public string DocumentExtension
        {
            get
            {
                if (PinnedItemType == PinnedType.File)
                    return Path.GetExtension(DocumentName);

                return string.Empty;
            }
        }

        public bool HasAdditionalIcon
        {
            get => _hasAdditionalIcon;
            set => Set(ref _hasAdditionalIcon, value);
        }

        public string AdditionalIconLocation
        {
            get => _additionalIconLocation;
            set => Set(ref _additionalIconLocation, value);
        }

        public string Token { get; }

        public string DocumentPath { get; }

        public IStorageItem File { get; set; }

        public void SetToFolder()
        {
            PinnedItemType = PinnedType.Folder;
            ThumbLoaded = true;
            ThumbnailLocation = "ms-appx:///Assets/DocumentPage/FolderIcon.png";
        }

        public void CheckAdditionalIcon()
        {
            if (UtilityFunctions.IsOneDrive(DocumentPath))
            {
                HasAdditionalIcon = true;
                AdditionalIconLocation = RecentDocumentProperties.ONEDRIVE_ICON_LOCATION;
            }
        }

        public static PinnedItem CreatePin()
        {
            string text = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources").GetString("DocumentsPage_FoldersPage_PinItem_Text");
            if (ApplicationView.GetForCurrentView().VisibleBounds.Width < 450)
                text = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources").GetString("DocumentsPage_FoldersPage_SmallPinItem_Text");

            return new PinnedItem(text, string.Empty, "pin");
        }
    }
}
