﻿using pdftron.PDF.Tools;
using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.IO;
using Windows.ApplicationModel;

namespace pdftron.PDF.Tools.Collections
{
    /// <summary>
    /// Represents a recently viewed item
    /// </summary>
    public class RecentItem : ViewModelBase
    {
        private bool _isSelected;
        private bool _needsBorder;

        /// <summary>
        /// Then name of the document
        /// </summary>
        public string DocumentName { get; }

        /// <summary>
        /// The path where the thumbnail is located
        /// </summary>
        public string ThumbnailLocation
        {
            get => Properties.ThumbnailLocation;
            set
            {
                if (Properties.ThumbnailLocation != value)
                {
                    Properties.ThumbnailLocation = value;
                    RaisePropertyChanged();
                    NeedsBorder = ThumbnailLocation?.Equals(RecentDocumentProperties.DEFAULT_THUMB_PATH, StringComparison.OrdinalIgnoreCase) == false &&
                        !ThumbnailLocation.Equals(RecentDocumentProperties.DEFAULT_PROTECTED_THUMB_PATH, StringComparison.OrdinalIgnoreCase) &&
                        !ThumbnailLocation.Equals(RecentDocumentProperties.DEFAULT_PACKAGE_THUMB_PATH, StringComparison.OrdinalIgnoreCase) &&
                        !ThumbnailLocation.Equals(RecentDocumentProperties.DEFAULT_FOLDER_PATH, StringComparison.OrdinalIgnoreCase);
                }
            }
        }

        public string AdditionalIconLocation
        {
            get
            {
                switch (Properties.AdditonalIcon)
                {
                    case AdditionalIcons.OneDrive:
                        return Path.Combine(Package.Current.InstalledLocation.Path, "Assets", "DocumentPage", "OneDrive_rgb_Blue2728.png");
                    default:
                        return null;
                }
            }
        }

        public bool HasAdditionalLogo => !string.IsNullOrEmpty(AdditionalIconLocation);

        public string Token { get; }

        public string DocumentPath
        {
            get => Properties.FilePath;
            private set => Properties.FilePath = value;
        }

        public bool IsProtected
        {
            get => Properties.IsProtected;
            set => Properties.IsProtected = value;
        }

        public int LastPageNumber
        {
            get => Properties.PageNumber;
            set => Properties.PageNumber = value;
        }

        public bool IsSelected
        {
            get => _isSelected;
            set => Set(ref _isSelected, value);
        }

        public bool NeedsBorder
        {
            get => _needsBorder;
            set => Set(ref _needsBorder, value);
        }

        public RecentDocumentProperties Properties { get; set; }

        public RecentItem(string documentName, string thumbLocation, string token, string documentPath, bool isProtected = false)
        {
            DocumentName = documentName;
            Properties = new RecentDocumentProperties(null, 1, isProtected);
            ThumbnailLocation = thumbLocation;
            Token = token;
            DocumentPath = documentPath;
            IsSelected = false;

            if (UtilityFunctions.IsOneDrive(documentPath))
            {
                Properties.AdditonalIcon = AdditionalIcons.OneDrive;
            }
            else
            {
                Properties.AdditonalIcon = AdditionalIcons.None;
            }
        }

        public RecentItem(RecentDocumentProperties properties, string token)
        {
            // don't parse a root system folder
            if (properties.FilePath.Split(new string[] { "\\" }, StringSplitOptions.RemoveEmptyEntries).Length <= 1)
            {
                DocumentName = properties.FilePath;
            }
            else
            {
                DocumentName = Path.GetFileName(properties.FilePath);
            }
            Properties = properties;
            Token = token;
            if (UtilityFunctions.IsOneDrive(DocumentPath))
            {
                Properties.AdditonalIcon = AdditionalIcons.OneDrive;
            }
            else
            {
                Properties.AdditonalIcon = AdditionalIcons.None;
            }
        }
    }
}
