﻿using System.Collections.Generic;
using System.Windows.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Commands.Events
{
    public static class PasswordChangedCommand
    {
        private static readonly Dictionary<DependencyObject, RoutedEventHandler> _textChangedHandlers = new Dictionary<DependencyObject, RoutedEventHandler>();

        public static readonly DependencyProperty CommandProperty = DependencyProperty.RegisterAttached(
            "Command",
            typeof(ICommand),
            typeof(PasswordChangedCommand),
            new PropertyMetadata(null, OnCommandPropertyChanged));

        public static void SetCommand(DependencyObject d, ICommand value) => d.SetValue(CommandProperty, value);

        public static ICommand GetCommand(DependencyObject d) => (ICommand)d.GetValue(CommandProperty);

        private static void OnCommandPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is PasswordBox control)
            {
                if (e.NewValue != null)
                {
                    RoutedEventHandler handler = new RoutedEventHandler(Control_PasswordChanged);
                    _textChangedHandlers[d] = handler;
                    control.PasswordChanged += handler;
                }
                else
                {
                    if (_textChangedHandlers.ContainsKey(d))
                    {
                        control.PasswordChanged -= _textChangedHandlers[d];
                        _textChangedHandlers.Remove(d);
                    }
                }
            }
        }

        private static void Control_PasswordChanged(object sender, RoutedEventArgs e)
        {
            PasswordBox control = sender as PasswordBox;
            var command = GetCommand(control);

            if (command?.CanExecute(control.Password) == true)
                command.Execute(control.Password);
        }
    }
}
