﻿using System;
using System.Collections.Generic;
using System.Windows.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Commands.Events
{
    public static class PointerContactCommand
    {
        private static readonly Dictionary<FrameworkElement, Dictionary<uint, Pointer>> _contactPoints = new Dictionary<FrameworkElement, Dictionary<uint, Pointer>>();

        public static readonly DependencyProperty CommandProperty = DependencyProperty.RegisterAttached(
            "Command",
            typeof(ICommand),
            typeof(PointerContactCommand),
            new PropertyMetadata(null, OnCommandPropertyChanged));

        public static void SetCommand(DependencyObject d, ICommand value) => d.SetValue(CommandProperty, value);

        public static ICommand GetCommand(DependencyObject d) => (ICommand)d.GetValue(CommandProperty);

        private static void OnCommandPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is FrameworkElement control)
            {
                control.PointerPressed += Control_PointerPressed;
                control.PointerMoved += Control_PointerMoved;
                control.PointerReleased += Control_PointerReleased;
                control.PointerCanceled += Control_PointerCanceled;
                control.PointerCaptureLost += Control_PointerCaptureLost;
            }
        }

        private static void Control_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            if (!_contactPoints.ContainsKey(control))
                _contactPoints[control] = new Dictionary<uint, Pointer>();

            _contactPoints[control][e.Pointer.PointerId] = e.Pointer;
            control.CapturePointer(e.Pointer);
            ExecuteCommand(sender, e);
        }

        private static void Control_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            if (_contactPoints.ContainsKey(control) && _contactPoints[control].ContainsKey(e.Pointer.PointerId))
                ExecuteCommand(sender, e);
        }

        private static void Control_PointerReleased(object sender, PointerRoutedEventArgs e) => RemovePointer(sender, e);

        private static void Control_PointerCanceled(object sender, PointerRoutedEventArgs e) => RemovePointer(sender, e);

        private static void Control_PointerCaptureLost(object sender, PointerRoutedEventArgs e) => RemovePointer(sender, e);

        private static void RemovePointer(object sender, PointerRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            Pointer ptr = e.Pointer;
            if (_contactPoints.ContainsKey(control) && _contactPoints[control].ContainsKey(ptr.PointerId))
            {
                _contactPoints[control].Remove(ptr.PointerId);
                if (_contactPoints[control].Count == 0)
                    _contactPoints.Remove(control);

                control.ReleasePointerCapture(e.Pointer);
            }
        }

        private static void ExecuteCommand(object sender, PointerRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            var command = GetCommand(control);

            if (command?.CanExecute(e) == true)
                command.Execute(new Tuple<FrameworkElement, PointerRoutedEventArgs>(control, e));
        }
    }
}
