﻿using System.Windows.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Commands.Events
{
    public static class RightTappedCommand
    {
        public static readonly DependencyProperty CommandProperty = DependencyProperty.RegisterAttached(
            "Command",
            typeof(ICommand),
            typeof(RightTappedCommand), new PropertyMetadata(null, OnCommandPropertyChanged));

        public static void SetCommand(DependencyObject d, ICommand value) => d.SetValue(CommandProperty, value);

        public static ICommand GetCommand(DependencyObject d) => (ICommand)d.GetValue(CommandProperty);

        private static void OnCommandPropertyChanged(DependencyObject d,
            DependencyPropertyChangedEventArgs e)
        {
            if (d is FrameworkElement control)
            {
                if (e.NewValue != null)
                {
                    control.RightTapped += Control_RightTapped;
                }
                else
                {
                    control.RightTapped -= Control_RightTapped;
                }
            }
        }

        private static void Control_RightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            var command = GetCommand(control);

            if (command?.CanExecute(e) == true)
                command.Execute(e);
        }
    }
}
