﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

/* NOTE:
 * This is a custom CommandBar class that adds some extra features like the
 * direction which the command bar opens (Auto, Downwards or Upwards)
 */

namespace pdftron.PDF.Tools.Controls
{
    public class CommandBarCustom : CommandBar
    {
        public enum DirectionType
        {
            Auto,
            Downwards,
            Upwards
        }

        protected override void OnApplyTemplate()
        {
            base.OnApplyTemplate();
            var layoutRoot = GetTemplateChild("LayoutRoot") as Grid;
            if (layoutRoot != null)
            {
                VisualStateManager.SetCustomVisualStateManager(layoutRoot, new CustomOpenCommandBarVisualStateManager(OpeningDirection));
            }
        }

        #region Properties

        public DirectionType OpeningDirection
        {
            get { return (DirectionType)GetValue(OpeningDirectionProperty); }
            set { SetValue(OpeningDirectionProperty, value); }
        }

        // Using a DependencyProperty as the backing store for OpeningDirection.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty OpeningDirectionProperty =
            DependencyProperty.Register("OpeningDirection", typeof(DirectionType),
                typeof(CommandBarCustom),  new PropertyMetadata(DirectionType.Auto));

        #endregion

        class CustomOpenCommandBarVisualStateManager : VisualStateManager
        {
            const string STATE_OPEN_UP = "CompactOpenUp";
            const string STATE_OPEN_DOW = "CompactOpenDown";

            DirectionType _directionType;

            public CustomOpenCommandBarVisualStateManager(DirectionType type)
            {
                _directionType = type;
            }

            protected override bool GoToStateCore(Control control, FrameworkElement templateRoot, string stateName, VisualStateGroup group, VisualState state, bool useTransitions)
            {
                if (_directionType == DirectionType.Auto)
                    return base.GoToStateCore(control, templateRoot, stateName, group, state, useTransitions);

                /* In case the state name is CompactOpenUp we replace to CompactOpenDown
                 * to force downwards behavior */
                if (_directionType == DirectionType.Downwards)
                {
                    if (!string.IsNullOrWhiteSpace(stateName) && stateName.Contains(STATE_OPEN_UP))
                        stateName = STATE_OPEN_DOW;
                }
                else
                {
                    if (!string.IsNullOrWhiteSpace(stateName) && stateName.Contains(STATE_OPEN_DOW))
                        stateName = STATE_OPEN_UP;
                }

                return base.GoToStateCore(control, templateRoot, stateName, group, state, useTransitions);
            }
        }
    }
}
