﻿using System.Collections.Generic;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public partial class ThumbnailsViewViewModel
    {
        /// <summary>
        /// Rotates all the selected pages. Applies XAML rotation to increase responsive first.
        /// After the rotated thumbnail is fetched, the XAML rotation is reset.
        /// </summary>
        /// <param name="parameter"></param>
        private void RotateSelectedPages(object parameter)
        {
            _pdfViewCtrl.CancelAllThumbRequests();
            List<int> pagesToRotate = new List<int>();
            try
            {
                _pdfViewCtrl.DocLock(true);
                if (parameter != null)
                {
                    HandleRotation(parameter as ThumbnailItem);
                    pagesToRotate.Add(ThumbnailsList.IndexOf(parameter as ThumbnailItem) + 1);
                }
                else
                {
                    foreach (ThumbnailItem item in _selectedItems)
                    {
                        ThumbnailItem thumbItem = item;
                        HandleRotation(thumbItem);
                        pagesToRotate.Add(ThumbnailsList.IndexOf(item) + 1);
                    }
                }

                // Call this to make the rotate take effect in the viewer page immediately
                _pdfViewCtrl.UpdatePageLayout();

                if (ToolManager?.UndoRedoAction != null)
                    ToolManager.UndoRedoAction.TakeSnapshot("Rotate Page", true);
            }
            finally
            {
                _pdfViewCtrl.DocUnlock();
            }

            foreach (int pageNum in pagesToRotate)
            {
                _pageDeleted?.Invoke(pageNum);
            }

            ThumbnailsList.RequestResources();
        }
    }
}
