﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Utilities;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer
{
    public partial class ViewerViewModel
    {
        private async Task GoToFilesAsync()
        {
            TabControlViewModel?.SelectedTab?.ToolManager?.CloseOpenDialog();
            CloseCurrentTool();
            if (TabControlViewModel.SelectedTab != null)
            {
                SaveNeededStates status = CheckSavingRequirements(TabControlViewModel.SelectedTab);
                if (status == SaveNeededStates.saveNeeded)
                {
                    SaveResult result = await SaveToOriginalAsync(TabControlViewModel.SelectedTab, true, true);
                    if (result != SaveResult.e_normal)
                    {
                        try
                        {
                            IsSecretlyModal = true;
                            if (TabControlViewModel.SelectedTab.PDFViewCtrl != null)
                                TabControlViewModel.SelectedTab.PDFViewCtrl.IsHitTestVisible = false;

                            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                            string message = string.Format(loader.GetString("ViewerPage_AutoSaveFailed_Close_Message_AccessDenied"),
                                TabControlViewModel.SelectedTab.Title, loader.GetString("Generic_Cancel_Text"));
                            if (result == SaveResult.e_unknown_error)
                                message = string.Format(loader.GetString("ViewerPage_AutoSaveFailed_Close_Message_Unknown"), TabControlViewModel.SelectedTab.Title);

                            ContentDialog dialog = new ContentDialog 
                            {
                                Title = loader.GetString("ViewerPage_AutoSaveFailed_Title"),
                                Content = message,
                                PrimaryButtonText = loader.GetString("ViewerPage_AutoSaveFailed_Close_Discard"),
                                CloseButtonText = loader.GetString("Generic_Cancel_Text")
                            };

                            dialog.PrimaryButtonCommand = new RelayCommand(async _ => 
                            {
                                await NavigateAwayAsync();
                            });

                            await DialogHelper.ShowContentDialogAsync(dialog);
                        }
                        finally
                        {
                            IsSecretlyModal = false;
                            if (TabControlViewModel.SelectedTab.PDFViewCtrl != null)
                                TabControlViewModel.SelectedTab.PDFViewCtrl.IsHitTestVisible = true;
                        }
                    }
                }
                else
                {
                    _pageNumberChanged?.Invoke(TabControlViewModel.SelectedTab);
                    await NavigateAwayAsync();
                }
            }
        }
    }
}
