﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.Storage;

using pdftron.PDF.Tools.Controls.ViewModels.Compare;
using pdftron.PDF.Tools.Controls.ControlBase;
using Windows.ApplicationModel.DataTransfer;
using System.Threading.Tasks;
using Windows.ApplicationModel.Resources;
using pdftron.PDF.Tools.Utilities;

// The User Control item template is documented at https://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls.Compare
{
    public sealed partial class PdfCompareControl : UserControl, ControlBase.ICloseableControl, IMultiWindowUserControl
    {
        ResourceLoader _loader;
        string mDraggingFeedback; // cached text for dragging feedback

        public PdfCompareControl()
        {
            this.InitializeComponent();
            this.DataContext = ViewModel = new PdfCompareViewModel();

            Init();
        }

        public PdfCompareControl(PDFViewCtrl ctrl)
        {
            this.InitializeComponent();
            this.DataContext = ViewModel = new PdfCompareViewModel(ctrl, this);

            Init();
        }

        public PdfCompareControl(IStorageFile file)
        {
            this.InitializeComponent();
            this.DataContext = ViewModel = new PdfCompareViewModel(file);

            Init();
        }

        /// <summary>
        /// Initialize some additional resources
        /// </summary>
        private void Init()
        {
            _loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            pageTitle.Text = _loader.GetString("PDFCompareControl_pageTitle");
            browsePdfA.Text = browsePdfB.Text = mDraggingFeedback = _loader.GetString("PDFCompareControl_browse");
            pdfACompareDesc.Text = pdfBCompareDesc.Text  = _loader.GetString("PDFCompareControl_compareDescription");
            pdfADropDesc.Text = pdfBDropDesc.Text = _loader.GetString("PDFCompareControl_dropFileDescription");
        }

        public PdfCompareViewModel ViewModel { get; set; }

        public event ControlBase.ControlClosedDelegate ControlClosed = delegate { };

        public void CloseControl()
        {
            ViewModel.CleanUp();
            ControlClosed();
        }

        private async void PdfA_OpenFile_DragOver(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, false, true, mDraggingFeedback);
        }

        private async void PdfB_OpenFile_DragOver(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, false, false, mDraggingFeedback);
        }

        private async void PdfA_OpenFile_Drop(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, true, true);
        }

        private async void PdfB_OpenFile_Drop(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, true, false);
        }

        /// <summary>
        /// Handle drag and drop operations
        /// </summary>
        /// <param name="e"></param>
        /// <param name="isDrop">True if dropping, otherwise dragging</param>
        /// <param name="isPdfA">True if Pdf A / left side, otherwise Pdf B / right side</param>
        /// <param name="drag_caption">Feedback when dragging PDF</param>
        /// <returns></returns>
        private async Task HandleDragAndDropOperation(DragEventArgs e, bool isDrop, bool isPdfA, string drag_caption = null)
        {
            DragOperationDeferral deferral = e.GetDeferral();
            
            if (e.DataView.Contains(StandardDataFormats.StorageItems))
            {
                IReadOnlyList<IStorageItem> items = await e.DataView.GetStorageItemsAsync();
                if (items?.Count == 1)
                {
                    if (items[0] is StorageFile storageFile && storageFile.FileType.Equals(".pdf", StringComparison.OrdinalIgnoreCase))
                    {
                        if (isDrop)
                        {
                            if (ViewModel != null)
                            {
                                if (isPdfA)
                                    await ViewModel.OpenAndCompareFileAAsync(storageFile);
                                else
                                    await ViewModel.OpenAndCompareFileBAsync(storageFile);
                            }
                        }
                        else
                        {
                            e.AcceptedOperation = DataPackageOperation.Copy;                            

                            if (string.IsNullOrEmpty(drag_caption) == false)
                                e.DragUIOverride.Caption = drag_caption;
                        }
                    }
                    else
                    {
                        e.AcceptedOperation = DataPackageOperation.None;
                    }
                }
            }

            deferral.Complete();
        }
    }
}
