﻿using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The Content Dialog item template is documented at https://go.microsoft.com/fwlink/?LinkId=234238

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class CreateDocControl : ContentDialog, IDisposable
    {
        private CreateDocControlViewModel _ViewModel;

        #region Events
        private CreateDocControlViewModel.NewDocumentCreatedDelegate _NewDocumentCreated;

        /// <summary>
        /// Raised when the document is successfully created
        /// </summary>
        public event CreateDocControlViewModel.NewDocumentCreatedDelegate NewDocumentCreated
        {
            add
            {
                if (_NewDocumentCreated?.GetInvocationList().Contains(value) != true)
                    _NewDocumentCreated += value;
            }
            remove => _NewDocumentCreated -= value;
        }

        private CreateDocControlViewModel.DocumentCreationFailedDelegate _documentCreationFailed;


        public event CreateDocControlViewModel.DocumentCreationFailedDelegate DocumentCreationFailed
        {
            add
            {
                if (_documentCreationFailed?.GetInvocationList().Contains(value) != true)
                    _documentCreationFailed += value;
            }
            remove => _documentCreationFailed -= value;
        }

        #endregion

        public CreateDocControl()
        {
            this.InitializeComponent();
            _ViewModel = new CreateDocControlViewModel();
            DataContext = _ViewModel;
            SetRefViewModels(ref _ViewModel);
            _ViewModel.NewDocumentCreated += RaiseNewDocumentCreated;
        }

        public void Dispose()
        {
            if (_ViewModel != null)
            {
                _ViewModel.NewDocumentCreated -= RaiseNewDocumentCreated;
                _ViewModel = null;
            }
        }

        #region Private Methods
        private void RaiseNewDocumentCreated(Windows.Storage.StorageFile file)
        {
            if (_NewDocumentCreated != null)
                _NewDocumentCreated.Invoke(file);
        }

        private void NavigationView_SelectionChanged(Microsoft.UI.Xaml.Controls.NavigationView sender, Microsoft.UI.Xaml.Controls.NavigationViewSelectionChangedEventArgs args)
        {
            Microsoft.UI.Xaml.Controls.NavigationViewItem item = args.SelectedItem as Microsoft.UI.Xaml.Controls.NavigationViewItem;
            if (item == null || _ViewModel == null)
                return;

            string tag = item.Tag.ToString();
            if (tag == string.Empty)
                return;

            if (tag == "Blank")
            {
                _ViewModel.DocumentType = CreateDocControlViewModel.CreateDocFrom.Blank;
                _ViewModel.ImageType = CreateDocControlViewModel.ImageCreationType.None;
            }
            else if (tag == "Image")
            {
                _ViewModel.DocumentType = CreateDocControlViewModel.CreateDocFrom.Image;
                _ViewModel.ImageType = CreateDocControlViewModel.ImageCreationType.File;
            }
            else if (tag == "Camera")
            {
                _ViewModel.DocumentType = CreateDocControlViewModel.CreateDocFrom.Image;
                _ViewModel.ImageType = CreateDocControlViewModel.ImageCreationType.Camera;
            }
        }

        private void SetRefViewModels(ref CreateDocControlViewModel viewModel)
        {
            if (BlankDocControl != null)
            {
                BlankDocControl.ViewModel = viewModel;
            }
            if (ImageDocControl != null)
            {
                ImageDocControl.ViewModel = viewModel;
            }

        }

        #endregion
    }
}
