﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class MagnifyingToolbar : UserControl
    {
        public static readonly DependencyProperty OrientationProperty = DependencyProperty.Register(
         "Orientation",
         typeof(Orientation),
         typeof(MagnifyingToolbar),
         new PropertyMetadata(Orientation.Horizontal, new PropertyChangedCallback(OnOrientationChanged)));

        public Orientation Orientation
        {
            get { return (Orientation)GetValue(OrientationProperty); }
            set { SetValue(OrientationProperty, value); }
        }

        private static void OnOrientationChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            Orientation ori = (Orientation)e.NewValue;
            MagnifyingToolbar toolbar = (d as MagnifyingToolbar);
            toolbar.InkOptionsToolbar.Orientation = ori;
            toolbar.NavigationOptionsToolbar.Orientation = ori;
            if (ori == Orientation.Horizontal)
            {
                toolbar.InkOptionsToolbar.HorizontalAlignment = HorizontalAlignment.Stretch;
                toolbar.InkOptionsToolbar.VerticalAlignment = VerticalAlignment.Top;
                toolbar.NavigationOptionsToolbar.HorizontalAlignment = HorizontalAlignment.Stretch;
                toolbar.NavigationOptionsToolbar.VerticalAlignment = VerticalAlignment.Bottom;
            }
            else
            {
                toolbar.InkOptionsToolbar.HorizontalAlignment = HorizontalAlignment.Left;
                toolbar.InkOptionsToolbar.VerticalAlignment = VerticalAlignment.Stretch;
                toolbar.NavigationOptionsToolbar.HorizontalAlignment = HorizontalAlignment.Right;
                toolbar.NavigationOptionsToolbar.VerticalAlignment = VerticalAlignment.Stretch;
            }
        }

        public static readonly DependencyProperty InkOptionsVisibilityProperty = DependencyProperty.Register(
         "InkOptionsVisibility",
         typeof(Orientation),
         typeof(MagnifyingToolbar),
         new PropertyMetadata(Visibility.Visible, new PropertyChangedCallback(OnInkOptionsVisibilityChanged)));

        public Visibility InkOptionsVisibility
        {
            get { return (Visibility)GetValue(InkOptionsVisibilityProperty); }
            set { SetValue(InkOptionsVisibilityProperty, value); }
        }

        private static void OnInkOptionsVisibilityChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            Visibility vis = (Visibility)e.NewValue;
            (d as MagnifyingToolbar).InkOptionsToolbar.Visibility = vis;
        }

        public static readonly DependencyProperty NavigationOptionsVisibilityProperty = DependencyProperty.Register(
         "NavigationOptionsVisibility",
         typeof(Orientation),
         typeof(MagnifyingToolbar),
         new PropertyMetadata(Visibility.Visible, new PropertyChangedCallback(OnNavigationOptionsVisibilityChanged)));

        public Visibility NavigationOptionsVisibility
        {
            get { return (Visibility)GetValue(NavigationOptionsVisibilityProperty); }
            set { SetValue(NavigationOptionsVisibilityProperty, value); }
        }

        private static void OnNavigationOptionsVisibilityChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            Visibility vis = (Visibility)e.NewValue;
            (d as MagnifyingToolbar).NavigationOptionsToolbar.Visibility = vis;
        }

        public static readonly DependencyProperty DragHandleProperty = DependencyProperty.Register(
         "DragHandle",
         typeof(UIElement),
         typeof(MagnifyingToolbar),
         new PropertyMetadata(null, new PropertyChangedCallback(OnDragHandleChanged)));

        public UIElement DragHandle
        {
            get { return (UIElement)GetValue(DragHandleProperty); }
            set { SetValue(DragHandleProperty, value); }
        }

        private static void OnDragHandleChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            UIElement element = e.NewValue as UIElement;
            (d as MagnifyingToolbar).DragHostElement.Child = element;
        }

        public FrameworkElement DragTarget { get { return DragHostElement; } }

        public MagnifyingToolbar()
        {
            this.InitializeComponent();
        }
    }
}
