﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Media.Audio;
using Windows.Media.MediaProperties;
using Windows.Media.Render;
using Windows.Storage;


namespace pdftron.PDF.Tools.Controls.Multimedia
{
    public class AudioGraphPlayer
    {
        private AudioGraph mGraph;
        private AudioFileInputNode mFileInput;
        private AudioDeviceOutputNode mDeviceOutput;

        public AudioGraphPlayer()
        {

        }

        /// <summary>
        /// Default to initialize in Lossless PCM
        /// </summary>
        public async Task<bool> InitAudioGraphAsync()
        {
            // Create an AudioGraph with lossless PCM settings
            AudioGraphSettings settings = new AudioGraphSettings(AudioRenderCategory.Media);
            settings.EncodingProperties = AudioEncodingProperties.CreatePcm(
                        UtilityFunctions.PCM_AUDIO_SAMPLE_RATE,
                        UtilityFunctions.PCM_AUDIO_NUM_CHANNELS,
                        UtilityFunctions.PCM_AUDIO_BITS_PER_SAMPLE);  

            CreateAudioGraphResult result = await AudioGraph.CreateAsync(settings);

            if (result.Status != AudioGraphCreationStatus.Success)
            {                
                return false;
            }

            mGraph = result.Graph;

            // Create a device output node
            CreateAudioDeviceOutputNodeResult deviceOutputNodeResult = await mGraph.CreateDeviceOutputNodeAsync();

            if (deviceOutputNodeResult.Status != AudioDeviceNodeCreationStatus.Success)
            {
                return false;
            }

            mDeviceOutput = deviceOutputNodeResult.DeviceOutputNode;

            return true;
        }

        public async Task LoadFileAsync(StorageFile file)
        {
            // File can be null if cancel is hit in the file picker
            if (mGraph == null || file == null)
                return;

            // If another file is already loaded into the FileInput node
            if (mFileInput != null)
            {
                // Release the file and dispose the contents of the node
                mFileInput.Dispose();
            }

            CreateAudioFileInputNodeResult fileInputResult = await mGraph.CreateFileInputNodeAsync(file);
            if (AudioFileNodeCreationStatus.Success != fileInputResult.Status)
            {
                // Cannot read input file
                // TODO: error
                return;
            }

            mFileInput = fileInputResult.FileInputNode;

            if (mFileInput.Duration <= TimeSpan.FromSeconds(0))
            {
                // TODO: error
                mFileInput.Dispose();
                mFileInput = null;
                return;
            }

            mFileInput.AddOutgoingConnection(mDeviceOutput);
        }

        public void Play()
        {
            if (mGraph == null || mFileInput == null)
                return;

            mGraph.Start();
        }

        public void Stop()
        {
            if (mGraph == null || mFileInput == null)
                return;

            mGraph.Stop();
        }
    }
}
