﻿using pdftron.PDF.Tools.Controls.ViewModels.Common;
using Windows.UI.Xaml.Navigation;

namespace pdftron.PDF.Tools.Controls
{
    public class NavigablePage : Windows.UI.Xaml.Controls.Page
    {
        protected override void OnNavigatedTo(NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            ActivateViewModel(e.Parameter);
        }

        protected void ActivateViewModel(object parameter)
        {
            if (DataContext is INavigable navigableViewModel)
            {
                navigableViewModel.NewINavigableAvailable += NavigableViewModel_NewINavigableAvailable;
                navigableViewModel.Activate(parameter);
            }
        }

        protected override void OnNavigatingFrom(NavigatingCancelEventArgs e)
        {
            base.OnNavigatingFrom(e);
            DeactivateViewModel(e.Parameter);
        }

        protected void DeactivateViewModel(object parameter)
        {
            if (DataContext is INavigable navigableViewModel)
            {
                DataContext = null;
                navigableViewModel.Deactivate(parameter);
                navigableViewModel.NewINavigableAvailable -= NavigableViewModel_NewINavigableAvailable;
            }
        }

        /// <summary>
        /// Override this to show the new content
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="newNavigable"></param>
        protected virtual void NavigableViewModel_NewINavigableAvailable(INavigable sender, INavigable newNavigable)
        {
        }
    }
}
