﻿using pdftron.Common;
using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class ReflowView : UserControl 
    {
        private const string BLANK_PAGE_STRING_PRE_SCRIPT = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> <html><head xmlns = \"http://www.w3.org/1999/xhtml\" > <style>a:link{color:#24a9db}</style>";
        private const string BLANK_PAGE_STRING_POST_SCRIPT = "</head> <body class=\"big\"></body> </html>";
        private string _BlankPageWithKeyEventsString;


        private Dictionary<int, string> HtmlPages = new Dictionary<int, string>();

        private ReflowViewModel _ReflowViewModel;
        public ReflowViewModel ReflowViewModel
        {
            get { return _ReflowViewModel; }
            set
            {
                ReflowViewModel oldVM = _ReflowViewModel;
                if (_ReflowViewModel != value)
                {
                    _ReflowViewModel = value;
                    if (oldVM != null)
                    {
                        oldVM.PropertyChanged -= ViewModel_PropertyChanged;
                    }
                    if (_ReflowViewModel != null)
                    {
                        _ReflowViewModel.PropertyChanged += ViewModel_PropertyChanged;
                    }
                }
            }
        }

        public int CurrentPage
        {
            get
            {
                if (ReflowViewModel != null)
                {
                    return ReflowViewModel.CurrentPageNumber;
                }
                return 0;
            }
            set
            {
                if (ReflowViewModel != null)
                {
                    ReflowViewModel.SetPageNumber(value);
                }
            }
        }

        public event TappedEventHandler ReflowTapped;

        public event PropertyChangedEventHandler PropertyChanged;
        public void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        public ReflowView()
        {
            Init();
        }

        public ReflowView(PDFDoc doc, int pageNumber)
        {
            Init();

            ReflowViewModel.Doc = doc;
            ReflowViewModel.SetPageNumber(pageNumber);
        }

        private void Init()
        {
            this.InitializeComponent();
            _BlankPageWithKeyEventsString = BLANK_PAGE_STRING_PRE_SCRIPT + ReflowViewModel.KEYDOWN_STRING + ReflowViewModel.TAP_STRING + BLANK_PAGE_STRING_POST_SCRIPT;
            ReflowViewModel = new ReflowViewModel();
            this.DataContext = ReflowViewModel;
            this.PointerWheelChanged += ReflowView_PointerWheelChanged;
            ReflowWebView.Loaded += ReflowView_Loaded;
            ReflowWebView.ScriptNotify += ReflowWebView_ScriptNotify;
            MainGrid.Tapped += MainGrid_Tapped;
            PreviousButtonHorizontal.Tapped += NavigationButtonHorizontal_Tapped;
            NextButtonHorizontal.Tapped += NavigationButtonHorizontal_Tapped;
        }

        private void NavigationButtonHorizontal_Tapped(object sender, TappedRoutedEventArgs e)
        {
            e.Handled = true;
        }

        private void MainGrid_Tapped(object sender, TappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Mouse)
            {
                ReflowTapped?.Invoke(this, new TappedRoutedEventArgs());
            }
        }

        private void ReflowView_Loaded(object sender, RoutedEventArgs e)
        {
            ReflowWebView.Focus(FocusState.Programmatic);
        }

        private void ViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("ReflowedContent"))
            {
                if (ReflowViewModel.HasReflowedContent)
                {
                    ReflowWebView.NavigateToString(ReflowViewModel.ReflowedContent);
                }
                else
                {
                    ReflowWebView.NavigateToString(_BlankPageWithKeyEventsString);
                }
            }
            else if (e.PropertyName.Equals("CurrentPageNumber", StringComparison.OrdinalIgnoreCase))
            {
                string progressText = ResourceHandler.GetString("Reflow_WorkingOnPage");
                ProgressTextBlock.Text = string.Format(progressText, ReflowViewModel.CurrentPageNumber);
            }
            else if (e.PropertyName.Equals("PostProcessingMode", StringComparison.OrdinalIgnoreCase))
            {
                UpdateColors();
            }
        }

        private void ReflowView_PointerWheelChanged(object sender, PointerRoutedEventArgs e)
        {
            int delta = e.GetCurrentPoint(this).Properties.MouseWheelDelta;
            if (delta < 0)
            {
                ReflowViewModel.NavigationCommand.Execute("next");
            }
            else if (delta > 0)
            {
                ReflowViewModel.NavigationCommand.Execute("prev");
            }
        }

        private void ReflowWebView_ScriptNotify(object sender, NotifyEventArgs e)
        {
            if ((e.Value == "tap") || (e.Value == "rightClick"))
            {
                ReflowTapped?.Invoke(this, new TappedRoutedEventArgs());
            }
            else if (!string.IsNullOrWhiteSpace(e.Value) && e.Value.Length > 4 && e.Value.Substring(0, 4).Equals("key:"))
            {
                string[] keysplit = e.Value.Split(new char[] { ' ' });
                if (keysplit.Length == 2)
                {
                    string keyString = keysplit[1];
                    int num = 0;
                    
                    if (int.TryParse(keyString, out num))
                    {
                        if (num == 37)
                        {
                            ReflowViewModel.NavigationCommand.Execute("prev");
                        }
                        else if (num == 39)
                        {
                            ReflowViewModel.NavigationCommand.Execute("next");
                        }
                    }
                    
                    
                }
            }
            else
            {
                ReflowViewModel.NavigateInternalLink(e.Value);
            }
        }

        private void UpdateColors()
        {
            if (ReflowViewModel != null)
            {
                try
                {
                    switch (ReflowViewModel.PostProcessingMode)
                    {
                        case PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map:
                            MainGrid.Background = new SolidColorBrush(ReflowViewModel.PostProcessingWhiteColor);
                            ReflowWebView.DefaultBackgroundColor = ReflowViewModel.PostProcessingWhiteColor;
                            ProgressRing.Foreground = new SolidColorBrush(ReflowViewModel.PostProcessingBlackColor);
                            ProgressTextBlock.Foreground = new SolidColorBrush(ReflowViewModel.PostProcessingBlackColor);
                            break;
                        case PDFRasterizerColorPostProcessMode.e_postprocess_night_mode:
                        case PDFRasterizerColorPostProcessMode.e_postprocess_invert:
                            MainGrid.Background = (SolidColorBrush)Resources["NightModeBackgroundBrush"];
                            ReflowWebView.DefaultBackgroundColor = (Color)Resources["NightModeBackgroundColor"];
                            ProgressRing.Foreground = (SolidColorBrush)Resources["NightModeForegroundBrush"];
                            ProgressTextBlock.Foreground = (SolidColorBrush)Resources["NightModeForegroundBrush"];
                            break;
                        default:
                            MainGrid.Background = (SolidColorBrush)Resources["DayModeBackgroundBrush"];
                            ReflowWebView.DefaultBackgroundColor = (Color)Resources["DayModeBackgroundColor"];
                            ProgressRing.Foreground = (SolidColorBrush)Resources["DayModeForegroundBrush"];
                            ProgressTextBlock.Foreground = (SolidColorBrush)Resources["DayModeForegroundBrush"];
                            break;
                    }
                }
                catch (Exception e)
                {
                    PDFNetException pdfe = new PDFNetException(e.HResult);
                    if (!pdfe.IsPDFNetException)
                    {
                        throw;
                    }
                    else
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                    }
                }
            }
        }

        #region Public Functions

        public void SetColorPostProcessingMode(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
        {
            if (_ReflowViewModel == null)
                return;

            _ReflowViewModel.SetColorPostProcessingMode(mode, whiteColor, blackColor);
        }

        public void DeactivateView()
        {
            if (ReflowViewModel != null)
            {
                ReflowViewModel.ShutDown();
            }
        }

        public void SetFocus(FocusState focusState)
        {
            ReflowWebView.Focus(focusState);
        }

        #endregion Public functions

    }
}
