﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Controls.Resources
{
    class NoTouchPointerOverButton : Windows.UI.Xaml.Controls.Button
    {
        protected override void OnPointerEntered(PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch)
            {
                base.OnPointerEntered(e);
            }
        }

        //private Windows.UI.Xaml.Media.Animation.Storyboard _DelayPressedStoryBoard;
        private bool _WaitingForPressedStoryboard = false;
        private bool _PressedWaitingForReleased = false;
        protected override async void OnPointerPressed(PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
            {
                try
                {
                    DependencyObject pressedObject = this.GetTemplateChild("Pressed");
                    VisualState pressedState = pressedObject as VisualState;
                    if (pressedState != null)
                    {
                        _WaitingForPressedStoryboard = true;
                        _PressedWaitingForReleased = true;
                        await Task.Delay(60);
                        if (_WaitingForPressedStoryboard)
                        {
                            pressedState.Storyboard.Begin();
                            _WaitingForPressedStoryboard = false;
                        }
                    }
                }
                catch (Exception) { }
            }
            else
            {
                base.OnPointerPressed(e);
            }
        }

        protected override async void OnPointerReleased(PointerRoutedEventArgs e)
        {
            if (_WaitingForPressedStoryboard)
            {
                _PressedWaitingForReleased = false;
            }
            base.OnPointerReleased(e);
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
            {
                try
                {
                    DependencyObject pressedObject = this.GetTemplateChild("Restore");
                    VisualState pressedState = pressedObject as VisualState;
                    if (pressedState != null)
                    {
                        await Task.Delay(60);
                        _WaitingForPressedStoryboard = false;
                        pressedState.Storyboard.Begin();
                    }
                }
                catch (Exception) { }
            }
        }

        protected override void OnPointerCanceled(PointerRoutedEventArgs e)
        {
            base.OnPointerCanceled(e);
            if (_PressedWaitingForReleased)
            {

                _WaitingForPressedStoryboard = false;
                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
                {
                    try
                    {
                        DependencyObject pressedObject = this.GetTemplateChild("Restore");
                        VisualState pressedState = pressedObject as VisualState;
                        if (pressedState != null)
                        {
                            pressedState.Storyboard.Begin();
                        }
                    }
                    catch (Exception) { }
                }
            }
        }

        protected override void OnPointerCaptureLost(PointerRoutedEventArgs e)
        {
            base.OnPointerCaptureLost(e);
            if (_PressedWaitingForReleased)
            {
                _WaitingForPressedStoryboard = false;
                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
                {
                    try
                    {
                        DependencyObject pressedObject = this.GetTemplateChild("Restore");
                        VisualState pressedState = pressedObject as VisualState;
                        if (pressedState != null)
                        {
                            pressedState.Storyboard.Begin();
                        }
                    }
                    catch (Exception) { }
                }
            }
        }
    }
}
