﻿using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using System;
using Windows.Storage;
using Windows.Storage.Streams;

namespace pdftron.PDF.Tools.Controls.ViewModels.FileOpening
{
    public class NewDocumentProperties
    {
        public StorageFile File { get; set; }

        public StorageFile TemporaryFile { get; set; }

        public StorageFile OriginalFileFromSystem { get; set; }

        public IRandomAccessStream FileStream { get; set; }

        public PDFDoc Doc { get; set; }

        public OpenedDocumentStates OpenedDocumentState { get; set; }

        public string MRUToken { get; set; }

        public string Password { get; set; }

        public int StartPage { get; set; }
        public PDFViewCtrlPagePresentationMode PresentationMode { get; set; }

        public bool IsReflow { get; set; }

        public PageRotate PageRotation { get; set; }

        public double Zoom { get; set; }

        public double HorizontalScrollPosition { get; set; }

        public double VerticalScrollPosition { get; set; }

        public DateTimeOffset LastModifiedDate { get; set; } = DateTimeOffset.MinValue;

        public ulong Filesize { get; set; }

        public bool OpenedThroughDrop { get; set; }

        public bool RecentPropertiesRestored { get; set; }

        public NewDocumentProperties()
        {
            OpenedDocumentState = OpenedDocumentStates.None;
            MRUToken = string.Empty;
            Password = string.Empty;
            StartPage = 1;

            PresentationMode = ViewerSettings.PagePresentationMode;
        }

        public void ApplyRecentProperties(IRecentDocumentProperties recentFileProperties)
        {
            StartPage = recentFileProperties.PageNumber;
            PageRotation = recentFileProperties.PageRotation;
            PresentationMode = recentFileProperties.PagePresentationMode;
            Zoom = recentFileProperties.Zoom;
            HorizontalScrollPosition = recentFileProperties.HorizontalScrollPos;
            VerticalScrollPosition = recentFileProperties.VerticalScrollPos;
            IsReflow = recentFileProperties.IsInReflowMode;
            RecentPropertiesRestored = true;
        }

        public void ApplyOpenProperties(PdfViewCtrlTabInfo tab)
        {
            StartPage = tab.MetaData.LastPage;
            PageRotation = tab.MetaData.PageRotation;
            PresentationMode = tab.MetaData.PagePresentationMode;
            Zoom = tab.MetaData.Zoom;
            HorizontalScrollPosition = tab.MetaData.HScrollPos;
            VerticalScrollPosition = tab.MetaData.VScrollPos;
            IsReflow = tab.MetaData.IsReflow;
            RecentPropertiesRestored = true;
        }
    }
}
