﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml.Data;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    /// <summary>
    /// This class lets us keep track of which items are currently visible, so that we can know which items to request thumbnails for.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class ObservableCollectionWithVisibleItemsTracking<T> : ObservableCollection<T>, IItemsRangeInfo
    {
        private EventHandler<ItemIndexRange> _visibleRangeChanged;
        public event EventHandler<ItemIndexRange> VisibleRangeChanged
        {
            add
            {
                if (_visibleRangeChanged?.GetInvocationList().Contains(value) != true)
                    _visibleRangeChanged += value;
            }
            remove => _visibleRangeChanged -= value;
        }

        private int _firstIndex;
        private int _lastIndex;
        private ItemIndexRange _previousRange = new ItemIndexRange(0, 1);
        private bool _firstRequest = true;

        public int FirstVisibleIndex => _firstIndex;

        public int LastVisibleIndex => Math.Min(_lastIndex, Count - 1);

        public ItemIndexRange PreviousRange => _previousRange;

        public void Dispose() { } // We maintain no state, so this can be empty

        public void RangesChanged(ItemIndexRange visibleRange, IReadOnlyList<ItemIndexRange> trackedItems)
        {
            if (!_firstRequest && _previousRange.FirstIndex == visibleRange.FirstIndex && _previousRange.LastIndex == visibleRange.LastIndex)
                return;

            _firstRequest = false;
            _firstIndex = visibleRange.FirstIndex;
            _lastIndex = visibleRange.LastIndex;
            _visibleRangeChanged?.Invoke(this, visibleRange);
            _previousRange = visibleRange;
        }
    }
}
