﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ControlBase;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Extensions;
using System;
using System.Threading.Tasks;
using System.Collections.Generic;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Media;
using UIRect = Windows.Foundation.Rect;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class AnnotationToolbarViewModel : ViewModelBase
    {
        #region Constants
        private const int MIN_POINTS_CREATE_POLYGON = 3;
        private const int MIN_POINTS_CREATE_POLYLINE = 2;
        #endregion

        private ICloseableControl _view;
        private bool _subscribed;

#region Commands
        public RelayCommand ToolButtonClickedCommand {get; private set; }
        public RelayCommand InkToolbarCommand { get; private set; }
        public RelayCommand InkPresetButtonCommand { get; private set; }
        public RelayCommand PolygonSaveToolbarCommand { get; private set; }
        public RelayCommand MeasureToolbarCommand { get; private set; }
        public RelayCommand RedactionToolbarCommand { get; private set; }
        public RelayCommand UndoButtonClickedCommand { get; private set; }
        public RelayCommand RedoButtonClickedCommand { get; private set; }
        public RelayCommand ToggleMainAppBarCommand { get; private set; }
        #endregion Commands

#if SUBSCRIPTIONS
        private bool _isRedactionTipVisible = true;
#else
        private bool _isRedactionTipVisible;
#endif
        public bool IsRedactionTipVisible
        {
            get
            {
#if SUBSCRIPTIONS
                if (!ViewerSettings.RedactionExplained)
                {
                    ViewerSettings.RedactionExplained = true;
                    return _isRedactionTipVisible;
                }
                else
                {
                    return false;
                }
#else
                return false;
#endif
            }
            set => Set(ref _isRedactionTipVisible, value);
        }

        /// <summary>
        /// Perform the escape key operation by closing any opened toolbar, saving chagnes, and reseting to pan tool
        /// </summary>
        public void PerformEscapeFromTool()
        {
            ToolType currentToolType = ToolManager.CurrentTool.ToolMode;

            if (currentToolType == ToolType.e_ink_create)
            {
                InkToolbarCommandImpl("S");
            }
            else if (currentToolType != ToolType.e_text_select
                && currentToolType != ToolType.e_pan)
            {
                ToolManager.CreateDefaultTool();
            }
        }

        public bool Subscribed
        {
            get => _subscribed;
            set => Set(ref _subscribed, value);
        }

        public void SetCloseableControl(ICloseableControl control) => _view = control;

#region Tool Visibility

        private bool _isStickyNoteToolButtonVisible = true;
        public bool IsStickyNoteToolButtonVisible
        {
            get { return _isStickyNoteToolButtonVisible; }
            set
            {
                if (value != _isStickyNoteToolButtonVisible)
                {
                    _isStickyNoteToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isHighlightToolButtonVisible = true;
        public bool IsHighlightToolButtonVisible
        {
            get { return _isHighlightToolButtonVisible; }
            set
            {
                if (value != _isHighlightToolButtonVisible)
                {
                    _isHighlightToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isUnderlineToolButtonVisible = true;
        public bool IsUnderlineToolButtonVisible
        {
            get { return _isUnderlineToolButtonVisible; }
            set
            {
                if (value != _isUnderlineToolButtonVisible)
                {
                    _isUnderlineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isStrikeoutToolButtonVisible = true;
        public bool IsStrikeoutToolButtonVisible
        {
            get { return _isStrikeoutToolButtonVisible; }
            set
            {
                if (value != _isStrikeoutToolButtonVisible)
                {
                    _isStrikeoutToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSquigglyToolButtonVisible = true;
        public bool IsSquigglyToolButtonVisible
        {
            get { return _isSquigglyToolButtonVisible; }
            set
            {
                if (value != _isSquigglyToolButtonVisible)
                {
                    _isSquigglyToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isStamperToolButtonVisible = true;
        public bool IsStamperToolButtonVisible
        {
            get { return _isStamperToolButtonVisible; }
            set
            {
                if (value != _isStamperToolButtonVisible)
                {
                    _isStamperToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSignatureToolButtonVisible = true;
        public bool IsSignatureToolButtonVisible
        {
            get { return _isSignatureToolButtonVisible; }
            set
            {
                if (value != _isSignatureToolButtonVisible)
                {
                    _isSignatureToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isInkToolButtonVisible = true;
        public bool IsInkToolButtonVisible
        {
            get { return _isInkToolButtonVisible; }
            set
            {
                if (value != _isInkToolButtonVisible)
                {
                    _isInkToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isEraserToolButtonVisible = true;
        public bool IsEraserToolButtonVisible
        {
            get { return _isEraserToolButtonVisible; }
            set
            {
                if (value != _isEraserToolButtonVisible)
                {
                    _isEraserToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMarkerToolButtonVisible = true;
        public bool IsMarkerToolButtonVisible
        {
            get { return _isMarkerToolButtonVisible; }
            set
            {
                if (value != _isMarkerToolButtonVisible)
                {
                    _isMarkerToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isTextBoxToolButtonVisible = true;
        public bool IsTextBoxToolButtonVisible
        {
            get { return _isTextBoxToolButtonVisible; }
            set
            {
                if (value != _isTextBoxToolButtonVisible)
                {
                    _isTextBoxToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isShapeToolsButtonVisible = true;
        public bool IsShapeToolsButtonVisible
        {
            get { return _isShapeToolsButtonVisible; }
            set
            {
                if (value != _isShapeToolsButtonVisible)
                {
                    _isShapeToolsButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMultiSelectToolButtonVisible = true;
        public bool IsMultiSelectToolButtonVisible
        {
            get { return _isMultiSelectToolButtonVisible; }
            set
            {
                if (value != _isMultiSelectToolButtonVisible)
                {
                    _isMultiSelectToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isLassoToolButtonVisible = true;
        public bool IsLassoToolButtonVisible
        {
            get { return _isLassoToolButtonVisible; }
            set
            {
                if (value != _isLassoToolButtonVisible)
                {
                    _isLassoToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPanToolButtonVisible = true;
        public bool IsPanToolsButtonVisible
        {
            get { return _isPanToolButtonVisible; }
            set
            {
                if (value != _isPanToolButtonVisible)
                {
                    _isPanToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isUndoRedoToolButtonVisible = true;
        public bool IsUndoRedoToolButtonVisible
        {
            get { return _isUndoRedoToolButtonVisible; }
            set
            {
                if (value != _isUndoRedoToolButtonVisible)
                {
                    _isUndoRedoToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isSquareToolButtonVisible = true;
        public bool IsSquareToolButtonVisible
        {
            get { return _isSquareToolButtonVisible; }
            set
            {
                if (value != _isSquareToolButtonVisible)
                {
                    _isSquareToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isCircleToolButtonVisible = true;
        public bool IsCircleToolButtonVisible
        {
            get { return _isCircleToolButtonVisible; }
            set
            {
                if (value != _isCircleToolButtonVisible)
                {
                    _isCircleToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isLineToolButtonVisible = true;
        public bool IsLineToolButtonVisible
        {
            get { return _isLineToolButtonVisible; }
            set
            {
                if (value != _isLineToolButtonVisible)
                {
                    _isLineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isArrowToolButtonVisible = true;
        public bool IsArrowToolButtonVisible
        {
            get { return _isArrowToolButtonVisible; }
            set
            {
                if (value != _isArrowToolButtonVisible)
                {
                    _isArrowToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMeasureToolsButtonVisible = true;
        public bool IsMeasureToolsButtonVisible
        {
            get { return _isMeasureToolsButtonVisible; }
            set
            {
                if (value != _isMeasureToolsButtonVisible)
                {
                    _isMeasureToolsButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isRulerToolButtonVisible = true;
        public bool IsRulerToolButtonVisible
        {
            get { return _isRulerToolButtonVisible; }
            set
            {
                if (value != _isRulerToolButtonVisible)
                {
                    _isRulerToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPerimeterToolButtonVisible = true;
        public bool IsPerimeterToolButtonVisible
        {
            get { return _isPerimeterToolButtonVisible; }
            set
            {
                if (value != _isPerimeterToolButtonVisible)
                {
                    _isPerimeterToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isAreaToolButtonVisible = true;
        public bool IsAreaToolButtonVisible
        {
            get { return _isAreaToolButtonVisible; }
            set
            {
                if (value != _isAreaToolButtonVisible)
                {
                    _isAreaToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPolylineToolButtonVisible = true;
        public bool IsPolylineToolButtonVisible
        {
            get { return _isPolylineToolButtonVisible; }
            set
            {
                if (value != _isPolylineToolButtonVisible)
                {
                    _isPolylineToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isPolygonToolButtonVisible = true;
        public bool IsPolygonToolButtonVisible
        {
            get { return _isPolygonToolButtonVisible; }
            set
            {
                if (value != _isPolygonToolButtonVisible)
                {
                    _isPolygonToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isCloudToolButtonVisible = true;
        public bool IsCloudToolButtonVisible
        {
            get { return _isCloudToolButtonVisible; }
            set
            {
                if (value != _isCloudToolButtonVisible)
                {
                    _isCloudToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isMarqueeZoomToolButtonVisible = true;
        public bool IsMarqueeZoomToolButtonVisible
        {
            get { return _isMarqueeZoomToolButtonVisible; }
            set
            {
                if (value != _isMarqueeZoomToolButtonVisible)
                {
                    _isMarqueeZoomToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isRedactionToolButtonVisible = true;
        public bool IsRedactionToolButtonVisible
        {
            get { return _isRedactionToolButtonVisible; }
            set
            {
                if (value != _isRedactionToolButtonVisible)
                {
                    _isRedactionToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isRedactionTextToolButtonVisible = true;
        public bool IsRedactionTextToolButtonVisible
        {
            get { return _isRedactionTextToolButtonVisible; }
            set
            {
                if (value != _isRedactionTextToolButtonVisible)
                {
                    _isRedactionTextToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isRedactionAreaToolButtonVisible = true;
        public bool IsRedactionAreaToolButtonVisible
        {
            get { return _isRedactionAreaToolButtonVisible; }
            set
            {
                if (value != _isRedactionAreaToolButtonVisible)
                {
                    _isRedactionAreaToolButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isCloseToolbarButtonVisible = true;
        public bool IsCloseToolbarButtonVisible
        {
            get { return _isCloseToolbarButtonVisible; }
            set
            {
                if (value != _isCloseToolbarButtonVisible)
                {
                    _isCloseToolbarButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _isExpandButtonVisible = true;
        public bool IsExpandButtonVisible
        {
            get { return _isExpandButtonVisible; }
            set
            {
                if (value != _isExpandButtonVisible)
                {
                    _isExpandButtonVisible = value;
                    RaisePropertyChanged();
                }
            }
        }

        #endregion

        private ToolManager _ToolManager;
        /// <summary>
        /// Gets or sets the ToolManager
        /// </summary>
        internal ToolManager ToolManager
        {
            get { return _ToolManager; }
            set
            {
                if (_ToolManager != null)
                {
                    _ToolManager.NewToolCreated -= _ToolManager_NewToolCreated;
                    _ToolManager.UndoRedoStatusChanged += _ToolManager_UndoRedoStatusChanged;

                }
                _ToolManager = value;
                if (_ToolManager != null)
                {
                    _ToolManager.NewToolCreated += _ToolManager_NewToolCreated;
                    _ToolManager.UndoRedoStatusChanged += _ToolManager_UndoRedoStatusChanged;
                }
            }
        }

        private bool _ButtonsStayDown;
        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each
        /// annotation is drawn.
        /// </summary>
        internal bool ButtonsStayDown
        {
            get { return _ButtonsStayDown; }
            set
            {
                _ButtonsStayDown = value;
                if (_ToolManager != null)
                {
                    _ToolManager.CurrentTool.ReturnToPanModeWhenFinished = !value;
                }
            }
        }

#region XAML properties

        private bool _showPolygonSaveToolbar = false;
        /// <summary>
        /// Controls whether the regular annotation bar or polygon bar is in view
        /// </summary>
        public bool ShowPolygonSaveToolbar
        {
            get { return _showPolygonSaveToolbar; }
            set
            {
                if (value != _showPolygonSaveToolbar)
                {
                    _showPolygonSaveToolbar = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _ShowInkToolbar = false;
        /// <summary>
        /// Controls whether the regular annotation bar or ink bar is in view
        /// </summary>
        public bool ShowInkToolbar
        {
            get { return _ShowInkToolbar; }
            set
            {
                if (value != _ShowInkToolbar)
                {
                    _ShowInkToolbar = value;
                    RaisePropertyChanged();

                    if (value)
                    {
                        CurrentPresetLabel = Settings.GetPropertyLabel(ToolType.e_ink_create);
                        SetInkPropertiesFromLabel();
                    }
                    else
                    {
                        CurrentPresetLabel = -1;
                    }
                }
            }
        }

        private bool _InkMultiStroke = false;
        /// <summary>
        /// Ink Muilti Stroke option will lock the PDF Viewer to the selected page and enable
        /// undo and redo options for multi strokes
        /// </summary>
        public bool InkMultiStroke
        {
            get { return _InkMultiStroke; }
            set
            {
                if (_InkMultiStroke != value)
                {
                    _InkMultiStroke = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _ShowMeasureToolbar = false;
        /// <summary>
        /// Controls whether the regular annotation bar or measure bar is in view
        /// </summary>
        public bool ShowMeasureToolbar
        {
            get { return _ShowMeasureToolbar; }
            set
            {
                if (value != _ShowMeasureToolbar)
                {
                    _ShowMeasureToolbar = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _ShowRedactionToolbar = false;
        public bool ShowRedactionToolbar
        {
            get { return _ShowRedactionToolbar; }
            set
            {
                if (value != _ShowRedactionToolbar)
                {
                    _ShowRedactionToolbar = value;
                    RaisePropertyChanged();
                }
            }
        }

        // Control whethers the ink app bar or regular tool bar or polygon tool bar is expanded or compact
        public bool _IsMainAppBarOpen = false;
        public bool IsMainAppBarOpen
        {
            get { return _IsMainAppBarOpen; }
            set
            {
                if (value != _IsMainAppBarOpen)
                {
                    _IsMainAppBarOpen = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanUndo = false;
        public bool CanUndo
        {
            get { return _CanUndo; }
            set
            {
                if (value != _CanUndo)
                {
                    _CanUndo = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanRedo = false;
        public bool CanRedo
        {
            get { return _CanRedo; }
            set
            {
                if (value != _CanRedo)
                {
                    _CanRedo = value;
                    RaisePropertyChanged();
                }
            }
        }

        private string _UndoToolTip = "Undo";
        public string UndoToolTip
        {
            get { return _UndoToolTip; }
            set
            {
                if (value != _UndoToolTip)
                {
                    _UndoToolTip = value;
                    RaisePropertyChanged();
                }
            }
        }

        private string _RedoToolTip = "Redo";
        public string RedoToolTip
        {
            get { return _RedoToolTip; }
            set
            {
                if (value != _RedoToolTip)
                {
                    _RedoToolTip = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanRedoInk = false;
        public bool CanRedoInk
        {
            get { return _CanRedoInk; }
            set
            {
                if (value != _CanRedoInk)
                {
                    _CanRedoInk = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _CanSaveInk = false;
        public bool CanSaveInk
        {
            get { return _CanSaveInk; }
            set
            {
                if (value != _CanSaveInk)
                {
                    _CanSaveInk = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _canSavePolygon = false;
        public bool CanSavePolygon
        {
            get { return _canSavePolygon; }
            set
            {
                if (value != _canSavePolygon)
                {
                    _canSavePolygon = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _canSaveMeasure = false;
        public bool CanSaveMeasure
        {
            get { return _canSaveMeasure; }
            set
            {
                if (value != _canSaveMeasure)
                {
                    _canSaveMeasure = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _GotToolFromToolManager = false;
        private ToolType _CurrentTool = ToolType.e_pan;
        public ToolType CurrentTool
        {
            get { return _CurrentTool; }
            private set
            {
                if (_CurrentTool != value)
                {
                    _CurrentTool = value;

                    if (!_GotToolFromToolManager)
                    {
                        SetCurrentTool();
                    }

                    RaisePropertyChanged();
                    RaisePropertyChanged(nameof(PanSelected));
                }
            }
        }

        public bool PanSelected { get { return _CurrentTool == ToolType.e_pan; } }

        private double _AppBarCompactHeight = 48;
        public double AppBarCompactHeight
        {
            get { return _AppBarCompactHeight; }
            set { Set(ref _AppBarCompactHeight, value); }
        }

        private Brush _currentToolColor;
        /// <summary>
        /// TODO:
        /// </summary>
        public Brush CurrentToolColor
        {
            get
            {
                return _currentToolColor;
            }
            set
            {
                if (_currentToolColor != value)
                {
                    _currentToolColor = value;
                    RaisePropertyChanged();
                }
            }
        }

#endregion XAML properties

        private FreehandCreate _CurrentInkTool;

        public delegate void ToolSelectedDelegate();
        public event ToolSelectedDelegate ToolSelected;

        /// <summary>
        /// Creates a new AnnotationToolbar that is not associated with a ToolManager.
        /// Use the ToolManager property to make the AnnotationToolbar Work with a ToolManager.
        /// </summary>
        public AnnotationToolbarViewModel(ICloseableControl view)
        {
            _view = view;
            Init();
        }

        /// <summary>
        /// Creates a new AnnotationToolbar that works with the ToolManager.
        /// </summary>
        /// <param name="ctrl"></param>
        /// <param name="toolManager"></param>
        public AnnotationToolbarViewModel(ICloseableControl view, ToolManager toolManager)
        {
            _view = view;
            Init();
            ToolManager = toolManager;
        }

        private void _ToolManager_UndoRedoStatusChanged(UndoRedoStatusEventArgs e)
        {
            CanUndo = e.CanUndo;
            CanRedo = e.CanRedo;
            UndoToolTip = e.UndoInfo;
            RedoToolTip = e.RedoInfo;
        }

        private void Init()
        {
            // set up relay commands
            ToolButtonClickedCommand = new RelayCommand(ToolButtonClicked);
            InkToolbarCommand = new RelayCommand(InkToolbarCommandImpl);
            InkPresetButtonCommand = new RelayCommand(InkPresetButtonCommandImpl);
            PolygonSaveToolbarCommand = new RelayCommand(PolygonSaveToolbarCommandImpl);
            MeasureToolbarCommand = new RelayCommand(MeasureToolbarCommandImpl);
            RedactionToolbarCommand = new RelayCommand(RedactionToolbarCommandImpl);
            UndoButtonClickedCommand = new RelayCommand(UndoButtonCommandImpl);
            RedoButtonClickedCommand = new RelayCommand(RedoButtonCommandImpl);
            ToggleMainAppBarCommand = new RelayCommand(ToggleMainAppBarCommandImpl);
            ReadyInkPresets();

            CurrentToolColor = new SolidColorBrush(Colors.Gray);
        }

        public void ToggleMainAppBarCommandImpl(object option) => IsMainAppBarOpen = !IsMainAppBarOpen;

        public void UndoButtonCommandImpl(object option)
        {
            if (_ToolManager?.UndoRedoAction != null)
            {
                _ToolManager.UndoRedoAction.DoUndo();
            }
        }

        public void RedoButtonCommandImpl(object option)
        {
            if (_ToolManager?.UndoRedoAction != null)
            {
                _ToolManager.UndoRedoAction.DoRedo();
            }
        }

        public void ToolButtonClicked(object option)
        {
            Button button = option as Button;
            if (button != null)
            {
                string toolName = button.Tag as string;

                if (toolName == "close")
                {
                    HandleCloseControl();
                    return;
                }

                ToolType toolType = GetToolTypeFromString(toolName);
#if SUBSCRIPTIONS
                if (toolType == ToolType.e_rect_redaction_create || toolType == ToolType.e_text_redaction)
                {
                    if (!Subscribed)
                    {
                        return;
                    }
                }
#endif

                if (CurrentTool == toolType)
                {
                    if (Settings.HasSettableProperties(CurrentTool) && CurrentTool != ToolType.e_ink_create)
                    {
                        ShowPopup(button);
                    }
                }
                else
                {
                    Utilities.AppCenterHandlerBase.CURRENT.SendEvent(toolType, new Dictionary<string, string>() { ["Tool Selected"] = toolName });

                    if ((CurrentTool == ToolType.e_polygon_create)
                        || (CurrentTool == ToolType.e_polyline_create)
                        || (CurrentTool == ToolType.e_cloud_create))
                    {
                        CommitPolygonTools();
                    }
                    else if ((CurrentTool == ToolType.e_area_create)
                        || (CurrentTool == ToolType.e_perimeter_create))
                    {
                        CommitMeasureTools();
                    }

                    CurrentTool = toolType;
                }

                if (ToolSelected != null)
                {
                    ToolSelected();
                    IsMainAppBarOpen = false;
                }
            }
        }

        void _ToolManager_NewToolCreated(ToolType toolType)
        {
            if (toolType == _CurrentTool)
            {
                return;
            }
            _GotToolFromToolManager = true;
            CurrentTool = toolType;
            _GotToolFromToolManager = false;

            CanSaveMeasure = false;
            CanSavePolygon = false;
        }

        ToolType FilterOutInvalidTools(ToolType toolType)
        {
            if (toolType == ToolType.e_annot_edit || toolType == ToolType.e_line_edit
                || toolType == ToolType.e_link_action || toolType == ToolType.e_none || toolType == ToolType.e_rich_media
                || toolType == ToolType.e_text_select)
            {
                return ToolType.e_pan;
            }
            return toolType;
        }

#region Polygon Save Toolbar

        private void PolygonSaveToolbarCommandImpl(Object option)
        {
            string opt = option as string;
            if (opt.Equals("C", StringComparison.OrdinalIgnoreCase))
            {
                CommitPolygonTools();
                ShowPolygonSaveToolbar = false;
                IsMainAppBarOpen = false;
                ToolManager.CreateTool(ToolType.e_pan, null);
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Close poly shape tools");
            }
            else if (opt.Equals("S", StringComparison.OrdinalIgnoreCase))
            {
                CommitPolygonTools();
            }
        }

        private void CommitPolygonTools()
        {
            if (_CurrentTool == ToolType.e_polyline_create)
            {
                PolylineCreate polylineTool = this.ToolManager.CurrentTool as PolylineCreate;
                if (polylineTool != null)
                    polylineTool.CommitAnnotation();
            }
            else if (_CurrentTool == ToolType.e_polygon_create)
            {
                PolygonCreate polygonTool = this.ToolManager.CurrentTool as PolygonCreate;
                if (polygonTool != null)
                    polygonTool.CommitAnnotation();
            }
            else if (_CurrentTool == ToolType.e_cloud_create)
            {
                CloudCreate cloudTool = this.ToolManager.CurrentTool as CloudCreate;
                if (cloudTool != null)
                    cloudTool.CommitAnnotation();
            }
        }

#endregion

#region InkToolbar

        /// <summary>
        /// Perform Freehand/Ink Undo
        /// </summary>
        public void PerformInkUndo()
        {
            if (_CurrentInkTool != null && _CurrentInkTool.CanUndoStroke())
                InkToolbarCommandImpl("U");
        }

        /// <summary>
        /// Perform Freehand/Ink Redo
        /// </summary>
        public void PerformInkRedo()
        {
            if (_CurrentInkTool != null && _CurrentInkTool.CanRedoStroke())
                InkToolbarCommandImpl("R");
        }

        private void CurrentInkTool_StrokeAdded()
        {
            CanSaveInk = true;
            CanRedoInk = false;
        }

        /// <summary>
        /// Perform ink toolbar operation according to operation param
        /// </summary>
        private void InkToolbarCommandImpl(Object option)
        {
            string opt = option as string;
            if (opt.Equals("S", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.CommitAnnotation();
                _ToolManager.CreateTool(ToolType.e_pan, null);
                ShowInkToolbar = false;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Save button selected");
            }
            else if (opt.Equals("U", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.UndoStroke();
                if (!_CurrentInkTool.CanUndoStroke())
                {
                    CanSaveInk = false;
                }
                CanRedoInk = true;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Undo button selected");
            }
            else if (opt.Equals("R", StringComparison.OrdinalIgnoreCase))
            {
                _CurrentInkTool.RedoStroke();
                if (!_CurrentInkTool.CanRedoStroke())
                {
                    CanRedoInk = false;
                }
                CanSaveInk = true;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Redo button selected");
            }
            else
            {
                ShowInkToolbar = false;
                IsMainAppBarOpen = false;
                ToolManager.CreateTool(ToolType.e_pan, null);
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Freehand Cancel button selected");
            }
        }

        private void ReadyInkPresets()
        {
            int label = Settings.GetPropertyLabel(ToolType.e_ink_create);
            _CurrentPresetLabel = label; // Don't want to edit the general ink when we initialize
            for (int i = 0; i < 3; i++)
            {
                SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, i);
                _PresetBrushes[i] = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
            }
        }

        private int _CurrentPresetLabel = -1;
        public int CurrentPresetLabel
        {
            get { return _CurrentPresetLabel; }
            set
            {
                if (value != _CurrentPresetLabel)
                {
                    _CurrentPresetLabel = value;
                    if (_CurrentPresetLabel < 100)
                    {
                        // Only save it if valid preset color index
                        if (value > -1)
                            Settings.SetPropertylLabel(ToolType.e_ink_create, _CurrentPresetLabel);

                        SetInkPropertiesFromLabel();

                        if (_CurrentInkTool != null)
                            _CurrentInkTool.IsEraser = false;
                    }
                    else
                    {
                        Settings.SetPropertylLabel(ToolType.e_ink_eraser, _CurrentPresetLabel);
                        SetEraserProperties();
                    }
                    RaisePropertyChanged();

                    // NOTE: Once a new preset color is selected update the Ink Annotation selected color
                    CurrentToolColor = Settings.GetAnnotationColor(ToolType.e_ink_create);
                }
            }
        }

        private void SetInkPropertiesFromLabel()
        {
            SettingsColor color = Settings.GetShapeColor(ToolType.e_ink_create, false, _CurrentPresetLabel);
            Settings.SetShapeColor(ToolType.e_ink_create, false, color);
            switch (CurrentPresetLabel)
            {
                case 0:
                    PresetBrush0 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 1:
                    PresetBrush1 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
                case 2:
                    PresetBrush2 = new SolidColorBrush(UtilityFunctions.ConvertToColor(color));
                    break;
            }
            double thickness = Settings.GetThickness(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetThickness(ToolType.e_ink_create, thickness);
            double opacity = Settings.GetOpacity(ToolType.e_ink_create, _CurrentPresetLabel);
            Settings.SetOpacity(ToolType.e_ink_create, opacity);

            if (_CurrentInkTool != null)
            {
                _CurrentInkTool.SetInkProperties(thickness, opacity, UtilityFunctions.ConvertToColor(color));
            }
        }

        private void SetEraserProperties()
        {
            if (CurrentPresetLabel >= 100 && _CurrentInkTool != null)
            {
                double thickness = Settings.GetThickness(ToolType.e_ink_eraser, CurrentPresetLabel);
                Windows.UI.Color eraserColor = FreehandCreate.DEFAULT_INK_ERASER_COLOR; ;
                double opacity = FreehandCreate.DEFAULT_INK_ERASER_OPACITY;
                _CurrentInkTool.SetEraserProperties(true, thickness, opacity, eraserColor);
                _CurrentInkTool.IsEraser = true;
            }
        }

        private SolidColorBrush[] _PresetBrushes = new SolidColorBrush[3];
        public SolidColorBrush PresetBrush0
        {
            get {
                return _PresetBrushes[0];
            }
            private set
            {
                _PresetBrushes[0] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush1
        {
            get { return _PresetBrushes[1]; }
            private set
            {
                _PresetBrushes[1] = value;
                RaisePropertyChanged();
            }
        }
        public SolidColorBrush PresetBrush2
        {
            get { return _PresetBrushes[2]; }
            private set
            {
                _PresetBrushes[2] = value;
                RaisePropertyChanged();
            }
        }

        private void InkPresetButtonCommandImpl(object obj)
        {
            Button button = obj as Button;
            if (button != null)
            {
                string tag = button.Tag as string;
                if (!string.IsNullOrEmpty(tag))
                {
                    int result = -1;
                    bool success = int.TryParse(tag, out result);
                    if (success)
                    {
                        if (result == CurrentPresetLabel)
                        {
                            ShowPopup(button);
                        }
                        else
                        {
                            CurrentPresetLabel = result;

                            if (InkMultiStroke == false)
                                CurrentTool = ToolType.e_ink_create;
                        }
                    }
                }
            }
        }

        #endregion InkToolbar

        #region Measure ToolBar

        /// <summary>
        /// Perform the saving of current measurement changes, set tool to pan and close toolbar
        /// </summary>
        private void MeasureToolbarCommandImpl(object option)
        {
            string opt = option as string;
            if (opt.Equals("C", StringComparison.OrdinalIgnoreCase))
            {
                CommitMeasureTools();

                _ToolManager.CreateTool(ToolType.e_pan, null);
                ShowMeasureToolbar = false;

                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "Closed Measure toolbar");
            }
            else if (opt.Equals("S", StringComparison.OrdinalIgnoreCase))
            {
                CommitMeasureTools();
            }
        }

        private void CommitMeasureTools()
        {
            if (CurrentTool == ToolType.e_perimeter_create)
            {
                PerimeterMeasureCreate perimeterTool = this.ToolManager.CurrentTool as PerimeterMeasureCreate;
                if (perimeterTool != null)
                {
                    perimeterTool.CommitAnnotation();
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "commit perimeter annotation");
                }
            }
            else if (CurrentTool == ToolType.e_area_create)
            {
                AreaMeasureCreate areaTool = this.ToolManager.CurrentTool as AreaMeasureCreate;
                if (areaTool != null)
                {
                    areaTool.CommitAnnotation();
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Annotation Toolbar", "commit area annotation");
                }
            }
        }

#endregion

#region Redaction Toolbar
        private void RedactionToolbarCommandImpl(object option)
        {
            string opt = option as string;
            if (opt.Equals("C", StringComparison.OrdinalIgnoreCase))
            {
                _ToolManager.CreateTool(ToolType.e_pan, null);
                ShowRedactionToolbar = false;
            }
            else if (opt.Equals("A", StringComparison.OrdinalIgnoreCase))
            {
                ToolManager.ApplyRedaction(true);
            }
        }
#endregion

#region Utilities

        public static ToolType GetToolTypeFromString(string option)
        {
            if (option.Equals("Pan", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_pan;
            }
            if (option.Equals("Line", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_line_create;
            }
            if (option.Equals("Arrow", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_arrow_create;
            }
            if (option.Equals("Square", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_rect_create;
            }
            if (option.Equals("Oval", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_oval_create;
            }
            if (option.Equals("Ink", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_create;
            }
            if (option.Equals("Eraser", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_eraser;
            }
            if (option.Equals("Marker", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ink_marker;
            }
            if (option.Equals("Signature", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_signature;
            }
            if (option.Equals("Stamper", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_stamper;
            }
            if (option.Equals("TextBox", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_annot_create;
            }
            if (option.Equals("Sticky", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_sticky_note_create;
            }
            if (option.Equals("Highlight", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_highlight;
            }
            if (option.Equals("Underline", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_underline;
            }
            if (option.Equals("Strikeout", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_strikeout;
            }
            if (option.Equals("Squiggly", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_squiggly;
            }
            if (option.Equals("Cloud", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_cloud_create;
            }
            if (option.Equals("Polygon", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_polygon_create;
            }
            if (option.Equals("Polyline", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_polyline_create;
            }
            if (option.Equals("annotation group", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_annot_edit_rect_group;
            }
            if (option.Equals("annotation group create", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_annot_edit_group_create;
            }
            if (option.Equals("lasso create", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_lasso_create;
            }
            if (option.Equals("Zoom", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_zoom_marquee;
            }
            if (option.Equals("Ruler", StringComparison.OrdinalIgnoreCase)
                || option.Equals("Measure", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_ruler_create;
            }
            if (option.Equals("Area", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_area_create;
            }
            if (option.Equals("Perimeter", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_perimeter_create;
            }
            if ((option.Equals("Redaction", StringComparison.OrdinalIgnoreCase))
                || (option.Equals("RectRedaction", StringComparison.OrdinalIgnoreCase)))
            {
                return ToolType.e_rect_redaction_create;
            }
            if (option.Equals("TextRedaction", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_text_redaction;
            }
            if (option.Equals("None", StringComparison.OrdinalIgnoreCase))
            {
                return ToolType.e_none;
            }
            return ToolType.e_none;
        }

        private void SetCurrentTool()
        {
            if (_ToolManager == null)
                return;

            if (CurrentTool != ToolType.e_ink_eraser && CurrentTool != ToolType.e_ink_create)
                ShowInkToolbar = false;

            bool shouldButtonStayDown = this.ButtonsStayDown;
            _ToolManager.CreateTool(CurrentTool, null, ButtonsStayDown);

            switch (CurrentTool)
            {
                case ToolType.e_sticky_note_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_text_highlight:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_text_underline:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_text_squiggly:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_text_strikeout:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_rect_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_oval_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_annot_edit_rect_group:
                case ToolType.e_annot_edit_group_create:
                case ToolType.e_lasso_create:
                case ToolType.e_pan:
                    shouldButtonStayDown = false;
                    break;

                case ToolType.e_ink_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);

                    IsMainAppBarOpen = false;
                    shouldButtonStayDown = false;
                    ShowInkToolbar = true;

                    _CurrentInkTool = _ToolManager.CurrentTool as FreehandCreate;
                    if (_CurrentInkTool != null)
                    {
                        if (InkMultiStroke)
                        {
                            _CurrentInkTool.MultiStrokeMode = true;
                            _CurrentInkTool.StrokeAdded += CurrentInkTool_StrokeAdded;
                            CanSaveInk = false;
                            CanRedoInk = false;

                            _CurrentInkTool.RetainInkWhenScalingOrSizing = true;
                        }
                        else
                        {
                            _CurrentInkTool.MultiStrokeMode = false;
                        }
                    }
                    break;

                case ToolType.e_line_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_arrow_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_polyline_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    CanSavePolygon = false;

                    var polylineTool = _ToolManager.CurrentTool as PolylineCreate;
                    if (polylineTool != null)
                    {
                        polylineTool.PointCreated += PolylineTool_PointCreated;
                    }
                    break;

                case ToolType.e_cloud_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);

                    var cloudTool = _ToolManager.CurrentTool as CloudCreate;
                    if (cloudTool != null)
                    {
                        cloudTool.PointCreated += CloudTool_PointCreated;
                    }
                    break;

                case ToolType.e_polygon_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);

                    IsMainAppBarOpen = false;
                    shouldButtonStayDown = false;
                    ShowPolygonSaveToolbar = true;
                    CanSavePolygon = false;

                    var polygonTool = _ToolManager.CurrentTool as PolygonCreate;
                    if (polygonTool != null)
                    {
                        polygonTool.PointCreated += PolygonTool_PointCreated;
                    }

                    break;

                case ToolType.e_ink_marker:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);

                    _CurrentInkTool = _ToolManager.CurrentTool as FreehandCreate;
                    if (_CurrentInkTool != null)
                    {
                        _CurrentInkTool.MultiStrokeMode = false;
                    }
                    break;
                case ToolType.e_signature:
                    shouldButtonStayDown = false;
                    break;
               case ToolType.e_stamper:
                    shouldButtonStayDown = false;
                    break;

                case ToolType.e_text_annot_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    break;

                case ToolType.e_zoom_marquee:
                    // NOTE: No action just guarantee it does not fall in e_none
                    break;

                case ToolType.e_ruler_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);

                    IsMainAppBarOpen = false;
                    ShowMeasureToolbar = true;
                    CanSaveMeasure = false;
                    break;

                case ToolType.e_area_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    CanSaveMeasure = false;

                    var areaTool = _ToolManager.CurrentTool as AreaMeasureCreate;
                    if (areaTool != null)
                    {
                        areaTool.PointCreated += AreaTool_PointCreated;
                    }
                    break;

                case ToolType.e_perimeter_create:
                    CurrentToolColor = Settings.GetAnnotationColor(CurrentTool);
                    CanSaveMeasure = false;

                    var perimeterTool = _ToolManager.CurrentTool as PerimeterMeasureCreate;
                    if (perimeterTool != null)
                    {
                        perimeterTool.PointCreated += PerimeterTool_PointCreated;
                    }
                    break;

                case ToolType.e_rect_redaction_create:
                    IsMainAppBarOpen = false;
                    ShowRedactionToolbar = true;
                    break;

                case ToolType.e_none: // NOTE: if none tool is selected, close control?
                    HandleCloseControl();
                    break;
            }
        }

        /// <summary>
        /// Update the current selected tool to a new one
        /// </summary>
        /// <param name="newCurrentTool">the tool to be selected</param>
        public void UpdateCurrentToolTo(ToolType newCurrentTool)
        {
            CurrentTool = newCurrentTool;
        }

        /// <summary>
        /// Perform the closing of the toolbar control
        /// </summary>
        private void HandleCloseControl()
        {
            IsMainAppBarOpen = false;
            _view.CloseControl();
            _ToolManager.CreateTool(ToolType.e_pan, null);
        }

        private void CloudTool_PointCreated(int total)
        {
            if (total >= MIN_POINTS_CREATE_POLYGON)
                CanSavePolygon = true;
        }

        private void AreaTool_PointCreated(int total)
        {
            if (total >= MIN_POINTS_CREATE_POLYGON)
                CanSaveMeasure = true;
        }

        private void PolygonTool_PointCreated(int total)
        {
            if (total >= MIN_POINTS_CREATE_POLYGON)
                CanSavePolygon = true;
        }

        private void PerimeterTool_PointCreated(int total)
        {
            if (total >= MIN_POINTS_CREATE_POLYLINE)
                CanSaveMeasure = true;
        }

        private void PolylineTool_PointCreated(int total)
        {
            if (total >= MIN_POINTS_CREATE_POLYLINE)
                CanSavePolygon = true;
        }

#endregion Utilities

#region Annotation Properties

        private Popup _Popup;
        private FrameworkElement _Target;
        private AnnotationPropertyDialog _Dialog;
        private void ShowPopup(FrameworkElement target)
        {
            _Target = target;

            _Dialog = new AnnotationPropertyDialog();
            _Dialog.ViewModel.ArrowHeadEnabled = true;

            UIRect rect = UtilityFunctions.GetElementRect(_Target);
            if (rect.Height > AppBarCompactHeight)
            {
                rect.Height = AppBarCompactHeight;
            }

            if (CurrentTool == ToolType.e_ink_create)
            {
                _Dialog.ViewModel.Sequencenumber = _CurrentPresetLabel;
            }
            if (_CurrentPresetLabel >= 100)
            {
                _Dialog.ViewModel.AssociatedTool = ToolType.e_ink_eraser;
            }
            else
            {
                _Dialog.ViewModel.AssociatedTool = CurrentTool;
            }

            AnnotationPropertiesDisplay display = AnnotationPropertiesDisplay.CreateAnnotationPropertiesDisplay(rect, _Dialog);
            display.AnnotationPropertyDialogClosed += display_AnnotationPropertyDialogClosed;
        }

        /// <summary>
        /// Once the Property Dialog is closed a few properties can be set
        /// </summary>
        /// <param name="dialog"></param>
        void display_AnnotationPropertyDialogClosed(AnnotationPropertyDialog dialog)
        {
            var color = dialog.ViewModel.MainColor;
            var brush = new SolidColorBrush(Windows.UI.Color.FromArgb(255, color.R, color.G, color.B));
            CurrentToolColor = brush;

            switch (CurrentTool)
            {
                case ToolType.e_ink_create:

                    if (_CurrentPresetLabel >= 100)
                    {
                        SetEraserProperties();
                    }
                    else
                    {
                        SetInkPropertiesFromLabel();
                    }
                    break;
            }
        }

        void _Popup_Closed(object sender, object e)
        {
            _Popup = null;
            _Dialog = null;

            if (CurrentTool == ToolType.e_ink_create)
            {
                SetInkPropertiesFromLabel();
            }
        }

#endregion Annotation Properties

#region Public Interface

        public override bool GoBack()
        {
            if (_Popup != null)
            {
                _Popup.IsOpen = false;
                return true;
            }
            if (ShowInkToolbar)
            {
                ShowInkToolbar = false;
                ToolManager.CreateDefaultTool();
                return true;
            }
            if (ShowPolygonSaveToolbar)
            {
                ShowPolygonSaveToolbar = false;
                ToolManager.CreateDefaultTool();
                return true;
            }


            return false;
        }
#endregion Public Interface

        public virtual bool GetSubscribed() => false;

        /// <summary>
        /// Commit changes for the selected tool and close any opened secondary toolbar
        /// </summary>
        /// <param name="currentTool">tools to commit changes</param>
        internal void CommitCurrentToolChanges(ToolType currentTool)
        {
            if (currentTool == ToolType.e_ink_create)
            {
                _CurrentInkTool.CommitAnnotation();
            }
            else if (currentTool == ToolType.e_perimeter_create 
                || currentTool == ToolType.e_area_create)
            {
                CommitMeasureTools();
            }
            else if (currentTool == ToolType.e_polyline_create 
                || currentTool == ToolType.e_polygon_create 
                || currentTool == ToolType.e_cloud_create)
            {
                CommitPolygonTools();

            }

            ShowInkToolbar = false;
            ShowMeasureToolbar = false;
            ShowPolygonSaveToolbar = false;
            ShowRedactionToolbar = false;
        }
    }
}
