﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.SDF;
using System;
using System.Diagnostics;
using System.Linq;
using Windows.ApplicationModel.Resources;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Password
{
    public class PasswordFileViewModel : ViewModelBase
    {
        private bool _isPopupOpen;
        private string _currentPassword;
        private string _isProtectedText;
        private bool _isProtected;

        private PopupClosedDelegate _popupClosed;
        public delegate void PopupClosedDelegate();
        public event PopupClosedDelegate PopupClosed
        {
            add
            {
                if (_popupClosed?.GetInvocationList().Contains(value) != true)
                    _popupClosed += value;
            }
            remove => _popupClosed -= value;
        }

        private PasswordConfirmedDelegate _passwordConfirmed;
        public delegate void PasswordConfirmedDelegate();
        public event PasswordConfirmedDelegate PasswordConfirmed
        {
            add
            {
                if (_passwordConfirmed?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _passwordConfirmed += value;
            }
            remove
            {
                if (_passwordConfirmed != null)
                {
                    PasswordConfirmedDelegate eventMethod = (PasswordConfirmedDelegate)Array.Find(
                        _passwordConfirmed.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _passwordConfirmed -= eventMethod;
                }
            }
        }

        public PasswordFileViewModel(/*bool isProtected*/)
        {
            //IsProtected = isProtected;
            //if (isProtected)
            //{
            //    ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            //    IsProtectedText = loader.GetString("PasswordPopup_PlaceHolderText");
            //}

            Init();
        }

        public bool IsPopupOpen
        {
            get => _isPopupOpen;
            set
            {
                if (Set(ref _isPopupOpen, value))
                {
                    if (!value)
                    {
                        _popupClosed?.Invoke();
                    }
                    else
                    {
                        SetDefaultValues();
                    }
                }
            }
        }

        public string CurrentPassword
        {
            get => _currentPassword;
            set
            {
                if (Set(ref _currentPassword, value))
                    RaisePropertyChanged(nameof(IsOkEnabled));
            }
        }

        public string IsProtectedText
        {
            get => _isProtectedText;
            set => Set(ref _isProtectedText, value);
        }

        //public bool IsProtected
        //{
        //    get => _isProtected;
        //    set
        //    {
        //        if (Set(ref _isProtected, value))
        //            RaisePropertyChanged(nameof(IsOkEnabled));
        //    }
        //}

        //public bool IsOkEnabled => IsProtected || (!IsProtected && !string.IsNullOrEmpty(CurrentPassword));
        public bool IsOkEnabled => !string.IsNullOrEmpty(CurrentPassword);

        private void Init()
        {
            CancelCommand = new RelayCommand(CancelCommandImpl);
            OkCommand = new RelayCommand(OkCommandImpl);
        }

        private void SetDefaultValues() => CurrentPassword = string.Empty;

        public RelayCommand CancelCommand { get; private set; }

        public RelayCommand OkCommand { get; private set; }

        private void CancelCommandImpl(object command) => IsPopupOpen = false;

        private void OkCommandImpl(object command)
        {
            _isPopupOpen = false;
            RaisePropertyChanged(nameof(IsPopupOpen));
            _passwordConfirmed?.Invoke();
        }

        public Tuple<bool, string> ApplyPassword(PDFDoc doc)
        {
            if (doc == null)
                return null;

            bool everythingIsOk = false;

            doc.Lock();
            const bool shouldUnlock = true;
            try
            {
                doc.RemoveSecurity();
                if (!string.IsNullOrEmpty(CurrentPassword))
                {
                    SecurityHandler handler = new SecurityHandler();
                    handler.ChangeRevisionNumber(6);
                    handler.ChangeUserPassword(CurrentPassword);

                    handler.SetPermission(SecurityHandlerPermission.e_print, true);
                    handler.SetPermission(SecurityHandlerPermission.e_extract_content, false);

                    doc.SetSecurityHandler(handler);
                    everythingIsOk = true;
                }
                else
                {
                    everythingIsOk = true;
                }
            }
            catch (Exception ex)
            {
                everythingIsOk = false;
                Debug.WriteLine(ex.StackTrace);
                //Utilities.AnalyticsHandler.CURRENT.SendEvent(Utilities.AnalyticsHandler.EXCEPTION_APP, pdftron.PDF.Tools.Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                    doc.Unlock();
            }

            return new Tuple<bool, string>(everythingIsOk, CurrentPassword);
        }
    }
}
