﻿using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using System;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab
{
    public class PdfViewCtrlTabMetaData : ViewModelBase
    {
        public string TabTitle { get; set; }

        public double HScrollPos { get; set; }

        public double VScrollPos { get; set; } = 0;

        public double Zoom { get; set; }

        public int LastPage { get; set; } = 1;

        public PageRotate PageRotation { get; set; }

        public PDFViewCtrlPagePresentationMode PagePresentationMode { get; set; } = ViewerSettings.PagePresentationMode;

        public bool IsReflow { get; set; }

        public bool IsRightToLeft { get; set; }

        public DateTime TabAddedTimeStamp { get; set; }

        public DateTime TabLastViewedTimeStamp { get; set; }

        public string Password { get; set; } = string.Empty;

        /// <summary>
        /// The token  of the file in the App's temporary folder that was opened
        /// </summary>
        public string FutureAccessListToken { get; set; }

        /// <summary>
        /// The file that is meant to be opened, which we save to in case of a normal save.
        /// </summary>
        public string OriginalFileToken { get; set; }

        /// <summary>
        /// In case of non-PDFs, and some read-only files, we keep a reference to the file the user initially picked, even if we can't modify it.
        /// This lets us avoid opening duplicates.
        /// </summary>
        public string FileSourceTokenIfNotSaveable { get; set; }

        public string ThumbnailLocation { get; set; }

        public bool IsThumbnailUpToDate { get; set; }

        public OpenedDocumentStates DocumentState { get; set; }

        public bool HasUserBeenWarnedAboutSaving { get; set; }

        public bool HasDocumentBeenModifiedSinceOpening { get; set; }

        public string MostRecentlyUsedToken { get; set; }

        public bool IsDocumentModifiedSinceLastSave { get; set; }

        public bool IsBrokenDocumentRestored { get; set; }

        public ulong FileSize { get; set; }

        public DateTimeOffset LastModifiedDate { get; set; } = DateTimeOffset.MinValue;

        public bool IsOpenedThroughDrop { get; set; }

        public void UpdateSettings(PDFViewCtrl ctrl)
        {
            HScrollPos = (int)ctrl.GetHScrollPos();
            VScrollPos = ctrl.GetVScrollPos();
            Zoom = ctrl.GetZoom();
            LastPage = ctrl.GetCurrentPage();
            PageRotation = ctrl.GetRotation();
            PagePresentationMode = ctrl.GetPagePresentationMode();
        }
    }
}
