﻿using Microsoft.UI.Xaml.Controls;
using pdftron.Common;
using pdftron.FDF;
using pdftron.Filters;
using pdftron.PDF;
using pdftron.PDF.Tools;
using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls;
using pdftron.PDF.Tools.Controls.Compare;
using pdftron.PDF.Tools.Controls.Multimedia;
using pdftron.PDF.Tools.Controls.Resources;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.ColorMode;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Crop;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Optimize;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Password;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Search;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Settings;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel.Core;
using Windows.ApplicationModel.DataTransfer;
using Windows.ApplicationModel.Resources;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.Graphics.Printing;
using Windows.Storage;
using Windows.Storage.AccessCache;
using Windows.Storage.Pickers;
using Windows.System;
using Windows.System.Display;
using Windows.UI;
using Windows.UI.Core;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;


namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer
{
    public partial class ViewerViewModel : ViewModelBase, INavigable, IDisposable
    {
        private const int MAX_TABS_PHONE = 6;
        private const int PAGES_NEEDED_FOR_THUMB_SLIDER = 5;

        private NewINavigableAvailableDelegate _newINavigableAvailable;
        public event NewINavigableAvailableDelegate NewINavigableAvailable
        {
            add
            {
                if (_newINavigableAvailable?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _newINavigableAvailable += value;
            }
            remove
            {
                if (_newINavigableAvailable != null)
                {
                    NewINavigableAvailableDelegate eventMethod = (NewINavigableAvailableDelegate)Array.Find(
                        _newINavigableAvailable.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _newINavigableAvailable -= eventMethod;
                }
            }
        }

        public event FindTextViewModel.FindTextResultFoundDelegate FindTextResultFound;

        private delegate void SaveAsDoneDelegate(bool cancelled, bool failed, Deferral deferral);
        private event SaveAsDoneDelegate SaveAsDone;
        public event EventHandler<Deferral> OpenSearch;

        private EventHandler<Deferral> _viewerActivate;
        public event EventHandler<Deferral> ViewerActivate
        {
            add
            {
                if (_viewerActivate?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _viewerActivate += value;
            }
            remove
            {
                if (_viewerActivate != null)
                {
                    EventHandler<Deferral> eventMethod = (EventHandler<Deferral>)Array.Find(
                        _viewerActivate.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _viewerActivate -= eventMethod;
                }
            }
        }

        private EventHandler<Deferral> _viewerActivateWithFile;
        public event EventHandler<Deferral> ViewerActivateWithFile
        {
            add
            {
                if (_viewerActivateWithFile?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _viewerActivateWithFile += value;
            }
            remove
            {
                if (_viewerActivateWithFile != null)
                {
                    EventHandler<Deferral> eventMethod = (EventHandler<Deferral>)Array.Find(
                        _viewerActivateWithFile.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _viewerActivateWithFile -= eventMethod;
                }
            }
        }

        private EventHandler<Deferral> _viewerDeactivate;
        public event EventHandler<Deferral> ViewerDeactivate
        {
            add
            {
                if (_viewerDeactivate?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _viewerDeactivate += value;
            }
            remove
            {
                if (_viewerDeactivate != null)
                {
                    EventHandler<Deferral> eventMethod = (EventHandler<Deferral>)Array.Find(
                        _viewerDeactivate.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _viewerDeactivate -= eventMethod;
                }
            }
        }

        private EventHandler<DragEventArgs> _viewerDragOver;
        public event EventHandler<DragEventArgs> ViewerDragOver
        {
            add
            {
                if (_viewerDragOver?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _viewerDragOver += value;
            }
            remove
            {
                if (_viewerDragOver != null)
                {
                    EventHandler<DragEventArgs> eventMethod = (EventHandler<DragEventArgs>)Array.Find(
                        _viewerDragOver.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _viewerDragOver -= eventMethod;
                }
            }
        }

        private EventHandler<DragEventArgs> _viewerDrop;
        public event EventHandler<DragEventArgs> ViewerDrop
        {
            add
            {
                if (_viewerDrop?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _viewerDrop += value;
            }
            remove
            {
                if (_viewerDrop != null)
                {
                    EventHandler<DragEventArgs> eventMethod = (EventHandler<DragEventArgs>)Array.Find(
                        _viewerDrop.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _viewerDrop -= eventMethod;
                }
            }
        }

        private BelongInRecentListDelegate _belongInRecentList;
        public delegate void BelongInRecentListDelegate(PdfViewCtrlTabInfo tab, bool isEncrypted, StorageFile tempFile);
        public event BelongInRecentListDelegate BelongInRecentList
        {
            add
            {
                if (_belongInRecentList?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _belongInRecentList += value;
            }
            remove
            {
                if (_belongInRecentList != null)
                {
                    BelongInRecentListDelegate eventMethod = (BelongInRecentListDelegate)Array.Find(
                        _belongInRecentList.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _belongInRecentList -= eventMethod;
                }
            }
        }

        private PageNumberChangedDelegate _pageNumberChanged;
        public delegate void PageNumberChangedDelegate(PdfViewCtrlTabInfo tab);
        public event PageNumberChangedDelegate PageNumberChanged
        {
            add
            {
                if (_pageNumberChanged?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _pageNumberChanged += value;
            }
            remove
            {
                if (_pageNumberChanged != null)
                {
                    PageNumberChangedDelegate eventMethod = (PageNumberChangedDelegate)Array.Find(
                        _pageNumberChanged.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _pageNumberChanged -= eventMethod;
                }
            }
        }

        public delegate void InternalControlStateDelegate(bool isOpen);

        InternalControlStateDelegate mPdfCompareState;

        /// <summary>
        /// Notify when the internal PDFCompare control has changed state Opened/Closed
        /// </summary>
        public event InternalControlStateDelegate PDFCompareState
        {
            add
            {
                if (mPdfCompareState?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    mPdfCompareState += value;
            }
            remove
            {
                if (mPdfCompareState != null)
                {
                    InternalControlStateDelegate eventMethod = (InternalControlStateDelegate)Array.Find(
                        mPdfCompareState.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        mPdfCompareState -= eventMethod;
                }
            }
        }

        private DocumentModifiedHandler _documentModified;
        public delegate void DocumentModifiedHandler();

        /// <summary>
        /// Raised when a document has been modified
        /// </summary>
        /// <remarks>
        /// Types of modification: Annotation added, removed, or edited 
        /// </remarks>
        public event DocumentModifiedHandler DocumentModified
        {
            add
            {
                if (_documentModified?.GetInvocationList().Contains(value) != true)
                    _documentModified += value;
            }
            remove => _documentModified -= value;
        }


        private readonly ResourceLoader _loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
        protected static readonly ContentDialog _contentDialog = new ContentDialog();

        private INavigable _navigateBack;
        private NavigablePage _navigateBackPage;

        protected NewDocumentProperties _propertiesToActivate;
        private bool _useTimedAppBar;
        private HotKeyHandler _hotKeyHandler;
        private DisplayRequest _displayRequest;
        private ThumbnailViewer _thumbnailViewer;
        private PdfCompareControl _PdfCompareViewer;
        private CreateDocControl _createDocControl;

        // Sound Annot
        private AudioControl _audioControl;

        private NewToolCreatedDelegate _toolManager_NewToolCreated;
        private SingleTapDelegate _toolManager_SingleTap;
        private PDFPrintManager _pdfPrintManager;

        // dealing with text to speech event
        private ToolManager.TextToSpeechDelegate _textToSpeechListener;

        // dealing with sound annot
        private ToolManager.SoundAnnotCreateHandler _soundAnnotCreateHandler;

        // dealing with modifying documents
        private ToolManager.AnnotationModificationHandler _annotAddedListener;
        private ToolManager.AnnotationModificationHandler _annotEditedListener;
        private ToolManager.AnnotationModificationHandler _annotRemovedListener;

        private ToolManager.AnnotationGroupModificationHandler _annotationGroupAddedListener;
        private ToolManager.AnnotationGroupModificationHandler _annotationGroupEditedListener;
        private ToolManager.AnnotationGroupModificationHandler _annotationGroupRemovedListener;

        private bool _hasPage1ChangeSinceThumbnailViewOpened;
        private bool _editedSinceThumbnailViewerOpened;
        private ThumbnailsViewViewModel.PageMovedDelegate _thumbnailViewerPageMovedDelegate;
        private ThumbnailsViewViewModel.PageDeletedDelegate _thumbnailViewerPageDeletedDelegate;
        private ThumbnailsViewViewModel.PageAddedDelegate _thumbnailViewerPageAddedDelegate;
        private ThumbnailsViewViewModel.PageRotatedDelegate _thumbnailViewerPageRotatedDelegate;

        protected bool _suppressRecentListUpdate;
        private bool _removingTab;

        private List<PdfViewCtrlTabInfo> _activeTabs = new List<PdfViewCtrlTabInfo>();

        private OutlineDialogViewModel _outlineDialogViewModel;
        private FindTextViewModel _findTextViewModel;
        private AnnotationCommandBar _annotationToolbar;
        private ThumbnailSlider _thumbnailSlider;
        private MediaElement _mediaElement;
        private bool _isPDFViewCtrlVisible;
        private bool _isModal;
        private bool _isSecretlyModal;
        private bool _isAppBarAvailable = true;
        private bool _isModalGrayout = true;
        private bool _isDismissableDialogOpen;
        private bool _isEntireAppBarOpen = ViewerSettings.PinCommandBar;
        private bool _isAppBarOpen;
        private bool _isQuickSettingsOpen;
        private bool _isDocumentEditable = true;
        private bool _isSaveButtonVisible;
        private bool _isUiPinned = ViewerSettings.PinCommandBar;
        private bool _isSaveAsButtonVisible;
        private bool _isSaveButtonEnabled = true;
        private bool _isSaveAsButtonEnabled = true;
        private bool _isSaveTimerOpen;
        private bool _isAnnotationToolbarOpen;
        private bool _outlineDialogNavigationCloses;
        private bool _outlineDialogHibernation;
        private bool _isOutlineDialogOpen;
        private bool _isFindTextDialogOpen;
        private bool _isAllowDropEnabled;
        private SaveAsOption _currentSaveAsOption;
        private string _saveAPasswordCopyText;
        private ClickablePageNumberIndicator _pageNumberIndicator;
        private bool _doToolsAllowZoom = true;
        private bool _isInputMouse;
        private bool _areAudioButtonsVisible;
        private bool _isAudioPlaying;
        // Used for the Phone, where we need to close the app when the user hits the back button in case they opened the app through file association.
        private bool _closeAppAfterSaving;
        private bool _savingInProgress;

        private SemaphoreSlim _resumeSemaphore;
        private SaveResult _autoSaveResultFromSuspend;
        private bool _showingAutoSaveResumeMessage;
        private AutoSaveHelper _autoSaveHelper;
        private bool _filePickerOpen;

        private OptimizeFileViewModel _optimizeFileViewModel;
        private PasswordFileViewModel _passwordFileViewModel;
        private CropPopupViewModel _cropPopupViewModel;
        private ViewerPageSettingsViewModel _viewerPageSettingsViewModel;
        private CustomColorViewModel _customColorViewModel;

        private NavigationState _navigationStateBeforeThumbnailsviewOpening;
        private bool _isThumbnailsViewOpen;
        private bool _isPdfCompareViewOpen;
        private CropView _cropView;
        private bool _isCropViewOpen;
        private Dictionary<int, Rect> _userCropRects = new Dictionary<int, Rect>();

        private EventHandler<RoutedEventArgs> _pdfUpdatedHandler;
        private EventHandler<RoutedEventArgs> _pdfIsReadyHandler;
        private PdfViewCtrlTabInfo _previousTab;

        private bool _tabIsSubscribed;
        private Dictionary<PdfViewCtrlTabInfo, EventHandler<TappedRoutedEventArgs>> _tabRightTapHandlers = new Dictionary<PdfViewCtrlTabInfo, EventHandler<TappedRoutedEventArgs>>();
        private TabConflictChoiceCanceller _currentConflictCanceller;
        private PdfViewCtrlTabInfo _tabThatIsMovingFile;
        private PdfViewCtrlTabInfo _tabToFallBackTo;

        private OnConversionEventHandler _onConversionChangedHandler;

        private DispatcherTimer _appBarClosingTimer;

        private StorageFile _fileToActivate;
        private NewDocumentProperties _fileToActivateDocumentProperties;
        private bool _launchedThroughFileActivation;
        private PdfViewCtrlTabInfo _tabToLoadAfterResolvingConflict;

        private DocumentOpener _documentOpener;
        private Tuple<PdfViewCtrlTabInfo, ReflowViewModel> _currentReflow;
        private bool _showPageNumberIndicator = true;
        private bool _showBackButton;
        private bool _isThumbnailViewVisible = true;
        private bool _isCropViewVisible = true;
        private bool _isActionButtonVisible = true;
        private bool _isActionsEnabled = true;
        private bool _isShareVisible = true;
        private bool _isPrintVisible = true;
        private bool _isExportXFDFVisible = true;
        private bool _isMultiWindowVisible = true;
        private bool _isMultiWindowEnabled = true;
        private bool _isMultiWindowActive = true;
        private bool _isPdfCompareVisible = true;
        private bool _isPdfCompareEnabled = true;
        private bool _isPdfCompareActive = true;
        private bool _isPivotItemThumbnailsEnabled = true;
        private bool _isPivotItemBookmarksEnabled = true;
        private bool _isPivotItemOutlineEnabled = true;
        private bool _isPivotItemAnnotationsEnabled = true;
        private bool _isCloseTabButtonEnabled = true;
        private bool _isEditButtonVisible = true;
        private bool _isEditButtonEnabled = true;
        private bool _isOutlineButtonVisible = true;
        private bool _isOutlineButtonEnabled = true;
        private bool _isTextSearchButtonVisible = true;
        private bool _isTextSearchButtonEnabled = true;
        private bool _isViewModeButtonVisible = true;
        private bool _isViewModeButtonEnabled = true;
        private bool _isSeeMoreButtonVisible = true;
        private bool _isSeeMoreButtonEnabled = true;
        private bool _isReadOnlyModeEnabled = false;
        private bool _isOutlineEditingEnabled = true;
        private bool _isAudioControlVisible = false;

#if SUBSCRIPTIONS
        private bool _isSubscribed;
        private bool _isPdfCompareSubVisible = false;
        private bool _isMultiWindowSubVisible = false;
        private bool _isDiamondSubVisible = false;
#endif

        private AppBarTabButton _saveCopyButton;

        #region Commands

        public RelayCommand GoToFilesCommand { get; private set; }

        public RelayCommand PinCommand { get; private set; }

        public RelayCommand UnpinCommand { get; private set; }

        public RelayCommand SaveCommand { get; private set; }

        public RelayCommand SaveAsCommand { get; private set; }

        public RelayCommand SearchCommand { get; private set; }

        public RelayCommand OutlineCommand { get; private set; }

        public RelayCommand EditCommand { get; private set; }

        public RelayCommand PdfCompareCommand { get; private set; }

        public RelayCommand MultiWindowCommand { get; private set; }

        public RelayCommand SaveAndFlattenCommand { get; private set; }

        public RelayCommand SaveReducedFileSizeCopyCommand { get; private set; }

        public RelayCommand SavePasswordCopyCommand { get; private set; }

        public RelayCommand SavePermCroppedCopyCommand { get; private set; }

        public RelayCommand PageEditingCommand { get; private set; }

        public RelayCommand PrintCommand { get; private set; }

        public RelayCommand ShareCommand { get; private set; }

        public RelayCommand ExportXFDFCommand { get; private set; }

        public RelayCommand ZoomCommand { get; private set; }

        public RelayCommand PlayAudioCommand { get; private set; }

        public RelayCommand PauseAudioCommand { get; private set; }

        public RelayCommand StopAudioCommand { get; private set; }

        public RelayCommand UndoCommand { get; private set; }

        public RelayCommand RedoCommand { get; private set; }

        public RelayCommand CloseSearchCommand { get; private set; }

        public RelayCommand FullScreenExitCommand { get; private set; }

        public RelayCommand DismissDialogCommand { get; private set; }

        public RelayCommand QuickSettingsOpenedCommand { get; private set; }

        public RelayCommand QuickSettingsClosedCommand { get; private set; }

        #endregion

        public ViewerViewModel(NewDocumentProperties properties = null, AnnotationToolbarViewModel viewModel = null)
        {
#if SUBSCRIPTIONS
            InitSubscription();
#endif

            InitToolControls(viewModel);
            InitCommands();
            InitAppBarManagement();

            InitializeTabControlViewModel();

            _propertiesToActivate = properties;
        }

        public void Dispose()
        {
            if (OutlineDialogViewModel != null)
            {
                OutlineDialogViewModel.RequestClosing -= OutlineDialogViewModel_RequestClosing;
                OutlineDialogViewModel.RequestSwitchSide -= OutlineDialogViewModel_RequestSwitchSide;
                OutlineDialogViewModel.DocumentModified -= OutlineDialogViewModel_DocumentModified;
                OutlineDialogViewModel.UserBookmarksEdited -= OutlineDialogViewModel_UserBookmarksEdited;

                OutlineDialogViewModel = null;
            }

            if (_autoSaveHelper != null)
            {
                _autoSaveHelper.DocumentWasSaved -= AutoSaveHelper_DocumentWasSaved;
                _autoSaveHelper.CommitChangesAsync -= AutoSaveHelper_CommitChangesAsync;

                _autoSaveHelper = null;
            }

            if (PdfCompareViewer != null)
            {
                PdfCompareViewer.ControlClosed -= PdfCompareViewer_ControlClosed;

                PdfCompareViewer = null;
            }

            if (_thumbnailViewer != null && _thumbnailViewer.ViewModel != null)
            {
                _thumbnailViewer.ViewModel.PageMoved -= _thumbnailViewerPageMovedDelegate;
                _thumbnailViewer.ViewModel.PageDeleted -= _thumbnailViewerPageDeletedDelegate;
                _thumbnailViewer.ViewModel.PageAdded -= _thumbnailViewerPageAddedDelegate;
                _thumbnailViewer.ViewModel.PageRotated -= _thumbnailViewerPageRotatedDelegate;
                _thumbnailViewer.ControlClosed -= ThumbnailViewer_ControlClosed;

                _thumbnailViewer = null;
            }

            if (CropView != null)
            {
                CropView.ControlClosed -= CropView_ControlClosed;

                if (CropView.ViewModel != null)
                {
                    CropView.ViewModel.DocumentEdited -= CropViewModel_DocumentEdited;
                }

                CropView = null;
            }

            if (ThumbnailSlider != null)
            {
                ThumbnailSlider.ManipulationStarted -= ThumbnailSlider_ManipulationStarted;
                ThumbnailSlider.ManipulationCompleted -= ThumbnailSlider_ManipulationCompleted;
                ThumbnailSlider.PointerReleased -= ThumbnailSlider_PointerReleased;
                ThumbnailSlider.PointerReleased -= ThumbnailSlider_PointerReleased;

                ThumbnailSlider = null;
            }

            if (_appBarClosingTimer != null)
            {
                _appBarClosingTimer.Tick -= AppBarClosingTimer_Tick;

                _appBarClosingTimer = null;
            }

            if (AnnotationToolbar != null)
            {
                AnnotationToolbar.ControlClosed -= AnnotationToolbar_ControlClosed;
                AnnotationToolbar.ToolManager = null;
            }

            if (_createDocControl != null)
            {
                _createDocControl.NewDocumentCreated -= _createDocControl_NewDocumentCreated;
                _createDocControl.DocumentCreationFailed -= _createDocControl_DocumentCreationFailed;

                _createDocControl.Dispose();
                _createDocControl = null;
            }
        }

        public bool IsPasswordProtected
        {
            get { return !string.IsNullOrEmpty(_propertiesToActivate.Password); }
        }

        public async void Activate(object parameter)
        {
            if (_viewerActivate != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    _viewerActivate.Invoke(this, deferral);
                    await tcs.Task;
                }
            }

            if (!DocumentManager.IsReady)
            {
                IsModal = true;
                await DocumentManager.Instance.InitializeInstanceAsync();
                IsModal = false;
            }

            if (TabControlViewModel.IsReady)
            {
                IsModal = true;
                await TabControlViewModel.Instance.UpdateFilePropertiesAsync();
                IsModal = false;
            }

            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone
                && ViewerSettings.PhoneFullScreen == 1)
            {
                await UtilityFunctions.SetFullScreenModeAsync(true);
            }

            if (_propertiesToActivate != null)
                await InitPdfViewAsync(_propertiesToActivate);

            _useTimedAppBar = true;

            _hotKeyHandler = HotKeyHandler.Current;
            _hotKeyHandler.KeyPressedEvent += HotKeyHandler_KeyPressedEvent;
            _hotKeyHandler.HotKeyPressedEvent += HotKeyHandler_HotKeyPressedEvent;
            _hotKeyHandler.AltHotKeyPressedEvent += HotKeyHandler_AltHotKeyPressedEvent;

            SystemNavigationManager.GetForCurrentView().BackRequested += BackButtonHandler_BackPressed;

            SharingHelper = SharingHelper ?? SharingHelper.GetSharingHelper();
            SharingHelper.RetrieveSharingString += SharingHelper_RetrieveSharingString;
            SharingHelper.RetrieveSharingStorageFile += SharingHelper_RetrieveSharingStorageFile;
            SharingHelper.DocumentSaver = SaveHelperAsync;

            if (ViewerSettings.ScreenSleepLock)
            {
                _displayRequest = new DisplayRequest();
                _displayRequest.RequestActive();
            }

            TabControlViewModel.TabWidthMode = ViewerSettings.TabScrolling ? TabViewWidthMode.SizeToContent : TabViewWidthMode.Equal;
        }

        /// <summary>
        /// Ctrl + Shift + key
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private async void HotKeyHandler_CtrlShiftHotKeyPressedEvent(CoreDispatcher sender, AcceleratorKeyEventArgs args)
        {
            switch (args.VirtualKey)
            {
                case VirtualKey.Tab: //Go to previous tab
                    await HotKeyNextTabHandlerAsync(true);
                    break;
                case (VirtualKey)187:
                    PDFViewCtrl.RotateClockwise();
                    break;
                case (VirtualKey)189:
                    PDFViewCtrl.RotateCounterClockwise();
                    break;
                case VirtualKey.Z: //Redo annotation
                    if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
                    {
                        var annotToolBar = AnnotationToolbar.DataContext as AnnotationToolbarViewModel;
                        annotToolBar.PerformInkRedo();
                        return;
                    }
                    else if (ToolManager.UndoRedoAction.CanRedo)
                    {
                        CloseCurrentTool();
                        ToolManager.UndoRedoAction.DoRedo();
                    }
                    break;
            }

            args.Handled = true;
        }

        /// <summary>
        /// ctrl + alt + something was pressed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private async void HotKeyHandler_AltHotKeyPressedEvent(Windows.UI.Core.CoreWindow sender, Windows.UI.Core.KeyEventArgs args)
        {
            if (IsModal)
            {
                return;
            }
            switch (args.VirtualKey)
            {
                case Windows.System.VirtualKey.S:
                    if (IsSaveAsButtonVisible)
                    {
                        await SaveAsAsync(TabControlViewModel.SelectedTab, false, false, SaveAsOption.Save);
                    }
                    break;
            }
        }

        public async void Deactivate(object parameter)
        {
            if (_viewerDeactivate != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    _viewerDeactivate.Invoke(this, deferral);
                    await tcs.Task;
                }
            }

            ThumbnailViewer = null;
            _hotKeyHandler.KeyPressedEvent -= HotKeyHandler_KeyPressedEvent;
            _hotKeyHandler.HotKeyPressedEvent -= HotKeyHandler_HotKeyPressedEvent;
            _hotKeyHandler.AltHotKeyPressedEvent -= HotKeyHandler_AltHotKeyPressedEvent;

            PDFPrintManager.UnRegisterForPrintingContract();

            SystemNavigationManager.GetForCurrentView().BackRequested -= BackButtonHandler_BackPressed;

            SharingHelper.RetrieveSharingString -= SharingHelper_RetrieveSharingString;
            SharingHelper.RetrieveSharingStorageFile -= SharingHelper_RetrieveSharingStorageFile;
            SharingHelper.DocumentSaver = null;

            TabControlViewModel.VisibleTabChanged -= TabControlViewModel_VisibleTabChanged;
            TabControlViewModel.BrowseDocumentRequested -= TabControlViewModel_BrowseDocumentRequested;
			TabControlViewModel.CreateDocumentRequested -= TabControlViewModel_CreateDocumentRequested;
            TabControlViewModel.CloseButtonClicked -= TabControlViewModel_CloseButtonClicked;

            foreach (PdfViewCtrlTabInfo tab in TabControlViewModel.Tabs)
            {
                tab?.NavigationStack?.Unsubscribe();
            }

            if (TabControlViewModel.SelectedTab != null)
                await DeactivateTabImplAsync(TabControlViewModel.SelectedTab);

            if (ViewerSettings.ScreenSleepLock)
                _displayRequest.RequestRelease();

            try
            {
                if (IsFullScreen)
                    await UtilityFunctions.SetFullScreenModeAsync(false);
            }
            catch
            {
                //Swallow exception
            }

            if (FindTextViewModel != null)
                await FindTextViewModel.WaitForTextSearchToCancel();

            if (OutlineDialogViewModel != null)
            {
                await OutlineDialogViewModel.CleanUpSubViewsAsync();
                OutlineDialogViewModel = null;
            }

            TabControlViewModel.Deactivate();
            ViewerPageSettingsViewModel = null;

            foreach (PdfViewCtrlTabInfo tab in _tabRightTapHandlers.Keys)
            {
                tab.ReflowTapped -= _tabRightTapHandlers[tab];
            }

            _tabRightTapHandlers.Clear();

            if (ToolManager != null)
            {
                if (ToolManager.UndoRedoAction != null)
                {
                    ToolManager.UndoRedoAction.OnUndoRedoStatusChanged -= UndoRedoAction_OnUndoRedoStatusChanged;
                }
                ToolManager.Dispose();
            } 
            if (DocumentOpener != null)
            {
                DocumentOpener.DocumentReady -= DocumentOpener_DocumentReady;
                DocumentOpener.ModalityChanged -= DocumentOpener_ModalityChanged;
                DocumentOpener.DocumentLoadingFailed -= DocumentOpener_DocumentLoadingFailed;
                DocumentOpener = null;
            }
        }

        private async void UndoRedoAction_OnUndoRedoStatusChanged(object sender, UndoRedoStatusEventArgs e)
        {
            PageNumberIndicator?.UpdatePageNumbers();

            // Notify document has changed and allow/enable Save button
            await HandleDocumentEditingAsync();

            // Make sure to update modified page thumbnail
            if (IsOutlineDialogOpen)
                OutlineDialogViewModel.Thumbnails.PageModified(e.ModifiedPage);
        }

        private async Task InitPdfViewAsync(NewDocumentProperties properties)
        {
            if (!TabControlViewModel.IsReady)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    "TabControlViewModel was not ready");
                return;
            }

            bool isFileOpenAndActive = await TabControlViewModel.CheckIfFileIsOpenAndActivateAsync(properties);
            bool isOriginalFileOpenAndActive = await TabControlViewModel.CheckIfOriginalIsOpenAndActivateAsync(properties.OriginalFileFromSystem);

            if (isFileOpenAndActive || (properties.OriginalFileFromSystem != null && isOriginalFileOpenAndActive))
            {
                PdfViewCtrlTabInfo tabInfo = TabControlViewModel.SelectedTab;
                if (tabInfo.MetaData.IsOpenedThroughDrop
                    && (properties?.OpenedThroughDrop != true)
                    && properties.OpenedDocumentState == OpenedDocumentStates.Normal
                    && properties.File != null)
                {
                    tabInfo.OriginalFile = properties.File;
                    tabInfo.DocumentState = OpenedDocumentStates.Normal;
                    tabInfo.MetaData.IsOpenedThroughDrop = false;
                    UpdateTabDependantProperties();
                }
            }
            else
            {
                if (TabControlViewModel.OpenTabs >= TabControlViewModel.MaximumItems)
                {
                    _fileToActivate = properties.File;
                    _fileToActivateDocumentProperties = properties;
                    _suppressRecentListUpdate = true;

                    await CloseOldestTabAsync();

                    return;
                }

                if (!ViewerSettings.RememberLastPage)
                    properties.StartPage = 1;

                await CreateNewTabAsync(properties);
            }
            ResolveColorMode(ViewerSettings.ColorMode);
        }

        private async Task CloseOldestTabAsync()
        {
            PdfViewCtrlTabInfo tabToclose = TabControlViewModel.OldestViewedTab;
            if (tabToclose == null)
                return;

            SaveNeededStates status = CheckSavingRequirements(tabToclose);
            bool conflict = !tabToclose.IsUpToDate && tabToclose.IsDocumentModifiedSinceLastSave;
            if (status == SaveNeededStates.saveAsNeeded)
                await TabControlViewModel.SelectTabAsync(tabToclose);

            if (!conflict)
                await CloseTabRequestedAsync(status, tabToclose, true);
        }

        private async Task CreateNewTabAsync(NewDocumentProperties properties)
        {
            PdfViewCtrlTabInfo tab = new PdfViewCtrlTabInfo(properties.File, properties.TemporaryFile, properties.Doc, properties.File.DisplayName)
            {
                FileSourceIfNotSaveable = properties.OriginalFileFromSystem,
                StreamForConversion = properties.FileStream
            };

            await TabControlViewModel.AddTabAsync(tab);
            tab.DocumentState = properties.OpenedDocumentState;
            tab.MetaData.IsOpenedThroughDrop = properties.OpenedThroughDrop;
            tab.HasDocumentBeenModifiedSinceOpening = false;
            tab.HasUserBeenWarnedAboutSaving = false;
            tab.IsDocumentModifiedSinceLastSave = false;
            if (!string.IsNullOrEmpty(properties.Password))
                tab.MetaData.Password = properties.Password;

            await TabControlViewModel.SelectTabAsync(tab);
            if (ViewerSettings.RememberLastPage)
            {
                tab.MetaData.PageRotation = properties.PageRotation;
                switch (tab.MetaData.PageRotation)
                {
                    case PageRotate.e_90:
                        tab.PDFViewCtrl.RotateClockwise();
                        break;
                    case PageRotate.e_180:
                        tab.PDFViewCtrl.RotateClockwise();
                        tab.PDFViewCtrl.RotateClockwise();
                        break;
                    case PageRotate.e_270:
                        tab.PDFViewCtrl.RotateCounterClockwise();
                        break;
                }

                tab.MetaData.PagePresentationMode = properties.PresentationMode;
                tab.PDFViewCtrl.SetPagePresentationMode(tab.MetaData.PagePresentationMode);
                tab.IsReflow = properties.IsReflow;

                tab.MetaData.LastPage = properties.StartPage;
                tab.MetaData.Zoom = properties.Zoom;
                tab.MetaData.HScrollPos = properties.HorizontalScrollPosition;
                tab.MetaData.VScrollPos = properties.VerticalScrollPosition;
            }

            tab.MetaData.LastPage = properties.StartPage;
            tab.MetaData.LastModifiedDate = properties.LastModifiedDate;
            tab.MetaData.FileSize = properties.Filesize;
            tab.SystemLastModifiedDate = properties.LastModifiedDate;
            tab.SystemFileSize = properties.Filesize;
            UpdateTabDependantProperties();
            ResolveReflow();

            if (ToolManager != null)
            {
                if (ToolManager.UndoRedoAction != null)
                    ToolManager.UndoRedoAction.OnUndoRedoStatusChanged += UndoRedoAction_OnUndoRedoStatusChanged;
            }
        }

#if SUBSCRIPTIONS
        /// <summary>
        /// Initialize subscription resources
        /// </summary>
        private void InitSubscription()
        {
            // TODO: remove IsSubscribed since we can move sub state to ViewerSettings
            ViewerSettings.IsSubscriptionActivated = GetSubscribed();

            if (ViewerSettings.IsSubscriptionActivated)
            {
                if (IsPdfCompareEnabled)
                    IsPdfCompareVisible = true;

                if (IsMultiWindowEnabled)
                    IsMultiWindowVisible = true;
            }
            else
            {
                IsPdfCompareSubVisible = true;
                IsMultiWindowSubVisible = true;
                IsDiamondSubVisible = true;
            }
        }
#endif

        private void InitToolControls(AnnotationToolbarViewModel viewModel)
        {
            _pageNumberIndicator = new ClickablePageNumberIndicator
            {
                Background = new SolidColorBrush(Colors.Transparent)
            };

            PageNumberIndicator.ShowingBehaviour = ClickablePageNumberIndicator.ShowingBehaviours.AlwaysShow;
            PageNumberIndicator.GotFocus += PageNumberIndicator_GotFocus;
            PageNumberIndicator.Show();

            AnnotationToolbar = new AnnotationCommandBar(viewModel);
            AnnotationToolbarViewModel.SetCloseableControl(AnnotationToolbar);

            AnnotationToolbar.ControlClosed += AnnotationToolbar_ControlClosed;
            AnnotationToolbar.ButtonsStayDown = ViewerSettings.ButtonsStayDown;
            AnnotationToolbar.InkMultiStroke = ViewerSettings.InkMultiStrokeOption;
        }

        private void InitializeTabControlViewModel()
        {
            TabControlViewModel.DefaultImageURL = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png";
            TabControlViewModel.ShowImagePreviews = true;
            TabControlViewModel.AutoSaveTabState = true;
            TabControlViewModel.VisibleTabChanged += TabControlViewModel_VisibleTabChanged;
            TabControlViewModel.BrowseDocumentRequested += TabControlViewModel_BrowseDocumentRequested;
            TabControlViewModel.CreateDocumentRequested += TabControlViewModel_CreateDocumentRequested;
            TabControlViewModel.CloseButtonClicked += TabControlViewModel_CloseButtonClicked;
            TabControlViewModel.ShowIfDocumentModified = ShowIfModifiedStates.DoNotShow;
        }

        private void OutlineDialogViewModel_RequestClosing(object sender, EventArgs e) => IsOutlineDialogOpen = false;

        private void OutlineDialogViewModel_RequestSwitchSide(object sender, EventArgs e)
        {
            if (ViewerSettings.OutlineDialogAnchorSide == AnchorSides.Left)
            {
                ViewerSettings.OutlineDialogAnchorSide = AnchorSides.Right;
            }
            else if (ViewerSettings.OutlineDialogAnchorSide == AnchorSides.Right)
            {
                ViewerSettings.OutlineDialogAnchorSide = AnchorSides.Left;
            }
        }

        private async void OutlineDialogViewModel_DocumentModified(PDFDoc doc) => await HandleDocumentEditingAsync();

        private async void OutlineDialogViewModel_UserBookmarksEdited(PDFDoc doc) => await HandleDocumentEditingAsync();

        private void NavigationStack_NavigationStackChanged()
        {
            RaisePropertyChanged(nameof(IsUndoEnabled));
            RaisePropertyChanged(nameof(IsRedoEnabled));
        }

        private void PDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages) => _pageNumberChanged?.Invoke(TabControlViewModel.SelectedTab);

        private void PDFViewCtrl_IsEnabledChanged(object sender, DependencyPropertyChangedEventArgs e) => IsModal = !(bool)e.NewValue;

        private void PageNumberIndicator_GotFocus(object sender, RoutedEventArgs e)
        {
            _appBarClosingTimer?.Stop();
        }

        private void AnnotationToolbar_ControlClosed()
        {
            IsAnnotationToolbarOpen = false;
            AppbarInteraction();
            CloseCurrentTool();
            UpdateUIVisiblity();
        }

        private PDFPrintManager PDFPrintManager
        {
            get
            {
                if (_pdfPrintManager == null)
                {
                    // printing
                    _pdfPrintManager = PDFPrintManager.GetInstance();
                    _pdfPrintManager.UnRegisterForPrintingContract();

                    ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Printing");
                    _pdfPrintManager.SetResourceLoader(loader);

                    // standard options
                    _pdfPrintManager.AddStandardPrintOption(StandardPrintTaskOptions.MediaSize);
                    _pdfPrintManager.AddStandardPrintOption(StandardPrintTaskOptions.Orientation);
                    _pdfPrintManager.AddStandardPrintOption(StandardPrintTaskOptions.Copies);

                    // PDFTron options
                    _pdfPrintManager.AddUserOptionAnnotations();
                    _pdfPrintManager.AddUserOptionAutoRotate();
                    _pdfPrintManager.AddUserOptionPageRange();
                }

                return _pdfPrintManager;
            }
        }

        public bool SaveDocumentAfterClosing { get; set; }

        public bool HasDocumentBeenModifiedSinceOpening
        {
            get => TabControlViewModel.SelectedTab.HasDocumentBeenModifiedSinceOpening;
            set => TabControlViewModel.SelectedTab.HasDocumentBeenModifiedSinceOpening = value;
        }

        public ToolManager ToolManager
        {
            get
            {
                if (TabControlViewModel == null || TabControlViewModel.SelectedTab == null)
                    return null;

                return TabControlViewModel.SelectedTab.ToolManager;
            }
        }

        public OutlineDialogViewModel OutlineDialogViewModel
        {
            get => _outlineDialogViewModel;
            set
            {
                OutlineDialogViewModel oldVM = _outlineDialogViewModel;
                if (Set(ref _outlineDialogViewModel, value))
                {
                    if (oldVM != null)
                    {
                        oldVM.RequestClosing -= OutlineDialogViewModel_RequestClosing;
                        oldVM.RequestSwitchSide -= OutlineDialogViewModel_RequestSwitchSide;
                        oldVM.DocumentModified -= OutlineDialogViewModel_DocumentModified;
                        oldVM.UserBookmarksEdited -= OutlineDialogViewModel_UserBookmarksEdited;
                    }

                    if (OutlineDialogViewModel != null)
                    {
                        OutlineDialogViewModel.CloseDialogOnItemClick = _outlineDialogNavigationCloses;
                        OutlineDialogViewModel.RequestClosing += OutlineDialogViewModel_RequestClosing;
                        OutlineDialogViewModel.RequestSwitchSide += OutlineDialogViewModel_RequestSwitchSide;
                        OutlineDialogViewModel.DocumentModified += OutlineDialogViewModel_DocumentModified;
                        OutlineDialogViewModel.UserBookmarksEdited += OutlineDialogViewModel_UserBookmarksEdited;
                    }
                }
            }
        }

        public FindTextViewModel FindTextViewModel
        {
            get => _findTextViewModel;
            set => Set(ref _findTextViewModel, value);
        }

        public TabControlViewModel TabControlViewModel => TabControlViewModel.Instance;

        public PDFViewCtrl PDFViewCtrl
        {
            get
            {
                if (TabControlViewModel == null || TabControlViewModel.SelectedTab == null)
                    return null;

                return TabControlViewModel.SelectedTab.PDFViewCtrl;
            }
        }

        public ReflowView ReflowView
        {
            get
            {
                if (TabControlViewModel == null || TabControlViewModel.SelectedTab == null)
                    return null;

                return TabControlViewModel.SelectedTab.ReflowView;
            }
        }

        public AnnotationCommandBar AnnotationToolbar
        {
            get => _annotationToolbar;
            set => Set(ref _annotationToolbar, value);
        }

        public AnnotationToolbarViewModel AnnotationToolbarViewModel
        {
            get => AnnotationToolbar.ViewModel;
            set => AnnotationToolbar.ViewModel = value;
        }

        public ThumbnailSlider ThumbnailSlider
        {
            get => _thumbnailSlider;
            set => Set(ref _thumbnailSlider, value);
        }

        public MediaElement MediaElement
        {
            get => _mediaElement;
            set => Set(ref _mediaElement, value);
        }

        public bool IsPDFViewCtrlVisible
        {
            get => _isPDFViewCtrlVisible;
            set => Set(ref _isPDFViewCtrlVisible, value);
        }

        public NewDocumentProperties PropertiesToActivate => _propertiesToActivate;

        public INavigable NavigateBack
        {
            get => _navigateBack;
            set => Set(ref _navigateBack, value);
        }

        public NavigablePage NavigateBackPage
        {
            get => _navigateBackPage;
            set => Set(ref _navigateBackPage, value);
        }

        public Type NavigablePage { get; set; }

        // gets or sets whether the current dialog is modal or not.
        public bool IsModal
        {
            get => _isModal;
            set
            {
                if (_isModal != value)
                {
                    _isModal = value;
                    if (AnnotationToolbar != null)
                        AnnotationToolbar.IsEnabled = !_isModal;

                    if (TabControlViewModel != null)
                        TabControlViewModel.IsEnabled = !value;

                    if (_isModal)
                    {
                        if (IsAppBarOpen)
                            IsAppBarOpen = false;

                        if (!ViewerSettings.PinCommandBar && !IsAnnotationToolbarOpen)
                            IsEntireAppBarOpen = false;
                    }

                    IsAppBarAvailable = !_isModal;
                    RaisePropertyChanged(nameof(IsModal));
                }
            }
        }

        /// <summary>
        /// This is added make the viewer modal without appearing modal. This can be used for async operations that
        /// are generally fast, but sometimes slow and require modality. Also remember to make the PDFViewCtrl not
        /// hit test visible.
        /// </summary>
        public bool IsSecretlyModal
        {
            get => _isSecretlyModal;
            set => Set(ref _isSecretlyModal, value);
        }

        public bool IsAppBarAvailable
        {
            get => _isAppBarAvailable;
            set => Set(ref _isAppBarAvailable, value);
        }

        /// <summary>
        /// When true, the modal dialog grays out the Viewer
        /// </summary>
        public bool IsModalGrayout
        {
            get => _isModalGrayout;
            set => Set(ref _isModalGrayout, value);
        }

        // Gets or sets whether or not the current dialog is dismissable
        public bool IsDismissableDialogOpen
        {
            get => _isDismissableDialogOpen;
            set
            {
                if (Set(ref _isDismissableDialogOpen, value) && !_isDismissableDialogOpen)
                    CloseDismissableDialogs();
            }
        }

        public bool IsFullScreen => UtilityFunctions.IsFullScreen();

        public bool IsReflow
        {
            get
            {
                if (TabControlViewModel?.SelectedTab == null)
                    return false;

                return TabControlViewModel.SelectedTab.IsReflow;
            }
        }

        public bool ShowPageNumberIndicator
        {
            get => PDFViewCtrl != null && _showPageNumberIndicator;
            set => Set(ref _showPageNumberIndicator, value);
        }

        public bool ShowBackButton
        {
            get => _showBackButton;
            set => Set(ref _showBackButton, value);
        }

        /// <summary>
        /// Change the visibility of the Right to Left reading direction option
        /// </summary>
        public bool IsRightToLeftVisibile
        {
            get => ViewerSettings.IsRightToLeftVisible;
            set
            {
                ViewerSettings.IsRightToLeftVisible = value;

                if (ViewerPageSettingsViewModel != null)
                {
                    ViewerPageSettingsViewModel.IsRightToLeftEnabled = value;
                }

            }
        }

        /// <summary>
        /// Choose whether or not the Edit button is Visible or Collapsed
        /// </summary>
        public bool IsEditButtonVisible
        {
            get
            {
                if (!IsEditButtonEnabled)
                    return false;

                return _isEditButtonVisible;
            }
            set => Set(ref _isEditButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the Outline button is Visible or Collapsed
        /// </summary>
        public bool IsOutlineButtonVisible
        {
            get
            {
                if (!IsOutlineButtonEnabled)
                    return false;

                if (!IsPivotItemAnnotationsEnabled && !IsPivotItemBookmarksEnabled && !IsPivotItemOutlineEnabled && !IsPivotItemThumbnailsEnabled)
                    return false;

                return _isOutlineButtonVisible;
            }
            set => Set(ref _isOutlineButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the TextSearch button is Visible or Collapsed
        /// </summary>
        public bool IsTextSearchButtonVisible
        {
            get
            {
                if (!IsDocumentSearchable)
                    return false;

                if (!IsTextSearchButtonEnabled)
                    return false;

                return _isTextSearchButtonVisible;
            }
            set => Set(ref _isTextSearchButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the ViewMode button is Visible or Collapsed
        /// </summary>
        public bool IsViewModeButtonVisible
        {
            get
            {
                if (!IsQuickSettingsEnabled)
                    return false;

                if (!IsViewModeButtonEnabled)
                    return false;

                return _isViewModeButtonVisible;
            }
            set => Set(ref _isViewModeButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the SeeMore button is Visible or Collapsed
        /// </summary>
        public bool IsSeeMoreButtonVisible
        {
            get
            {
                if (!IsSeeMoreButtonEnabled)
                    return false;

                return _isSeeMoreButtonVisible;
            }
            set => Set(ref _isSeeMoreButtonVisible, value);
        }

        /// <summary>
        /// Choose whether Read Only Mode is enabled
        /// </summary>
        public bool IsReadOnlyModeEnabled
        {
            get
            {
                return _isReadOnlyModeEnabled;
            }
            set
            {
                Set(ref _isReadOnlyModeEnabled, value);

                if (_isReadOnlyModeEnabled)
                {
                    IsEditButtonEnabled = false;
                }
            }
        }

        // 
        /// <summary>
        /// Choose whether or not editing outlines is enabled
        /// </summary>
        public bool IsOutlineEditingEnabled
        {
            get { return _isOutlineEditingEnabled; }
            set { _isOutlineEditingEnabled = value; }
        }

        /// <summary>
        /// Change the visibility of the Outline flyout menu Thumbnails button 
        /// </summary>
        public bool IsPivotItemThumbnailsEnabled
        {
            get => _isPivotItemThumbnailsEnabled;
            set => Set(ref _isPivotItemThumbnailsEnabled, value);
        }

        /// <summary>
        /// Change the visibility of the Outline flyout menu Annotations button 
        /// </summary>
        public bool IsPivotItemBookmarksEnabled
        {
            get => _isPivotItemBookmarksEnabled;
            set => Set(ref _isPivotItemBookmarksEnabled, value);
        }

        /// <summary>
        /// Change the visibility of the Outline flyout menu Bookmarks button 
        /// </summary>
        public bool IsPivotItemOutlineEnabled
        {
            get => _isPivotItemOutlineEnabled;
            set => Set(ref _isPivotItemOutlineEnabled, value);
        }

        /// <summary>
        /// Change the visibility of the Outline flyout menu Outline button 
        /// </summary>
        public bool IsPivotItemAnnotationsEnabled
        {
            get => _isPivotItemAnnotationsEnabled;
            set => Set(ref _isPivotItemAnnotationsEnabled, value);
        }

        /// <summary>
        /// Change the visibility of the buttons to open the ThumbnailView
        /// </summary>
        public bool IsThumbnailViewVisible
        {
            get => _isThumbnailViewVisible;
            set
            {
                if (_isThumbnailViewVisible != value)
                {
                    _isThumbnailViewVisible = value;
                    RaisePropertyChanged();

                    if (ViewerPageSettingsViewModel != null)
                    {
                        ViewerPageSettingsViewModel.IsThumbnailViewVisible = IsThumbnailViewVisible;
                    }
                }
            }
        }

        /// <summary>
        /// Change the visibility of the buttons to open the Crop View
        /// </summary>
        public bool IsCropViewVisible
        {
            get => _isCropViewVisible;
            set
            {
                if (_isCropViewVisible != value)
                {
                    _isCropViewVisible = value;
                    RaisePropertyChanged();

                    if (ViewerPageSettingsViewModel != null)
                    {
                        ViewerPageSettingsViewModel.IsCropViewVisible = IsCropViewVisible;
                    }
                }
            }
        }

        /// <summary>
        /// Choose whether the ViewerControl will keep the tabs opened
        /// </summary>
        /// <remarks>Used when there is a customized navigation pages which will
        /// handle file opening</remarks>
        public bool KeepOpenedTabs
        {
            get => TabControlViewModel.KeepOpenedTabs;
            set => TabControlViewModel.KeepOpenedTabs = value;
        }

        /// <summary>
        /// Ensure to deactivate all opened tabs when navigating away from the page that hosts
        /// the ViewerControl
        /// </summary>
        public bool IsDeactivationOnBackNavigationEnabled { get; set; } = false;

        /// <summary>
        /// When enabled it will force the ViewerControl to navigate back to the previous Frame page
        /// listening to the SystemNavigationManager.GetForCurrentView().BackRequested event
        /// </summary>
        /// <remarks>It will ignore any unsaved changes to the opened documents and just navigate away/back in the Frame navigation stack</remarks>
        public bool IsForceGoBackAllowed { get; set; } = false;

        /// <summary>
        /// Choose whether or not Tab Reordering is enabled
        /// </summary>
        public bool IsTabReorderEnabled
        {
            get => TabControlViewModel.IsTabReorderEnabled;
            set => TabControlViewModel.IsTabReorderEnabled = value;
        }

        public AppBarTabButton SaveCopyButton
        {
            get => _saveCopyButton;
            set => Set(ref _saveCopyButton, value);
        }

        public bool InitializeOnCreation { get; set; }

        public SharingHelper SharingHelper { get; set; }

        /// <summary>
        /// Controls whether the App Bar is visible on the screen or not
        /// </summary>
        public bool IsEntireAppBarOpen
        {
            get => _isEntireAppBarOpen;
            set
            {
                if (!value && IsFindTextDialogOpen)
                    return;

                if (Set(ref _isEntireAppBarOpen, value))
                {
                    CloseCurrentTool(true);
                    if (value)
                        AppbarInteraction();

                    UpdateUIVisiblity();
                }
            }
        }

        /// <summary>
        /// Controls whether the App Bar is in a expanded or compact state
        /// </summary>
        public bool IsAppBarOpen
        {
            get => _isAppBarOpen;
            set
            {
                if (value == IsAppBarOpen)
                    return;

                Set(ref _isAppBarOpen, value);

#if DEBUG
                Debug.WriteLine("IsAppBarOpen set to " + _isAppBarOpen);
#endif
                CloseCurrentTool(true);
                if (_isAppBarOpen)
                {
                    CloseDialogsOnAppBarOpen();
                    AppbarInteraction();
                }
                else
                {
                    _suppressRecentListUpdate = false;
                }

                UpdateUIVisiblity();
            }
        }

        public bool IsQuickSettingsOpen
        {
            get => _isQuickSettingsOpen;
            set
            {
                if (Set(ref _isQuickSettingsOpen, value))
                {
                    CloseCurrentTool();
                    if (_isQuickSettingsOpen && ViewerPageSettingsViewModel != null)
                        ViewerPageSettingsViewModel.UpdateSettings();
                }
            }
        }

        public bool IsDocumentEditable
        {
            get => PDFViewCtrl != null && _isDocumentEditable && !IsReflow;
            set => Set(ref _isDocumentEditable, value);
        }

        public bool IsOutlineEnabled => PDFViewCtrl != null;

        public bool IsQuickSettingsEnabled => PDFViewCtrl != null;

        public bool IsDocumentSearchable => PDFViewCtrl != null && !IsReflow && !IsPreparingForConversion;

        public bool IsPreparingForConversion => TabControlViewModel?.SelectedTab?.IsWaitingForConversionToStart == true;

        /// <summary>
        /// Change the visbility of the Save button
        /// </summary>
        public bool IsSaveButtonVisible
        {
            get
            {
                if (!IsSaveButtonEnabled)
                    return false;

                return _isSaveButtonVisible;
            }
            set => Set(ref _isSaveButtonVisible, value);
        }

        public bool IsUndoEnabled
        {
            get
            {
                if (TabControlViewModel?.SelectedTab?.NavigationStack != null)
                    return TabControlViewModel.SelectedTab.NavigationStack.IsUndoEnabled;

                return false;
            }
        }

        public bool IsRedoEnabled
        {
            get
            {
                if (TabControlViewModel?.SelectedTab?.NavigationStack != null)
                {
                    return TabControlViewModel.SelectedTab.NavigationStack.IsRedoEnabled;
                }

                return false;
            }
        }

        public bool IsUiPinned
        {
            get => _isUiPinned;
            set
            {
                if (Set(ref _isUiPinned, value))
                {
                    ViewerSettings.PinCommandBar = value;
                    UpdateUIVisiblity();
                }
            }
        }

        /// <summary>
        /// Whether or the current selected document has changes since last time it got saved
        /// </summary>
        public bool IsDocumentModifiedSinceLastSave
        {
            get
            {
                if (TabControlViewModel?.SelectedTab != null)
                {
                    return TabControlViewModel.SelectedTab.IsDocumentModifiedSinceLastSave;
                }

                return false;
            }
            set
            {
                if (value != TabControlViewModel?.SelectedTab?.IsDocumentModifiedSinceLastSave)
                {
                    if (TabControlViewModel.SelectedTab != null)
                        TabControlViewModel.SelectedTab.IsDocumentModifiedSinceLastSave = value;

                    RaisePropertyChanged();
                    RaisePropertyChanged(nameof(CanSaveDocument));
                }
            }
        }

        /// <summary>
        /// Choose wheather or not the Multi Window is Visible or Collapsed
        /// </summary>
        public bool IsMultiWindowVisible
        {
            get
            {
                if (!IsMultiWindowEnabled)
                    return false;

                return _isMultiWindowVisible;
            }
            set => Set(ref _isMultiWindowVisible, value);
        }

        /// <summary>
        /// Choose wheather or not the Multi Window is enabled or disabled
        /// </summary>
        public bool IsMultiWindowEnabled
        {
            get { return _isMultiWindowEnabled; }
            set => Set(ref _isMultiWindowEnabled, value);
        }

        /// <summary>
        /// Choose whether or not the Edit (annotation toolbar) is enabled or disabled
        /// </summary>
        public bool IsEditButtonEnabled
        {
            get { return _isEditButtonEnabled; }
            set => Set(ref _isEditButtonEnabled, value);
        }

        /// <summary>
        /// Choose whether or not the Outline button is enabled or disabled
        /// </summary>
        public bool IsOutlineButtonEnabled
        {
            get { return _isOutlineButtonEnabled; }
            set => Set(ref _isOutlineButtonEnabled, value);
        }

        /// <summary>
        /// Choose whether or not the TextSearch button is enabled or disabled
        /// </summary>
        public bool IsTextSearchButtonEnabled
        {
            get { return _isTextSearchButtonEnabled; }
            set => Set(ref _isTextSearchButtonEnabled, value);
        }

        /// <summary>
        /// Choose whether or not the ViewMode button is enabled or disabled
        /// </summary>
        public bool IsViewModeButtonEnabled
        {
            get { return _isViewModeButtonEnabled; }
            set => Set(ref _isViewModeButtonEnabled, value);
        }

        /// <summary>
        /// Choose whether or not the SeeMore is enabled or disabled
        /// </summary>
        public bool IsSeeMoreButtonEnabled
        {
            get { return _isSeeMoreButtonEnabled; }
            set => Set(ref _isSeeMoreButtonEnabled, value);
        }

        /// <summary>
        /// Choose wheather or not the Multi Window on the viewer is Active
        /// </summary>
        public bool IsMultiWindowActive
        {
            get
            {
                if (PDFViewCtrl == null)
                    return false;

                return _isMultiWindowActive;
            }
            set => Set(ref _isMultiWindowActive, value);
        }

        /// <summary>
        /// Choose wheather or not the PDF Compare is Visible or Collapsed
        /// </summary>
        public bool IsPdfCompareVisible
        {
            get 
            {
                if (!IsPdfCompareEnabled)
                    return false;

                return _isPdfCompareVisible;
            }
            set => Set(ref _isPdfCompareVisible, value);
        }

        /// <summary>
        /// Choose wheather or not the PDF Compare is enabled or disabled
        /// </summary>
        public bool IsPdfCompareEnabled
        {
            get { return _isPdfCompareEnabled; }
            set => Set(ref _isPdfCompareEnabled, value);
        }

        /// <summary>
        /// Choose wheather or not the PDF Compare on the viewer is Active
        /// </summary>
        public bool IsPdfCompareActive
        {
            get
            {
                if (PDFViewCtrl == null)
                    return false;

                return _isPdfCompareActive;
            }
            set => Set(ref _isPdfCompareActive, value);
        }

        /// <summary>
        /// Enable or disable the Save button
        /// </summary>
        public bool IsSaveButtonEnabled
        {
            get { return _isSaveButtonEnabled; }
            set 
            {
                _isSaveButtonEnabled = value;
                RaisePropertyChanged(nameof(IsSaveButtonVisible));
            }
        }

        /// <summary>
        /// Enable or disable the Add tab button "+" next to last tab
        /// </summary>
        public bool IsAddTabButtonEnabled
        {
            get => TabControlViewModel.IsAddTabButtonEnabled;
            set => TabControlViewModel.IsAddTabButtonEnabled = value;        
        }

        /// <summary>
        /// Enable or disable the Close tab button "X" within active tabs
        /// </summary>
        public bool IsCloseTabButtonEnabled
        {
            get { return _isCloseTabButtonEnabled; }
            set => Set(ref _isCloseTabButtonEnabled, value);
        }

        /// <summary>
        /// Tells when the selected document can be saved
        /// </summary>
        public bool CanSaveDocument 
        {
            get
            {
                if ((IsDocumentModifiedSinceLastSave == true)
                    && (CurrentDocumentState == OpenedDocumentStates.Normal)
                    || (CurrentDocumentState == OpenedDocumentStates.NeedsFullSave))

                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }

        /// <summary>
        /// Enable or disable the Save As button
        /// </summary>
        public bool IsSaveAsButtonEnabled
        {
            get { return _isSaveAsButtonEnabled; }
            set
            {
                _isSaveAsButtonEnabled = value;
                RaisePropertyChanged(nameof(IsSaveAsButtonVisible));
            }
        }

        /// <summary>
        /// Change the visibility of the Save As button
        /// </summary>
        public bool IsSaveAsButtonVisible
        {
            get 
            {
                if (!IsSaveAsButtonEnabled)
                    return false;

                return _isSaveAsButtonVisible;
            }
            set => Set(ref _isSaveAsButtonVisible, value);
        }

        /// <summary>
        /// Enable or disable the Actions button in the viewer
        /// </summary>
        public bool IsActionsEnabled
        {
            get => _isActionsEnabled;
            set
            {
                _isActionsEnabled = value;
                IsActionsButtonVisible = IsActionsEnabled;
            }
        }

        /// <summary>
        /// Change the visibility of the Share document button in the viewer
        /// </summary>
        public bool IsShareVisible
        {
            get => _isShareVisible;
            set => Set(ref _isShareVisible, value);
        }

        /// <summary>
        /// Change the visibility of the Print document button in the viewer
        /// </summary>
        public bool IsPrintVisible
        {
            get => _isPrintVisible;
            set => Set(ref _isPrintVisible, value);
        }

        /// <summary>
        /// Change the visibility of the Export XFDF button in the viewer
        /// </summary>
        public bool IsExportXFDFVisible
        {
            get => _isExportXFDFVisible;
            set => Set(ref _isExportXFDFVisible, value);
        }

        /// <summary>
        /// Change the visibility of the Actions button in the viewer
        /// </summary>
        public bool IsActionsButtonVisible
        {
            get => _isActionButtonVisible;
            set => Set(ref _isActionButtonVisible, value);
        }

        public bool IsSaveTimerOpen
        {
            get => _isSaveTimerOpen;
            set
            {
                if (Set(ref _isSaveTimerOpen, value))
                {
                    IsModal = value;
                    RaisePropertyChanged();
                }
            }
        }

        public bool IsAnnotationToolbarOpen
        {
            get => _isAnnotationToolbarOpen;
            set
            {
                if (Set(ref _isAnnotationToolbarOpen, value))
                    IsAppBarOpen = false;
            }
        }

        /// <summary>
        /// Used to Set whether navigating using the outline dialog should close it.
        /// </summary>
        public bool OutlineDialogNavigationCloses
        {
            get => _outlineDialogNavigationCloses;
            set
            {
                if (Set(ref _outlineDialogNavigationCloses, value) && OutlineDialogViewModel != null)
                {
                    OutlineDialogViewModel.CloseDialogOnItemClick = _outlineDialogNavigationCloses;
                }
            }
        }

        public bool IsOutlineDialogOpen
        {
            get => _isOutlineDialogOpen;
            set
            {
                if (Set(ref _isOutlineDialogOpen, value))
                {
                    ViewerSettings.OutlineDialogOpen = value;
                    PdfViewCtrlTabInfo tab = TabControlViewModel.SelectedTab;
                    if (tab != null)
                    {
                        if (_isOutlineDialogOpen)
                        {
                            OutlineDialogViewModel = tab.OutlineDialogViewModel;
                            OutlineDialogViewModel.IsPivotItemThumbnailsEnabled = IsPivotItemThumbnailsEnabled;
                            OutlineDialogViewModel.IsPivotItemBookmarksEnabled = IsPivotItemBookmarksEnabled;
                            OutlineDialogViewModel.IsPivotItemOutlineEnabled = IsPivotItemOutlineEnabled;
                            OutlineDialogViewModel.IsPivotItemAnnotationsEnabled = IsPivotItemAnnotationsEnabled;
                            OutlineDialogViewModel.IsOutlineEditingEnabled = IsOutlineEditingEnabled;
                        }
                        else
                        {
                            tab.CloseOutline();
                        }
                    }

                    if (!_isOutlineDialogOpen)
                    {
                        OutlineDialogViewModel = null;
                        if (PDFViewCtrl != null)
                        {
                            OutlineDialogViewModel = null;
                            PDFViewCtrl.Focus(FocusState.Programmatic);
                        }
                    }
                }
            }
        }

        public bool IsFindTextDialogOpen
        {
            get => _isFindTextDialogOpen;
            set
            {
                if (Set(ref _isFindTextDialogOpen, value))
                {
                    CloseCurrentTool();
                    if (_isFindTextDialogOpen)
                    {
                        IsAppBarOpen = false;
                        _appBarClosingTimer?.Stop();
                        IsEntireAppBarOpen = true;

#if SUBSCRIPTIONS
                        if (ViewerSettings.IsSubscriptionActivated)
                        {
                            if (IsPdfCompareEnabled)
                                IsPdfCompareVisible = false;

                            if (IsMultiWindowEnabled)
                                IsMultiWindowVisible = false;
                        }
                        else
                        {
                            IsPdfCompareSubVisible = false;
                            IsMultiWindowSubVisible = false;
                            IsDiamondSubVisible = false;
                        }
#else
                            if (IsPdfCompareEnabled)
                                IsPdfCompareVisible = false;

                            if (IsMultiWindowEnabled)
                                IsMultiWindowVisible = false;
#endif
                            if (IsEditButtonEnabled)
                                IsEditButtonVisible = false;

                            if (IsOutlineButtonEnabled)
                                IsOutlineButtonVisible = false;

                            if (IsViewModeButtonEnabled)
                                IsViewModeButtonVisible = false;

                            if (IsSeeMoreButtonEnabled)
                                IsSeeMoreButtonVisible = false;

                        FindTextViewModel = new FindTextViewModel(PDFViewCtrl, ViewerSettings.ColorMode);
                        FindTextViewModel.FindTextClosed += FindTextViewModel_FindTextClosed;
                        FindTextViewModel.FindTextResultFound += FindTextViewModel_FindTextResultFound;
                    }
                    else
                    {
#if SUBSCRIPTIONS
                        if (ViewerSettings.IsSubscriptionActivated)
                        {
                            if (IsPdfCompareEnabled)
                                IsPdfCompareVisible = true;

                            if (IsMultiWindowEnabled)
                                IsMultiWindowVisible = true;
                        }
                        else
                        {
                            IsPdfCompareSubVisible = true;
                            IsMultiWindowSubVisible = true;
                            IsDiamondSubVisible = true;
                        }
#else
                            if (IsPdfCompareEnabled)
                                IsPdfCompareVisible = true;

                            if (IsMultiWindowEnabled)
                                IsMultiWindowVisible = true;
#endif
                            if (IsEditButtonEnabled)
                                IsEditButtonVisible = true;

                            if (IsOutlineButtonEnabled)
                                IsOutlineButtonVisible = true;

                            if (IsTextSearchButtonEnabled)
                                IsTextSearchButtonVisible = true;

                            if (IsViewModeButtonEnabled)
                                IsViewModeButtonVisible = true;

                            if (IsSeeMoreButtonEnabled)
                                IsSeeMoreButtonVisible = true;

                        if (FindTextViewModel != null)
                        {
                            FindTextViewModel.ClearViewModel();
                            AppbarInteraction();
                        }
                    }
                }

                if (FindTextViewModel == null && PDFViewCtrl != null)
                    FindTextViewModel = new FindTextViewModel(PDFViewCtrl, ViewerSettings.ColorMode);

                if (FindTextViewModel != null)
                    FindTextViewModel.IsInSearchMode = value;
            }
        }

        public bool IsAllowDropEnabled
        {
            get => _isAllowDropEnabled;
            set => Set(ref _isAllowDropEnabled, value);
        }

        public SaveAsOption CurrentSaveAsOption
        {
            get => _currentSaveAsOption;
            set => Set(ref _currentSaveAsOption, value);
        }

        public string SaveAPasswordCopyText
        {
            get => _saveAPasswordCopyText;
            set => Set(ref _saveAPasswordCopyText, value);
        }

        private void FindTextViewModel_FindTextClosed(Deferral deferral)
        {
            try
            {
                IsFindTextDialogOpen = false;
            }
            finally
            {
                deferral.Complete();
            }
        }

        private void FindTextViewModel_FindTextResultFound(Deferral deferral) => FindTextResultFound?.Invoke(deferral);

        public ClickablePageNumberIndicator PageNumberIndicator => _pageNumberIndicator;

        private bool DoToolsAllowZoom
        {
            set
            {
                bool before = AreZoombuttonsVisible;
                _doToolsAllowZoom = value;
                if (before != AreZoombuttonsVisible)
                    RaisePropertyChanged(nameof(AreZoombuttonsVisible));
            }
        }

        private bool IsInputMouse
        {
            get => _isInputMouse;
            set
            {
                bool before = AreZoombuttonsVisible;
                _isInputMouse = value;
                if (before != AreZoombuttonsVisible)
                    RaisePropertyChanged(nameof(AreZoombuttonsVisible));
            }
        }

        public bool AreAudioButtonsVisible
        {
            get => _areAudioButtonsVisible;
            set => Set(ref _areAudioButtonsVisible, value);
        }

        public bool IsAudioPlaying
        {
            get => _isAudioPlaying;
            set => Set(ref _isAudioPlaying, value);
        }

        public bool IsAudioControlVisible
        {
            get => _isAudioControlVisible;
            set => Set(ref _isAudioControlVisible, value);
        }

#if SUBSCRIPTIONS
        /// <summary>
        /// Choose wheather or not the PDF Compare Subscription is Visible or Collapsed
        /// </summary>
        public bool IsPdfCompareSubVisible
        {
            get
            {
                return _isPdfCompareSubVisible;
            }
            set => Set(ref _isPdfCompareSubVisible, value);
        }

        /// <summary>
        /// Choose wheather or not the Multi Window Subscription is Visible or Collapsed
        /// </summary>
        public bool IsMultiWindowSubVisible
        {
            get
            {
                return _isMultiWindowSubVisible;
            }
            set => Set(ref _isMultiWindowSubVisible, value);
        }

        /// <summary>
        /// Choose wheather or not the Subscription Diamond is Visible or Collapsed
        /// </summary>
        public bool IsDiamondSubVisible
        {
            get
            {
                return _isDiamondSubVisible;
            }
            set => Set(ref _isDiamondSubVisible, value);
        }

#endif

        private void PDFViewCtrl_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType == PointerDeviceType.Mouse)
                IsInputMouse = true;
        }

        private void PDFViewCtrl_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerDeviceType != PointerDeviceType.Mouse)
                IsInputMouse = false;
        }

        private void ToolManager_NewToolCreated(ToolType toolType)
        {
            DoToolsAllowZoom = toolType != ToolType.e_ink_create && toolType != ToolType.e_text_annot_create;

            if (toolType == ToolType.e_ink_create)
            {
                //IsAppBarAvailable = false;        // Don't hide app bar with annotation command bar
            }
            else
            {
                //IsAppBarAvailable = !IsModal;
            }
        }

        private void ToolManager_SingleTap(SingleTapEvents eventType, TappedRoutedEventArgs args)
        {
            Debug.WriteLine("Tapped: " + eventType.ToString());
            if (eventType == SingleTapEvents.BeforeTool && IsEntireAppBarOpen && !IsAnnotationToolbarOpen && !ViewerSettings.PinCommandBar)
            {
                if (!IsAppBarOpen)
                    IsEntireAppBarOpen = false;

                IsAppBarOpen = false;
                args.Handled = true;
            }
            else if (eventType == SingleTapEvents.BeforeTool && IsAnnotationToolbarOpen)
            {
                if (AnnotationToolbar.IsAppBarOpen)
                {
                    AnnotationToolbar.IsAppBarOpen = false;
                    args.Handled = true;
                }
            }
            else if (eventType == SingleTapEvents.SingleTapConfirmed && !ViewerSettings.PinCommandBar && !IsEntireAppBarOpen)
            {
                IsAppBarOpen = false;
                IsEntireAppBarOpen = true;
                args.Handled = true;
            }
        }

        public bool AreZoombuttonsVisible
        {
            get
            {
                if (PDFViewCtrl == null)
                    return false;

                if (!_isInputMouse || !_doToolsAllowZoom)
                    return false;

                if (ViewerSettings.AlwaysShowPageNavigationTools)
                    return true;

                if (!ViewerSettings.PinCommandBar)
                {
                    if (IsAnnotationToolbarOpen)
                    {
                        return AnnotationToolbar.IsAppBarOpen;
                    }
                    else
                    {
                        return IsEntireAppBarOpen;
                    }
                }

                return true;
            }
        }

        public bool IsPageNumberIndicatorVisible
        {
            get
            {
                if (ViewerSettings.AlwaysShowPageNavigationTools)
                    return true;

                if (ViewerSettings.PinCommandBar)
                {
                    return true;
                }
                else
                {
                    if (!IsAnnotationToolbarOpen)
                    {
                        return IsEntireAppBarOpen;
                    }
                    else
                    {
                        return _annotationToolbar.IsAppBarOpen;
                    }
                }
            }
        }

        public bool IsThumbnailSliderVisible
        {
            get
            {
                if (PDFViewCtrl == null)
                    return false;

                if (ThumbnailSlider == null)
                    return false;

                if (ViewerSettings.AlwaysShowPageNavigationTools)
                    return true;

                if (IsPdfCompareViewOpen)
                    return false;

                if (ViewerPageSettingsViewModel != null && ViewerPageSettingsViewModel.IsRightToLeftEnabled)
                {
                    ThumbnailSlider.FlowDirection = FlowDirection.RightToLeft;
                }
                else
                {
                    ThumbnailSlider.FlowDirection = FlowDirection.LeftToRight;
                }

                if (IsUiPinned)
                {
                    return IsAppBarOpen || AnnotationToolbar.IsAppBarOpen ;
                }
                else if (IsAnnotationToolbarOpen)
                {
                    return AnnotationToolbar.IsAppBarOpen;
                }

                return IsEntireAppBarOpen;
            }
        }

        public bool IsConverting => PDFViewCtrl == null || TabControlViewModel?.SelectedTab?.PDFViewCtrl?.IsConverting == true;

        private void ResolveThumbnailSlider()
        {
            if (ThumbnailSlider == null && !PDFViewCtrl.IsConverting && PDFViewCtrl.GetPageCount() >= PAGES_NEEDED_FOR_THUMB_SLIDER)
            {
                AddThumbnailSlider(TabControlViewModel.SelectedTab);
            }
            else if (ThumbnailSlider != null && PDFViewCtrl.GetPageCount() < PAGES_NEEDED_FOR_THUMB_SLIDER)
            {
                ThumbnailSlider.PDFViewCtrl = null;
                ThumbnailSlider = null;
            }

            ThumbnailSlider?.UpdateMaximum();
        }

        public void UpdateUIVisiblity()
        {
            RaisePropertyChanged(nameof(IsPageNumberIndicatorVisible));
            RaisePropertyChanged(nameof(IsThumbnailSliderVisible));
            RaisePropertyChanged(nameof(AreZoombuttonsVisible));
            RaisePropertyChanged(nameof(AreAudioButtonsVisible));
        }

        private StorageFile CurrentFile => TabControlViewModel.SelectedTab.OriginalFile;

        private StorageFile TemporaryFile => TabControlViewModel.SelectedTab.PdfFile;

        private string CurrentPassword
        {
            get => TabControlViewModel.SelectedTab.MetaData.Password;
            set => TabControlViewModel.SelectedTab.MetaData.Password = value;
        }

        protected string MRUToken
        {
            set => TabControlViewModel.SelectedTab.MetaData.MostRecentlyUsedToken = value;
        }

        public static Dictionary<OpenedDocumentFileTypes, string> FileTypeNameDictionary = new Dictionary<OpenedDocumentFileTypes, string>
        {
            { OpenedDocumentFileTypes.pdf, "pdf" },
            { OpenedDocumentFileTypes.xps, "xps" }
        };

        public OpenedDocumentStates CurrentDocumentState
        {
            get
            {
                if (TabControlViewModel.SelectedTab == null)
                    return OpenedDocumentStates.None;

                return TabControlViewModel.SelectedTab.DocumentState;
            }
            set
            {
                if (TabControlViewModel.SelectedTab != null)
                    TabControlViewModel.SelectedTab.DocumentState = value;
            }
        }

        public bool CanSaveToSamefile => CurrentDocumentState == OpenedDocumentStates.Normal || CurrentDocumentState == OpenedDocumentStates.Created;

        public OpenedDocumentFileTypes CurrentDocumentFileType { get; set; } = OpenedDocumentFileTypes.pdf;

        private void InitCommands()
        {
            GoToFilesCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(GoToFilesAsync);
                command.Execute(parameter);
            });
            PinCommand = new RelayCommand(PinCommandImpl);
            UnpinCommand = new RelayCommand(UnpinCommandImpl);
            SaveCommand = new RelayCommand(parameter =>
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Save" });

                var command = AsyncCommand.Create(SaveAsync);
                command.Execute(parameter);
            });
            SaveAsCommand = new RelayCommand(parameter =>
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Save As" });

                var command = AsyncCommand.Create(SaveAsAsync);
                command.Execute(parameter);
            });

            SearchCommand = new RelayCommand(SearchCommandImpl);
            OutlineCommand = new RelayCommand(OutlineCommandImpl);
            EditCommand = new RelayCommand(EditCommandImpl);
            SaveAndFlattenCommand = new RelayCommand(parameter =>
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Flatten File" });

                var command = AsyncCommand.Create(SaveAndFlattenAsync);
                command.Execute(parameter);
            });
            SaveReducedFileSizeCopyCommand = new RelayCommand(SaveReducedFileSizeCopy);
            SavePasswordCopyCommand = new RelayCommand(SavePasswordCopy);
            SavePermCroppedCopyCommand = new RelayCommand(parameter =>
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Crop Pages" });

                var command = AsyncCommand.Create(SavePermCroppedCopyAsync);
                command.Execute(parameter);
            }, _ => true);

            PageEditingCommand = new RelayCommand(EditPages);
            PdfCompareCommand = new RelayCommand(ComparePDF);

            MultiWindowCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(MultiWindowImpl);
                command.Execute(parameter);
            });

            PrintCommand = new RelayCommand(parameter =>
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Print File" });

                var command = AsyncCommand.Create(PrintAsync);
                command.Execute(parameter);
            });
            ShareCommand = new RelayCommand(ShareCommandImpl);
            ZoomCommand = new RelayCommand(ZoomCommandImpl);

            ExportXFDFCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExportXFDFAsync);
                command.Execute(parameter);
            });

            PlayAudioCommand = new RelayCommand(PlayAudioCommandImpl);
            PauseAudioCommand = new RelayCommand(PauseAudioCommandImpl);
            StopAudioCommand = new RelayCommand(StopAudioCommandImpl);

            UndoCommand = new RelayCommand(UndoCommandImpl);
            RedoCommand = new RelayCommand(RedoCommandImpl);

            CloseSearchCommand = new RelayCommand(CloseSearchCommandImpl);
            FullScreenExitCommand = new RelayCommand(FullScreenExitCommandImpl);

            DismissDialogCommand = new RelayCommand(DismissDialogCommandImpl);
            QuickSettingsOpenedCommand = new RelayCommand(QuickSettingsOpenedCommandImpl);
            QuickSettingsClosedCommand = new RelayCommand(QuickSettingsClosedCommandImpl);
        }

        public virtual bool GetSubscribed() => false;

        private void PinCommandImpl(object CommandName) => IsUiPinned = true;

        private void UnpinCommandImpl(object CommandName)
        {
            IsUiPinned = false;
            IsEntireAppBarOpen = false;
        }

        private void SearchCommandImpl(object CommandName)
        {
            CloseCurrentTool();
            AppbarInteraction();
            TabControlViewModel?.SelectedTab?.ToolManager?.CloseOpenDialog();
            IsFindTextDialogOpen = true;

            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Text Search" });
        }

        private void OutlineCommandImpl(object CommandName)
        {
            CloseCurrentTool();
            AppbarInteraction();
            TabControlViewModel?.SelectedTab?.ToolManager?.CloseOpenDialog();
            IsOutlineDialogOpen = !IsOutlineDialogOpen;
            IsAppBarOpen = false;
            if (!ViewerSettings.PinCommandBar)
                IsEntireAppBarOpen = false;

            if (IsOutlineDialogOpen)
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Open Outline" });
        }

        private void EditCommandImpl(object CommandName)
        {
            AppbarInteraction();
            if (IsConverting)
                return;

            CloseCurrentTool();
            TabControlViewModel?.SelectedTab?.ToolManager?.CloseOpenDialog();
            _appBarClosingTimer?.Stop();
            IsAnnotationToolbarOpen = true;
            IsAppBarOpen = false;
            UpdateUIVisiblity();

            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Open Annotation Toolbar" });
        }

        private async Task TryPrintAsync()
        {
            if (PDFViewCtrl != null)
            {
                PDFDoc doc = PDFViewCtrl.GetDoc();
                if (doc != null && TabControlViewModel.SelectedTab != null)
                {
                    PDFPrintManager.RegisterForPrintingContract(doc, TabControlViewModel.SelectedTab.Title);
                    try
                    {
                        await PrintManager.ShowPrintUIAsync();
                    }
                    catch
                    {
                        //Swallow exception
                    }
                }
            }
        }

        private void ShareCommandImpl(object CommandName) => DataTransferManager.ShowShareUI();

        private async Task ExportXFDFAsync(object action)
        {
            if (PDFViewCtrl != null)
            {
                PDFDoc doc = PDFViewCtrl.GetDoc();

                var fdf = await Task.Run(() => doc.FDFExtract(PDFDocExtractFlag.e_both));

                var savePicker = new FileSavePicker
                {
                    SuggestedStartLocation = PickerLocationId.DocumentsLibrary
                };

                savePicker.FileTypeChoices.Add("XML File", new List<string> { ".xml" });
                savePicker.SuggestedFileName = TabControlViewModel.SelectedTab.Title + "_XFDF";

                StorageFile file = await savePicker.PickSaveFileAsync();
                if (file != null)
                {
                    var stream = await file.OpenAsync(FileAccessMode.ReadWrite);
                    IFilter filter = new RandomAccessStreamFilter(stream);

                    await fdf.SaveAsXFDFAsync(filter);
                }
            }
        }

        private void PlayAudioCommandImpl(object action)
        {
            IsAudioPlaying = true;
            MediaElement.Play();
        }

        private void PauseAudioCommandImpl(object action)
        {
            IsAudioPlaying = false;
            MediaElement.Pause();
        }

        private void StopAudioCommandImpl(object action)
        {
            AreAudioButtonsVisible = false;
            MediaElement.Stop();
        }

        private void UndoCommandImpl(object param)
        {
            CloseCurrentTool();
            AppbarInteraction();
            TabControlViewModel.SelectedTab.NavigationStack.Undo();
        }

        private void RedoCommandImpl(object param)
        {
            CloseCurrentTool();
            AppbarInteraction();
            TabControlViewModel.SelectedTab.NavigationStack.Redo();
        }

        private void CloseSearchCommandImpl(object parameter)
        {
            IsFindTextDialogOpen = false;
        }

        private void FullScreenExitCommandImpl(object parameter) => IsEntireAppBarOpen = true;

        private void DismissDialogCommandImpl(object commandName) => IsDismissableDialogOpen = false;

        private void QuickSettingsOpenedCommandImpl(object sender)
        {
            if (TabControlViewModel == null || TabControlViewModel.SelectedTab == null)
                return;

            if (ViewerPageSettingsViewModel == null)
            {
                ViewerPageSettingsViewModel = new ViewerPageSettingsViewModel(TabControlViewModel.SelectedTab);
                ViewerPageSettingsViewModel.IsRightToLeftVisible = IsRightToLeftVisibile;
                ViewerPageSettingsViewModel.IsThumbnailViewVisible = IsThumbnailViewVisible;
                ViewerPageSettingsViewModel.IsCropViewVisible = IsCropViewVisible;
                ViewerPageSettingsViewModel.PropertyChanged += ViewerPageSettingsViewModel_PropertyChanged;
                ViewerPageSettingsViewModel.ViewRequestedHandler += ViewerPageSettingsViewModel_ViewRequestedHandler;
            }

            CloseCurrentTool();
            AppbarInteraction();
            IsQuickSettingsOpen = true;
            ViewerPageSettingsViewModel.UpdateSettings();
            _appBarClosingTimer?.Stop();

            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Quick Settings" });
        }

        private void QuickSettingsClosedCommandImpl(object sender)
        {
            CloseCurrentTool();
            IsQuickSettingsOpen = false;
            AppbarInteraction();
        }

        private bool HasUserBeenWarnedAboutSaving
        {
            get => TabControlViewModel.SelectedTab.HasUserBeenWarnedAboutSaving;
            set => TabControlViewModel.SelectedTab.HasUserBeenWarnedAboutSaving = value;
        }

        public void ResolveAutoSaverAndSaveButton()
        {
            if (TabControlViewModel.SelectedTab == null || TabControlViewModel.SelectedTab.Doc == null)
            {
                TurnOffAutoSave();
                IsSaveButtonVisible = false;
            }

            bool autoSaveOn = ViewerSettings.AutoSaveOn;
            if (autoSaveOn)
            {
                TabControlViewModel.ShowIfDocumentModified = ShowIfModifiedStates.ShowIfModidiedAndCantSave;
            }
            else
            {
                TabControlViewModel.ShowIfDocumentModified = ShowIfModifiedStates.ShowIfModified;
            }

            if (TabControlViewModel.SelectedTab != null)
                TabControlViewModel.SelectedTab.ShowIfDocumentModified = TabControlViewModel.ShowIfDocumentModified;

            IsSaveAsButtonVisible = CurrentDocumentState != OpenedDocumentStates.Uneditable;
            
            if (IsActionsEnabled)
                IsActionsButtonVisible = CurrentDocumentState != OpenedDocumentStates.Uneditable;
            
            if (autoSaveOn)
            {
                IsSaveButtonVisible = false;
                if (CurrentDocumentState == OpenedDocumentStates.Normal)
                {
                    TurnOnAutoSaving();
                }
                else
                {
                    TurnOffAutoSave();
                }
            }
            else
            {
                IsSaveButtonVisible = CurrentDocumentState == OpenedDocumentStates.Normal || CurrentDocumentState == OpenedDocumentStates.NeedsFullSave;
                TurnOffAutoSave();
            }

            RaisePropertyChanged(nameof(CanSaveDocument));
        }

        private void TurnOnAutoSaving()
        {
            if (_autoSaveHelper == null)
            {
                _autoSaveHelper = new AutoSaveHelper(ViewerSettings.AutoSaveInterval);
                _autoSaveHelper.DocumentWasSaved += AutoSaveHelper_DocumentWasSaved;
                _autoSaveHelper.CommitChangesAsync += AutoSaveHelper_CommitChangesAsync;
            }

            _autoSaveHelper.Stop();
            _autoSaveHelper.CurrentFile = CurrentFile;
            _autoSaveHelper.TemporaryFile = TemporaryFile;
            _autoSaveHelper.PDFDoc = TabControlViewModel.SelectedTab.Doc;
            _autoSaveHelper.Start();
        }

        private Task AutoSaveHelper_CommitChangesAsync()
        {
            //if (OutlineDialogViewModel != null && OutlineDialogViewModel.HasUnsavedUserbookmarks)
            //{
            //    OutlineDialogViewModel.UserBookmarks.SaveBookmarks();
            //    return OutlineDialogViewModel.UserBookmarks.WaitForBookmarkSavingAsync();
            //}
            return null;
        }

        private void AutoSaveHelper_DocumentWasSaved(DateTimeOffset newModifiedDate, ulong newFileSize)
        {
            IsDocumentModifiedSinceLastSave = false;
            if (TabControlViewModel.SelectedTab != null)
            {
                TabControlViewModel.SelectedTab.MetaData.LastModifiedDate = newModifiedDate;
                TabControlViewModel.SelectedTab.MetaData.FileSize = newFileSize;
                TabControlViewModel.SelectedTab.SystemLastModifiedDate = newModifiedDate;
                TabControlViewModel.SelectedTab.SystemFileSize = newFileSize;
            }
        }

        private void TurnOffAutoSave()
        {
            if (_autoSaveHelper != null)
            {
                _autoSaveHelper.Stop();
                _autoSaveHelper = null;
            }
        }

        public void PauseAutoSaving() => _autoSaveHelper?.Stop();

        public void ResumeAutoSaving() => _autoSaveHelper?.Start();

        private async void ToolManager_AnnotationModified(IAnnot annotation, int pageNumber)
        {
            await HandleDocumentEditingAsync();
            if (pageNumber == 1)
                TabControlViewModel.SelectedTab.MetaData.IsThumbnailUpToDate = false;

            if (IsOutlineDialogOpen)
                OutlineDialogViewModel.Thumbnails.PageModified(pageNumber);
        }

        private async void ToolManager_AnnotationGroupModified(Dictionary<IAnnot, int> annotationGroup)
        {
            await HandleDocumentEditingAsync();
            if (annotationGroup.ContainsValue(1))
                TabControlViewModel.SelectedTab.MetaData.IsThumbnailUpToDate = false;

            if (IsOutlineDialogOpen)
            {
                foreach (var item in annotationGroup)
                {
                    OutlineDialogViewModel.Thumbnails.PageModified(item.Value);
                }
            }
        }

        private void ToolManager_TextToSpeechActivated(MediaElement media)
        {
            AreAudioButtonsVisible = true;
            IsAudioPlaying = true;
            MediaElement = media;
        }

        #region Sound Annotation / AudioControl

        private async void ToolManager_OnSoundAnnotRequested(Windows.Foundation.Point target, int pageNumber)
        {
            if (SoundControl != null)
                return;

            if (UtilityFunctions.IsAudioCaptureAvailable() == false)
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("pdftron_Controls_AudioControl_SoundAnnot_Title"),
                    Content = _loader.GetString("pdftron_Controls_AudioControl_MicNotFound"),
                    CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);

                return;
            }

            SoundControl = new AudioControl(PDFViewCtrl, target);
            SoundControl.CloseRequested += SoundControl_CloseRequested;
            SoundControl.RecordingCompleted += SoundControl_RecordingCompleted;

            var result = await SoundControl.InitializeAsync();

            if (result == false)
            {
                await DisposeAudioControlAsync();
                return;
            }

            IsAudioControlVisible = true;
        }

        private async void SoundControl_RecordingCompleted(StorageFile file, pdftron.PDF.Rect target, int pageNumber)
        {
            if (file == null)
                return;

            bool isSuccess = false;
            pdftron.PDF.Annots.Sound soundAnnot = null;

            try
            {
                if (PDFViewCtrl.DocTryLockRead() == false)
                    return;

                soundAnnot = pdftron.PDF.Annots.Sound.CreateWithData(
                    PDFViewCtrl.GetDoc().GetSDFDoc(),
                    target, 
                    new MappedFile(file.Path),
                    (int)UtilityFunctions.PCM_AUDIO_BITS_PER_SAMPLE,
                    (int)UtilityFunctions.PCM_AUDIO_SAMPLE_RATE,
                    (int)UtilityFunctions.PCM_AUDIO_NUM_CHANNELS);

                soundAnnot.SetIcon(pdftron.PDF.Annots.SoundIcon.e_Speaker);
                soundAnnot.GetSoundStream().PutName("E", "Signed");
                soundAnnot.RefreshAppearance();

                UtilityFunctions.SetUniqueId(soundAnnot, Guid.NewGuid().ToString());

                var page = PDFViewCtrl.GetDoc().GetPage(pageNumber);
                page.AnnotPushBack(soundAnnot);

                PDFViewCtrl.UpdateWithAnnot(soundAnnot, pageNumber);

                isSuccess = true;
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
            finally
            {
                PDFViewCtrl.DocUnlockRead();
            }

            if (isSuccess)
            {
                ToolManager.RaiseAnnotationAddedEvent(soundAnnot, pageNumber);

                UtilityFunctions.ShowToastNotification(
                    _loader.GetString("pdftron_Controls_AudioControl_RecComplete_Title"), 
                    _loader.GetString("pdftron_Controls_AudioControl_RecComplete_Desc") + " " + pageNumber.ToString());
            }

            await DisposeAudioControlAsync();
        }

        private async void SoundControl_CloseRequested()
        {
            await DisposeAudioControlAsync();
        }

        private async Task DisposeAudioControlAsync()
        {
            if (SoundControl != null)
            {
                SoundControl.CloseRequested -= SoundControl_CloseRequested;
                SoundControl.RecordingCompleted -= SoundControl_RecordingCompleted;
                
                SoundControl.Dispose();
                SoundControl = null;
            }

            IsAudioControlVisible = false;

            await DocumentManager.CleanupTempAudioFilesFromDiskAsync();
        }

        #endregion

        private void ThumbnailViewerPageMovedHandler(int pageNumber, int newLocation)
        {
            _navigationStateBeforeThumbnailsviewOpening = null;
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            _editedSinceThumbnailViewerOpened = true;
            if (pageNumber == 1 || newLocation == 1)
                _hasPage1ChangeSinceThumbnailViewOpened = true;
        }

        private void ThumbnailViewerPageDeletedHandler(int pageNumber)
        {
            _navigationStateBeforeThumbnailsviewOpening = null;
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            _editedSinceThumbnailViewerOpened = true;
            if (pageNumber == 1)
                _hasPage1ChangeSinceThumbnailViewOpened = true;

            PageNumberIndicator?.UpdatePageNumbers();
        }

        private void ThumbnailViewerPageAddedHandler(int pageNumber, ThumbnailItem item)
        {
            _navigationStateBeforeThumbnailsviewOpening = null;
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            _editedSinceThumbnailViewerOpened = true;
            if (pageNumber == 1)
                _hasPage1ChangeSinceThumbnailViewOpened = true;

            // Need to call this since adding page doesnt cause core to fire PageNumberChanged
            PageNumberIndicator?.UpdatePageNumbers();
        }

        private void ThumbnailViewerPageRotatedDelegate(int pageNumber)
        {
            _navigationStateBeforeThumbnailsviewOpening = null;
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            _editedSinceThumbnailViewerOpened = true;
            if (pageNumber == 1)
                _hasPage1ChangeSinceThumbnailViewOpened = true;
        }

        /// <summary>
        /// Make sure any doc modification has been notified
        /// </summary>
        private async Task HandleDocumentEditingAsync()
        {
            HasDocumentBeenModifiedSinceOpening = true;
            if (!HasUserBeenWarnedAboutSaving)
            {
                HasUserBeenWarnedAboutSaving = true;
                if (CurrentDocumentState == OpenedDocumentStates.Created || CurrentDocumentState == OpenedDocumentStates.NonePDF)
                {
                    await NotifyCreatedDocumentAsync();
                }
                else if (CurrentDocumentState == OpenedDocumentStates.ReadOnly)
                {
                    await NotifyReadOnlyDocumentAsync();
                }
                else if (CurrentDocumentState == OpenedDocumentStates.Corrupted || CurrentDocumentState == OpenedDocumentStates.CorruptedAndModified)
                {
                    await NotyfyReparablePdfDocAsync();
                }
            }

            UpdateTabDependantProperties();

            if (_editedSinceThumbnailViewerOpened)
            {
                _editedSinceThumbnailViewerOpened = false;
                PDFViewCtrl.UpdatePageLayout();
                IsDocumentModifiedSinceLastSave = true;
                ResolveThumbnailSlider();
                UpdateUIVisiblity();
            }

            if (_hasPage1ChangeSinceThumbnailViewOpened)
            {
                _hasPage1ChangeSinceThumbnailViewOpened = false;
                TabControlViewModel.SelectedTab.UpdatePreview();
            }

            if (CurrentDocumentState == OpenedDocumentStates.Corrupted)
                CurrentDocumentState = OpenedDocumentStates.CorruptedAndModified;

            IsDocumentModifiedSinceLastSave = true;

            _documentModified?.Invoke();
        }

        public async Task SaveImplAsync()
        {
            try
            {
                _savingInProgress = true;
                await SaveHelperAsync(null, true, true);
                ResolveAutoSaverAndSaveButton();
            }
            catch (Exception e)
            {
                Debug.WriteLine(e.ToString());
            }
            finally
            {
                _savingInProgress = false;
                if (PDFViewCtrl != null)
                    PDFViewCtrl.IsEnabled = true;

                IsSaveTimerOpen = false;
            }
        }

        public async Task SaveHelperAsync() => await SaveHelperAsync(null);

        /// <summary>
        /// It starts the saving process on the selected document/tab
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="includeFullSave"></param>
        /// <param name="alwaysSave">Means that we want to save to the original file, not just to the temporary one</param>
        /// <returns></returns>
        public async Task SaveHelperAsync(PdfViewCtrlTabInfo tab, bool includeFullSave = false, bool alwaysSave = false)
        {
            if (tab == null)
                tab = TabControlViewModel.SelectedTab;

            if (tab == null)
                return;

            if (OutlineDialogViewModel?.HasUnsavedUserbookmarks == true)
            {
                OutlineDialogViewModel.SaveBookmarks();
                await OutlineDialogViewModel.UserBookmarks.WaitForBookmarkSavingAsync();
            }

            SaveResult result = await SaveToOriginalAsync(tab, includeFullSave, alwaysSave).ConfigureAwait(true);
            if (result != SaveResult.e_normal)
            {
                IsSaveTimerOpen = false;
                string messageString = string.Format(_loader.GetString("ViewerPage_TabDialog_SaveError_AccessDenied"), tab.Title);
                if (result == SaveResult.e_unknown_error)
                {
                    messageString = string.Format(_loader.GetString("ViewerPage_TabDialog_SaveError"), tab.Title);
                }

                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("ViewerPage_TabDialog_SaveError_Title"),
                    Content = messageString,
                    CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
            }
        }

        private async Task<SaveResult> SaveToOriginalAsync(PdfViewCtrlTabInfo tab, bool includeFullSave = false, bool alwaysSave = false)
        {
            if (tab == null)
                tab = TabControlViewModel.SelectedTab;

            if (tab == null)
                return SaveResult.e_normal;

            SaveResult result = SaveResult.e_normal;
            try
            {
                if (tab.DocumentState != OpenedDocumentStates.Corrupted && tab.DocumentState != OpenedDocumentStates.CorruptedAndModified)
                {
                    result = SaveResult.e_unknown_error;

                    if (tab.Doc != null)
                    {
                        await tab.SaveDocumentImplAsync();
                    }

                    if ((tab.DocumentState == OpenedDocumentStates.Normal && (ViewerSettings.AutoSaveOn || alwaysSave))
                        || (tab.DocumentState == OpenedDocumentStates.NeedsFullSave && includeFullSave))
                    {
                        await DocumentManager.Instance.AddChangesToOriginalAsync(tab.OriginalFile, tab.PdfFile);
                        Tuple<DateTimeOffset, ulong> lastModifiedAndSize = await UtilityFunctions.GetDateModifiedAndSizeAsync(tab.OriginalFile);
                        tab.IsDocumentModifiedSinceLastSave = false;
                        tab.MetaData.LastModifiedDate = lastModifiedAndSize.Item1;
                        tab.MetaData.FileSize = lastModifiedAndSize.Item2;
                        tab.SystemLastModifiedDate = lastModifiedAndSize.Item1;
                        tab.SystemFileSize = lastModifiedAndSize.Item2;
                        TabControlViewModel.SaveTabState();
                    }

                    result = SaveResult.e_normal;
                }
            }
            catch (UnauthorizedAccessException)
            {
                result = SaveResult.e_access_denied;
            }
            catch (Exception)
            {
                result = SaveResult.e_unknown_error;
            }

            return result;
        }

        private async Task SaveTempFileOnlyAsync(PdfViewCtrlTabInfo tab) => await tab.SaveDocumentImplAsync();

        public async Task SaveAsAsync(PdfViewCtrlTabInfo tab, bool closeIfsuccessful, bool navigateIfSucessful, SaveAsOption option)
        {
            if (tab == null || _filePickerOpen)
                return;

            StorageFile file = null;

            FileSavePicker fileSavePicker = new FileSavePicker
            {
                CommitButtonText = "Save"
            };

            fileSavePicker.FileTypeChoices.Add("PDF Document", new List<string> { ".pdf" });
            string suggestedFileName = string.Empty;
            if (tab.OriginalFile != null)
            {
                suggestedFileName = tab.OriginalFile.DisplayName;
            }
            else
            {
                if (tab.PdfFile == null)
                    return;

                suggestedFileName = tab.PdfFile.DisplayName;
            }
            if (tab.DocumentState == OpenedDocumentStates.Normal
                || tab.DocumentState == OpenedDocumentStates.NeedsFullSave
                || tab.DocumentState == OpenedDocumentStates.Corrupted
                || tab.DocumentState == OpenedDocumentStates.CorruptedAndModified)
            {
                string suffix = "";
                switch (option)
                {
                    case SaveAsOption.Flatten:
                        suffix = " - Flattened";
                        break;
                    case SaveAsOption.Optimize:
                        suffix = " - Reduced";
                        break;
                    case SaveAsOption.Password:
                        if (!string.IsNullOrEmpty(PasswordFileViewModel.CurrentPassword))
                        {
                            suffix = " - Protected";
                        }
                        else
                        {
                            suffix = " - Not_Protected";
                        }
                        break;
                    case SaveAsOption.Crop:
                        suffix = " - Cropped";
                        break;
                    default:
                        suffix = " - Copy";
                        break;
                }

                suggestedFileName += suffix;
            }

            fileSavePicker.SuggestedFileName = suggestedFileName;

            try
            {
                _filePickerOpen = true;
                file = await fileSavePicker.PickSaveFileAsync();
                SaveNeededStates saveState = CheckSavingRequirements(TabControlViewModel.SelectedTab);
                if (saveState == SaveNeededStates.saveNeeded)
                    await SaveToOriginalAsync(TabControlViewModel.SelectedTab, false, true);
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _filePickerOpen = false;
            }

            if (file != null)
            {
                // check if a user selects an already open tab
                PdfViewCtrlTabInfo matchedTab = TabControlViewModel.Tabs.FirstOrDefault(x => x.OriginalFile?.Path == file.Path);
                if (matchedTab != null)
                {
                    // If selecting the same open tab, just save
                    if (matchedTab.IsSelected)
                    {
                        ContentDialog dialog = new ContentDialog();

                        switch (option)
                        {
                            case SaveAsOption.Flatten:
                                dialog.Title = _loader.GetString("ViewerPage_TabConflictDialog_SaveAsFlatten_Title");
                                dialog.Content = string.Format(_loader.GetString("ViewerPage_TabConflictDialog_SaveAsFlatten"), matchedTab.Title);

                                break;
                            case SaveAsOption.Optimize:
                                dialog.Title = _loader.GetString("ViewerPage_TabConflictDialog_SaveAsOptimize_Title");
                                dialog.Content = string.Format(_loader.GetString("ViewerPage_TabConflictDialog_SaveAsOptimize"), matchedTab.Title);

                                break;
                            case SaveAsOption.Password:
                                dialog.Title = _loader.GetString("ViewerPage_TabConflictDialog_SaveAsPassword_Title");
                                dialog.Content = string.Format(_loader.GetString("ViewerPage_TabConflictDialog_SaveAsPassword"), matchedTab.Title);

                                break;
                            case SaveAsOption.Crop:
                                dialog.Title = _loader.GetString("ViewerPage_TabConflictDialog_SaveAsCrop_Title");
                                dialog.Content = string.Format(_loader.GetString("ViewerPage_TabConflictDialog_SaveAsCrop"), matchedTab.Title);

                                break;
                            default:
                                await SaveImplAsync();
                                break;
                        }

                        if (option != SaveAsOption.Save)
                        {
                            await DialogHelper.ShowContentDialogAsync(dialog);
                            IsModal = false;
                        }
                    }
                    // If selecting an open, non-selected tab, close that tab, and continue
                    else
                    {
                        await CloseTabAsync(matchedTab);
                        await ContinueSavingAsync(file, tab, closeIfsuccessful, navigateIfSucessful, option);
                    }
                }
                else
                {
                    await ContinueSavingAsync(file, tab, closeIfsuccessful, navigateIfSucessful, option);
                }
            }
            else
            {
                if (SaveAsDone != null)
                {
                    var tcs = new TaskCompletionSource<bool>();
                    using (var deferral = new Deferral(() => tcs.SetResult(true)))
                    {
                        SaveAsDone.Invoke(true, false, deferral);
                        await tcs.Task;
                    }
                }
            }

            if (file == null)
                IsModal = false;
        }

        private IAsyncAction FlattenAsync(PDFDoc doc)
        {
            Task t = new Task(() => doc.FlattenAnnotations());
            t.Start();

            return t.AsAsyncAction();
        }

        private IAsyncAction OptimizeAsync(PDFDoc doc)
        {
            Task t = new Task(() => OptimizeFileViewModel.Optimize(doc));
            t.Start();

            return t.AsAsyncAction();
        }

        private IAsyncOperation<Tuple<bool, string>> ApplyPasswordAsync(PDFDoc doc)
        {
            Task<Tuple<bool, string>> t = new Task<Tuple<bool, string>>(() => PasswordFileViewModel.ApplyPassword(doc));
            t.Start();

            return t.AsAsyncOperation();
        }

        private IAsyncAction ApplyCropAsync(PDFDoc doc)
        {
            Task t = new Task(() => ApplyCrop(doc));
            t.Start();

            return t.AsAsyncAction();
        }

        private bool GetUserCropBoxes()
        {
            _userCropRects.Clear();
            PDFDoc pdfDoc = PDFViewCtrl.GetDoc();

            try
            {
                pdfDoc.LockRead();
                PageIterator iter = pdfDoc.GetPageIterator();
                int pageNum = 1;

                while (iter.HasNext())
                {
                    pdftron.PDF.Page page = iter.Current();

                    pdftron.PDF.Rect userCropBox = page.GetBox(PageBox.e_user_crop);
                    pdftron.PDF.Rect cropBox = page.GetCropBox();

                    if (!CropViewViewModel.AreRectsSimilar(userCropBox, cropBox, CropViewViewModel.RECT_COMPARE_THRESHOLD))
                        _userCropRects[pageNum] = page.GetBox(PageBox.e_user_crop);

                    pageNum++;
                    iter.Next();
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                    throw;
            }
            finally
            {
                pdfDoc.UnlockRead();
            }

            return _userCropRects.Count != 0;
        }

        private void ApplyCrop(PDFDoc doc)
        {
            try
            {
                doc.LockRead();
                PageIterator iter = doc.GetPageIterator();
                int pageNum = 1;

                while (iter.HasNext())
                {
                    pdftron.PDF.Page page = iter.Current();

                    if (_userCropRects.ContainsKey(pageNum))
                        page.SetBox(PageBox.e_crop, _userCropRects[pageNum]);

                    pageNum++;
                    iter.Next();
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                    throw;
            }
            finally
            {
                doc.UnlockRead();
            }
        }

        public async Task ContinueSavingAsync(StorageFile file, PdfViewCtrlTabInfo tab, bool closeIfsuccessful, bool navigateIfSucessful, SaveAsOption option, bool destinationIsTempFile = false)
        {
            if (file != null)
            {
                bool success = false;
                bool passwordError = false;
                string errorMessage = null;
                string oldFilePath = file.Path;

                string path = file.Path;
                if (string.IsNullOrEmpty(path))
                    path = file.Name;

                // The old File record is no longer needed, so we decrement it and create a new one here.
                bool isBox = UtilityFunctions.IsBox(file.Path);
                bool isDropBox = UtilityFunctions.IsDropBox(file.Path);
                bool isOneDrive = UtilityFunctions.IsOneDrive(file.Path);
                //bool isRemote = isBox || isDropBox || isOneDrive;
                //bool useRWStream = Utilities.UtilityFunctions.IsBox(path) || Utilities.UtilityFunctions.IsDropBox(path);

                NewDocumentProperties docProps = null;
                bool createNewTab = option != SaveAsOption.Save;

                try
                {
                    CurrentSaveAsOption = option;

                    IsSaveTimerOpen = true;
                    IsModal = true;

                    List<UserBookmarksViewModel.UserBookmarkItem> bookmarkItems = null;
                    string documentTag = string.Empty;
                    if (tab.OriginalFile != null)
                    {
                        documentTag = UtilityFunctions.GetMaximalAvailablePath(tab.OriginalFile);

                        if (!isDropBox)
                            bookmarkItems = await UserBookmarkManager.GetBookmarkListAsync(documentTag);
                    }

                    bool generateNewFileName = false;
                    if (oldFilePath == file.Path)
                        generateNewFileName = true;

                    StorageFile temporaryFile = await DocumentManager.Instance.OpenTemporaryCopyAsync(file, generateNewFileName);

                    if (tab.Doc != null)
                    {
                        await tab.SaveDocumentImplAsync();

                        PDFDoc saveDoc = tab.Doc;
                        if (createNewTab)
                        {
                            // Replace the temporary file created with a copy of the original one: tab.PDFFile
                            await tab.PdfFile.CopyAndReplaceAsync(temporaryFile);

                            docProps = new NewDocumentProperties
                            {
                                Password = tab.MetaData.Password,
                                StartPage = tab.PDFViewCtrl.GetCurrentPage(),
                                Zoom = tab.PDFViewCtrl.GetZoom(),
                                HorizontalScrollPosition = tab.PDFViewCtrl.GetHScrollPos(),
                                VerticalScrollPosition = tab.PDFViewCtrl.GetVScrollPos(),
                                PageRotation = tab.PDFViewCtrl.GetRotation(),
                                PresentationMode = tab.PDFViewCtrl.GetPagePresentationMode(),
                                File = file,
                                TemporaryFile = temporaryFile,
                                OpenedDocumentState = OpenedDocumentStates.Normal
                            };

                            if (UtilityFunctions.IsBox(file.Path) || UtilityFunctions.IsDropBox(file.Path) || destinationIsTempFile)
                                docProps.OpenedDocumentState = OpenedDocumentStates.ReadOnly;

                            docProps.Doc = new PDFDoc(docProps.TemporaryFile);
                            if (!docProps.Doc.InitSecurityHandler())
                                docProps.Doc.InitStdSecurityHandler(docProps.Password);

                            saveDoc = docProps.Doc;
                        }
                        else
                        {
                            await tab.Doc.SaveAsync(temporaryFile, SDFDocSaveOptions.e_incremental);
                        }

                        switch (option)
                        {
                            case SaveAsOption.Flatten:
                                await FlattenAsync(saveDoc);
                                break;
                            case SaveAsOption.Optimize:
                                await OptimizeAsync(saveDoc);
                                OptimizeFileViewModel.IsPopupOpen = false;
                                break;
                            case SaveAsOption.Password:
                                Tuple<bool, string> passwordResult = await ApplyPasswordAsync(saveDoc);
                                passwordError = !passwordResult.Item1;
                                if (!passwordError)
                                    docProps.Password = passwordResult.Item2;

                                PasswordFileViewModel.IsPopupOpen = false;
                                break;
                            case SaveAsOption.Crop:
                                await ApplyCropAsync(saveDoc);
                                break;
                        }

                        if (createNewTab)
                        {
                            await docProps.Doc.SaveAsync(SDFDocSaveOptions.e_remove_unused);

                            // Copy all changes from the temporary file to the output original one (flattened, optimized)
                            await DocumentManager.Instance.AddChangesToOriginalAsync(docProps.File, docProps.TemporaryFile, UtilityFunctions.IsDropBox(file.Path));

                            Tuple<DateTimeOffset, ulong> lastModifiedAndSize = await UtilityFunctions.GetDateModifiedAndSizeAsync(file);
                            docProps.LastModifiedDate = lastModifiedAndSize.Item1;
                            docProps.Filesize = lastModifiedAndSize.Item2;
                        }
                    }
                    else
                    {
                        await tab.PdfFile.CopyAndReplaceAsync(temporaryFile);
                    }

                    success = !passwordError;

                    if (!createNewTab && success)
                    {
                        await DocumentManager.Instance.AddChangesToOriginalAsync(file, temporaryFile, UtilityFunctions.IsDropBox(file.Path));

                        Tuple<DateTimeOffset, ulong> lastModifiedAndSize = await UtilityFunctions.GetDateModifiedAndSizeAsync(file);
                        tab.MetaData.LastModifiedDate = lastModifiedAndSize.Item1;
                        tab.MetaData.FileSize = lastModifiedAndSize.Item2;
                        tab.SystemLastModifiedDate = lastModifiedAndSize.Item1;
                        tab.SystemFileSize = lastModifiedAndSize.Item2;

                        if (bookmarkItems?.Count > 0)
                            await UserBookmarkManager.SaveBookmarkListAsync(documentTag, bookmarkItems);

                        if (tab.OriginalFile != null)
                            DocumentManager.Instance.CloseFile(tab.OriginalFile);

                        tab.OriginalFile = file;
                        tab.PdfFile = temporaryFile;
                        tab.Title = file.DisplayName;
                        tab.MetaData.OriginalFileToken = StorageApplicationPermissions.FutureAccessList.Add(tab.OriginalFile);
                        tab.MetaData.FutureAccessListToken = StorageApplicationPermissions.FutureAccessList.Add(tab.PdfFile);

                        tab.DocumentState = OpenedDocumentStates.Normal;
                        if (UtilityFunctions.IsBox(file.Path) || UtilityFunctions.IsDropBox(file.Path) || destinationIsTempFile)
                            tab.DocumentState = OpenedDocumentStates.ReadOnly;

                        bool isEncrypted = !string.IsNullOrEmpty(tab.MetaData.Password);
                        _belongInRecentList?.Invoke(tab, isEncrypted, temporaryFile);
                    }
                }
                catch (Exception ex)
                {
                    errorMessage = AnalyticsHandlerBase.GetExceptionMessage(ex);
                    AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, errorMessage);
                    success = false;
                }
                finally
                {
                    IsSaveTimerOpen = false;
                    IsModal = false;
                }

                if (success)
                {
                    if (SaveAsDone != null)
                    {
                        var tcs = new TaskCompletionSource<bool>();
                        using (var deferral = new Deferral(() => tcs.SetResult(true)))
                        {
                            SaveAsDone.Invoke(false, false, deferral);
                            await tcs.Task;
                        }
                    }

                    if (closeIfsuccessful)
                    {
                        await CloseTabAsync(tab);
                    }
                    else
                    {
                        if (createNewTab)
                        {
                            await ActivateWithFileAsync(file, docProps);
                        }
                        else
                        {
                            tab.IsDocumentModifiedSinceLastSave = false;
                            if (_autoSaveHelper != null)
                            {
                                _autoSaveHelper.CurrentFile = tab.OriginalFile;
                                _autoSaveHelper.TemporaryFile = tab.PdfFile;
                            }
                        }
                    }

                    if (navigateIfSucessful)
                    {
                        await NavigateAwayAsync();
                        return;
                    }

                    if (!closeIfsuccessful && tab.Doc == null)
                    {
                        await ActivateTabAsync(tab);
                        await tab.ActivateAsync();
                    }

                    ResolveAutoSaverAndSaveButton();

                    SharingHelper.DocumentSaver = SaveHelperAsync;

                    // TODO Phone
                    //CloseAppIfApplicable();

                }
                else
                {
                    if (passwordError)
                    {
                        ContentDialog dialog = new ContentDialog
                        {
                            Title = _loader.GetString("ViewerPage_SaveAsFailed_Title"),
                            Content = _loader.GetString("ViewerPage_SaveWithPasswordFailed_Content"),
                            CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                    }
                    else
                    {
                        if (string.IsNullOrEmpty(errorMessage))
                            errorMessage = string.Empty;

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = _loader.GetString("ViewerPage_SaveAsFailed_Title"),
                            Content = string.Format(_loader.GetString("ViewerPage_SaveAsFailed_Content"), ViewerSettings.SupportName, Environment.NewLine + errorMessage),
                            CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                        };

                        await DialogHelper.ShowContentDialogAsync(dialog);
                    }

                    if (SaveAsDone != null)
                    {
                        var tcs = new TaskCompletionSource<bool>();
                        using (var deferral = new Deferral(() => tcs.SetResult(true)))
                        {
                            SaveAsDone.Invoke(false, true, deferral);
                            await tcs.Task;
                        }
                    }
                }
            }
        }

        public async Task SuspendAsync()
        {
            try
            {
                _autoSaveResultFromSuspend = SaveResult.e_normal;
                if (PDFViewCtrl?.GetDoc() != null)
                {
                    SaveNeededStates saveState = CheckSavingRequirements(TabControlViewModel.SelectedTab);
                    if (saveState == SaveNeededStates.saveNeeded)
                        _autoSaveResultFromSuspend = await SaveToOriginalAsync(TabControlViewModel.SelectedTab, false, true);
                }

                TabControlViewModel.SaveTabState();
                if (CropPopupViewModel?.IsAutomaticCropping == true)
                    CropPopupViewModel.CancelAutomaticCommand.Execute(null);

                await DeactivateTabImplAsync(TabControlViewModel.SelectedTab);
            }
            catch (Exception ex)
            {
                string error;
                PDFNetException pdfEx = new PDFNetException(ex.HResult);
                if (pdfEx.IsPDFNetException)
                {
                    error = pdfEx.ToString();
                }
                else
                {
                    error = ex.ToString();
                }

                Debug.WriteLine(error);
            }
        }

        private SemaphoreSlim ResumeSemaphore => _resumeSemaphore ?? (_resumeSemaphore = new SemaphoreSlim(1));

        public async Task ResumeAsync()
        {
            await ResumeSemaphore.WaitAsync();
            try
            {
                if (TabControlViewModel != null)
                {
                    IsModal = true;
                    await TabControlViewModel.UpdateFilePropertiesAsync();
                    IsModal = false;
                }
            }
            finally
            {
                ResumeSemaphore.Release();
            }

            if (_fileToActivate == null && TabControlViewModel?.SelectedTab != null)
            {
                if (!TabControlViewModel.SelectedTab.IsUpToDate && TabControlViewModel.SelectedTab.IsDocumentModifiedSinceLastSave)
                {
                    await ShowConflictingTabOptionsAsync(TabControlViewModel.SelectedTab, null);
                }
                else
                {
                    if (!TabControlViewModel.SelectedTab.IsUpToDate)
                        await TabControlViewModel.SelectedTab.UpdateFileAsync();

                    _previousTab = TabControlViewModel.SelectedTab;
                    await ActivateTabAsync(TabControlViewModel.SelectedTab);
                    if (_autoSaveResultFromSuspend != SaveResult.e_normal)
                    {
                        try
                        {
                            _showingAutoSaveResumeMessage = true;
                            string message = string.Format(
                                _loader.GetString("ViewerPage_AutoSaveFailed_Suspend_Message_AccessDenied"),
                                TabControlViewModel.SelectedTab.Title);
                            if (_autoSaveResultFromSuspend == SaveResult.e_unknown_error)
                            {
                                message = string.Format(
                                    _loader.GetString("ViewerPage_AutoSaveFailed_Suspend_Message_unknown"),
                                    TabControlViewModel.SelectedTab.Title);
                            }

                            ContentDialog dialog = new ContentDialog
                            {
                                Title = _loader.GetString("ViewerPage_AutoSaveFailed_Title"),
                                Content = message,
                                CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                            };

                            await DialogHelper.ShowContentDialogAsync(dialog);
                        }
                        finally
                        {
                            _showingAutoSaveResumeMessage = false;
                        }
                    }
                }
            }
        }

        public OptimizeFileViewModel OptimizeFileViewModel
        {
            get => _optimizeFileViewModel;
            set => Set(ref _optimizeFileViewModel, value);
        }

        public PasswordFileViewModel PasswordFileViewModel
        {
            get => _passwordFileViewModel;
            set => Set(ref _passwordFileViewModel, value);
        }

        public CropPopupViewModel CropPopupViewModel
        {
            get => _cropPopupViewModel;
            set => Set(ref _cropPopupViewModel, value);
        }

        public ViewerPageSettingsViewModel ViewerPageSettingsViewModel
        {
            get => _viewerPageSettingsViewModel;
            set
            {
                ViewerPageSettingsViewModel old = _viewerPageSettingsViewModel;
                if (Set(ref _viewerPageSettingsViewModel, value) && old != null)
                    old.CleanUp();
            }
        }

        public CustomColorViewModel CustomColorViewModel
        {
            get => _customColorViewModel;
            set => Set(ref _customColorViewModel, value);
        }

        public string CurrentPagePresentationModeSymbol
        {
            get
            {
                if (PDFViewCtrl != null)
                {
                    switch (PDFViewCtrl.GetPagePresentationMode())
                    {
                        case PDFViewCtrlPagePresentationMode.e_single_continuous:
                            return new string((char)0x0052, 1);
                        case PDFViewCtrlPagePresentationMode.e_single_page:
                            return new string((char)0x0061, 1);
                        case PDFViewCtrlPagePresentationMode.e_facing:
                            return new string((char)0x0060, 1);
                        case PDFViewCtrlPagePresentationMode.e_facing_cover:
                            return new string((char)0x0055, 1);
                        case PDFViewCtrlPagePresentationMode.e_facing_continuous_cover:
                            return new string((char)0x0056, 1);
                        case PDFViewCtrlPagePresentationMode.e_facing_continuous:
                            return new string((char)0x0054, 1);
                    }
                }

                return new string((char)0x0061, 1);
            }
        }

        public Thickness CurrentPagePresentationModeSymbolMargin
        {
            get
            {
                if (PDFViewCtrl != null)
                {
                    switch (PDFViewCtrl.GetPagePresentationMode())
                    {
                        case PDFViewCtrlPagePresentationMode.e_facing_cover:
                            return new Thickness(7, 0, 0, 4);
                        case PDFViewCtrlPagePresentationMode.e_facing:
                            return new Thickness(0, 0, 0, 4);
                    }
                }

                return new Thickness(0, 0, 0, 2);
            }
        }

        public double CurrentPagePresentationModeFontSize
        {
            get
            {
                if (PDFViewCtrl != null)
                {
                    switch (PDFViewCtrl.GetPagePresentationMode())
                    {
                        case PDFViewCtrlPagePresentationMode.e_single_continuous:
                        case PDFViewCtrlPagePresentationMode.e_single_page:
                            return 25;
                        case PDFViewCtrlPagePresentationMode.e_facing:
                        case PDFViewCtrlPagePresentationMode.e_facing_continuous:
                        case PDFViewCtrlPagePresentationMode.e_facing_cover:
                        case PDFViewCtrlPagePresentationMode.e_facing_continuous_cover:
                            return 22;
                    }
                }

                return 22;
            }
        }

        private void ViewerPageSettingsViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("PagePresentationMode", StringComparison.OrdinalIgnoreCase))
            {
                IsAppBarOpen = false;
                IsQuickSettingsOpen = false;
                RaisePropertyChanged(nameof(IsQuickSettingsOpen));
                RaisePropertyChanged(nameof(CurrentPagePresentationModeSymbol));
                RaisePropertyChanged(nameof(CurrentPagePresentationModeSymbolMargin));
                RaisePropertyChanged(nameof(CurrentPagePresentationModeFontSize));
            }

            if (e.PropertyName.Equals("ColorMode", StringComparison.OrdinalIgnoreCase))
            {
                IsAppBarOpen = false;
                IsQuickSettingsOpen = false;
                RaisePropertyChanged(nameof(IsQuickSettingsOpen));
                if (ViewerPageSettingsViewModel != null && ViewerPageSettingsViewModel.ColorMode != CustomColorModeType.Custom)
                    ResolveColorMode(ViewerPageSettingsViewModel.ColorMode);
            }

            if (e.PropertyName.Equals("CustomColor", StringComparison.OrdinalIgnoreCase))
            {
                if (CustomColorViewModel == null)
                {
                    CustomColorViewModel = new CustomColorViewModel();
                    CustomColorViewModel.CustomColorRequested += CustomColorViewModel_CustomColorRequested;
                    CustomColorViewModel.PopupClosed += PopupClosed;
                }

                IsAppBarOpen = false;
                IsQuickSettingsOpen = false;
                RaisePropertyChanged(nameof(IsQuickSettingsOpen));
                IsModal = true;
                CustomColorViewModel.IsPopupOpen = true;
            }

            if (e.PropertyName.Equals("FullScreen", StringComparison.OrdinalIgnoreCase))
                RaisePropertyChanged(nameof(IsFullScreen));

            if (e.PropertyName.Equals("IsReflow", StringComparison.OrdinalIgnoreCase))
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Reflow" });

                IsAppBarOpen = false;
                IsQuickSettingsOpen = false;
                RaisePropertyChanged(nameof(IsDocumentEditable));
                RaisePropertyChanged(nameof(IsDocumentSearchable));
                ResolveReflow();
            }

            // TODO Phone
            //IsViewModeDialogOpen = false;
        }

        private void ViewerPageSettingsViewModel_ViewRequestedHandler(RequestableViewType view)
        {
            if (view == RequestableViewType.Thumbnails)
            {
                IsThumbnailsViewOpen = true;
                RaisePropertyChanged(nameof(IsThumbnailsViewOpen));
            }

            if (view == RequestableViewType.Crop)
            {
                IsAppBarOpen = false;
                IsQuickSettingsOpen = false;
                IsModal = true;
                CropPopupViewModel.IsPopupOpen = true;
            }

            // TODO Phone
            //IsViewModeDialogOpen = false;
        }

        private void CustomColorViewModel_CustomColorRequested(CustomColorIconViewModel icon)
        {
            ViewerPageSettingsViewModel.ColorMode = CustomColorModeType.Custom;
            ResolveColorMode(CustomColorModeType.Custom);
        }

        private void PopupClosed() => IsModal = false;

        private async void OptimizeFileViewModel_OptimizePopupOptimize()
        {
            IsAppBarOpen = false;
            IsModal = true;
            await SaveAsAsync(TabControlViewModel.SelectedTab, false, false, SaveAsOption.Optimize);
        }

        private async void PasswordFileViewModel_PasswordConfirmed()
        {
            IsAppBarOpen = false;
            IsModal = true;
            await SaveAsAsync(TabControlViewModel.SelectedTab, false, false, SaveAsOption.Password);
        }

        private void CropPopupViewModel_ManualCropRequested()
        {
            IsCropViewOpen = true;
            RaisePropertyChanged(nameof(IsCropViewOpen));
        }

        private async void CropPopupViewModel_DocumentEdited()
        {
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            await HandleDocumentEditingAsync();
        }

        public PdfCompareControl PdfCompareViewer
        {
            get { return _PdfCompareViewer; }
            set
            {
                if (_PdfCompareViewer != value)
                {
                    _PdfCompareViewer = value;
                    RaisePropertyChanged();
                }
            }
        }

        public bool IsPdfCompareViewOpen
        {
            get { return _isPdfCompareViewOpen; }
            set 
            {
                if (_isPdfCompareViewOpen != value)
                {
                    _isPdfCompareViewOpen = value;
                    IsEntireAppBarOpen = !_isPdfCompareViewOpen;

                    if (IsPdfCompareViewOpen)
                    {
                        IsOutlineDialogOpen = _outlineDialogHibernation;

                        PdfCompareViewer = new PdfCompareControl(PDFViewCtrl);
                        PdfCompareViewer.ControlClosed += PdfCompareViewer_ControlClosed;

                        IsAppBarOpen = false;
                    }
                    else
                    {
                        _outlineDialogHibernation = IsOutlineDialogOpen;
                        IsOutlineDialogOpen = false;

                        if (PdfCompareViewer != null)
                        {
                            PdfCompareViewer.ControlClosed -= PdfCompareViewer_ControlClosed;
                            PdfCompareViewer.ViewModel?.CleanUp();
                            PdfCompareViewer = null;
                        }
                    }

                    IsModal = value;
                    mPdfCompareState?.Invoke(_isPdfCompareViewOpen);
                    IsAllowDropEnabled = !_isPdfCompareViewOpen; // PDFCompare has it's own drag n drop so disable ViewerControl one
                    RaisePropertyChanged();
                }
            }
        }

        public ThumbnailViewer ThumbnailViewer
        {
            get => _thumbnailViewer;
            set
            {
                if (value != _thumbnailViewer)
                {
                    if (_thumbnailViewer != null)
                    {
                        _thumbnailViewer.ViewModel.CleanUp();
                        if (_thumbnailViewerPageMovedDelegate != null)
                        {
                            _thumbnailViewer.ViewModel.PageMoved -= _thumbnailViewerPageMovedDelegate;
                            _thumbnailViewer.ViewModel.PageDeleted -= _thumbnailViewerPageDeletedDelegate;
                            _thumbnailViewer.ViewModel.PageAdded -= _thumbnailViewerPageAddedDelegate;
                            _thumbnailViewer.ViewModel.PageRotated -= _thumbnailViewerPageRotatedDelegate;
                            _thumbnailViewerPageMovedDelegate = null;
                            _thumbnailViewerPageDeletedDelegate = null;
                        }

                        _thumbnailViewer.PDFViewCtrl = null;
                    }

                    _thumbnailViewer = value;
                    if (_thumbnailViewer != null)
                    {
                        _thumbnailViewer.ViewModel.ReadOnly = CurrentDocumentState == OpenedDocumentStates.Uneditable;
                        _thumbnailViewerPageMovedDelegate = new ThumbnailsViewViewModel.PageMovedDelegate(ThumbnailViewerPageMovedHandler);
                        _thumbnailViewerPageDeletedDelegate = new ThumbnailsViewViewModel.PageDeletedDelegate(ThumbnailViewerPageDeletedHandler);
                        _thumbnailViewerPageAddedDelegate = new ThumbnailsViewViewModel.PageAddedDelegate(ThumbnailViewerPageAddedHandler);
                        _thumbnailViewerPageRotatedDelegate = new ThumbnailsViewViewModel.PageRotatedDelegate(ThumbnailViewerPageRotatedDelegate);
                        _thumbnailViewer.ViewModel.PageMoved += _thumbnailViewerPageMovedDelegate;
                        _thumbnailViewer.ViewModel.PageDeleted += _thumbnailViewerPageDeletedDelegate;
                        _thumbnailViewer.ViewModel.PageAdded += _thumbnailViewerPageAddedDelegate;
                        _thumbnailViewer.ViewModel.PageRotated += _thumbnailViewerPageRotatedDelegate;
                    }

                    RaisePropertyChanged();
                }
            }
        }

        public bool IsThumbnailsViewOpen
        {
            get => _isThumbnailsViewOpen;
            set
            {
                if (value != _isThumbnailsViewOpen)
                {
                    if (value == true)
                        Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Viewer", new Dictionary<string, string>() { ["Command"] = "Open Thumbnails View" });

                    _isThumbnailsViewOpen = value;
                    IsEntireAppBarOpen = !_isThumbnailsViewOpen;

                    if (!_isThumbnailsViewOpen)
                    {
                        IsOutlineDialogOpen = _outlineDialogHibernation;

                        _thumbnailViewer?.CloseControl();

                        ThumbnailViewer = null;
                        if (_editedSinceThumbnailViewerOpened)
                        {
                            /* NOTE: running it synchronous with discard for now as it does not adds a noticeable performance issue
                             * an alternative would be to use AwaitableRunAsync from Microsoft.Toolkit.Uwp.Helpers
                             * CoreApplication.MainView.CoreWindow.Dispatcher.AwaitableRunAsync(async () => await HandleDocumentEditingAsync());
                             */
                            _ = HandleDocumentEditingAsync();
                        }
                    }
                    else
                    {
                        _outlineDialogHibernation = IsOutlineDialogOpen;
                        IsOutlineDialogOpen = false;

                        _editedSinceThumbnailViewerOpened = false;
                        PDFViewCtrl.CancelRendering();
                        ToolManager?.CreateTool(ToolType.e_pan, ToolManager.CurrentTool);

                        _navigationStateBeforeThumbnailsviewOpening = NavigationState.TakeSnapshot(PDFViewCtrl);
                        ThumbnailViewer = new ThumbnailViewer(PDFViewCtrl, UtilityFunctions.GetMaximalAvailablePath(CurrentFile))
                        {
                            ToolManager = TabControlViewModel.SelectedTab.ToolManager
                        };

                        if (ViewerSettings.ColorMode == CustomColorModeType.Night)
                        {
                            ThumbnailViewer.Background = new SolidColorBrush(Color.FromArgb(255, 44, 44, 44));
                            ThumbnailViewer.BlankPageDefaultColor = Colors.Black;
                        }
                        else if (ViewerSettings.ColorMode == CustomColorModeType.Sepia)
                        {
                            ThumbnailViewer.BlankPageDefaultColor = Color.FromArgb(255, 255, 232, 206);
                        }

                        ThumbnailViewer.ControlClosed += ThumbnailViewer_ControlClosed;
                        IsAppBarOpen = false;
                    }

                    IsModal = value;
                    RaisePropertyChanged();
                }
            }
        }

        public CropView CropView
        {
            get => _cropView;
            set => Set(ref _cropView, value);
        }

        public bool IsCropViewOpen
        {
            get => _isCropViewOpen;
            set
            {
                if (value != _isCropViewOpen)
                {
                    _isCropViewOpen = value;
                    IsEntireAppBarOpen = !_isCropViewOpen;

                    if (!_isCropViewOpen)
                    {
                        _cropView?.CloseControl();
                        _cropView = null;
                        PDFViewCtrl.UpdatePageLayout();
                    }
                    else
                    {
                        PDFViewCtrl.CancelRendering();
                        ToolManager?.CreateTool(ToolType.e_pan, ToolManager.CurrentTool);
                        CropView = new CropView(PDFViewCtrl);
                        if (ViewerSettings.ColorMode == CustomColorModeType.Sepia)
                        {
                            CropView.SetColorPostProcessing(
                                PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map,
                                Color.FromArgb(255, 255, 232, 206),
                                Colors.Black);
                        }
                        else
                        {
                            CropView.SetColorPostProcessing(PDFViewCtrl.GetColorPostProcessMode(), null, null);
                        }

                        CropView.ControlClosed += CropView_ControlClosed;
                        CropView.ViewModel.DocumentEdited += CropViewModel_DocumentEdited;
                        IsAppBarOpen = false;
                    }

                    IsModal = value;
                    RaisePropertyChanged();
                }
            }
        }

        public AudioControl SoundControl
        {
            get => _audioControl;
            set => Set(ref _audioControl, value);
        }

        private async void PdfCompareViewer_ControlClosed()
        {
            IsPdfCompareViewOpen = false;
            
            // Ensure to notify the doc has changed
            if (PDFViewCtrl.GetDoc().IsModified())
            {
                await HandleDocumentEditingAsync();
            }

            PDFViewCtrl.Update(true); // Ensure to draw any updates done at PDFCompare
        }

        private void ThumbnailViewer_ControlClosed()
        {
            if (_navigationStateBeforeThumbnailsviewOpening != null)
            {
                NavigationState afterState = NavigationState.TakeSnapshot(PDFViewCtrl);
                TabControlViewModel.SelectedTab.NavigationStack.RegisterJump(_navigationStateBeforeThumbnailsviewOpening, afterState);
                _navigationStateBeforeThumbnailsviewOpening = null;
            }

            IsThumbnailsViewOpen = false;
        }

        private void CropView_ControlClosed()
        {
            int cropViewPage = CropView.ViewModel.CurrCropItem.Page;
            TabControlViewModel.SelectedTab.NavigationStack.SetCurrentPage(cropViewPage);
            IsCropViewOpen = false;
        }

        private async void CropViewModel_DocumentEdited()
        {
            TabControlViewModel.SelectedTab.NavigationStack.Clear();
            await HandleDocumentEditingAsync();
        }

        private async void TabControlViewModel_BrowseDocumentRequested(TabControlViewModel tabControlViewModel)
        {
            IsAppBarOpen = false;
            await BrowseAsync();
        }

        private async void TabControlViewModel_CreateDocumentRequested(TabControlViewModel tabControlViewModel)
        {
            IsAppBarOpen = false;

            if (_createDocControl == null)
            {
                _createDocControl = new CreateDocControl();
                _createDocControl.NewDocumentCreated += _createDocControl_NewDocumentCreated;
                _createDocControl.DocumentCreationFailed += _createDocControl_DocumentCreationFailed;
            }

            await DialogHelper.ShowContentDialogAsync(_createDocControl);
        }

        private void _createDocControl_DocumentCreationFailed()
        {

        }

        private async void _createDocControl_NewDocumentCreated(StorageFile file)
        {
            NewDocumentProperties docProps = new NewDocumentProperties()
            {
                OpenedDocumentState = OpenedDocumentStates.Created
            };

            await DocumentOpener.OpenAsync(file, docProps);
        }

        private async void TabControlViewModel_CloseButtonClicked(TabControlViewModel tabControlViewModel, PdfViewCtrlTabInfo clickedTab)
        {
            CloseCurrentTool();
            AppbarInteraction();

            bool conflict = !clickedTab.IsUpToDate && clickedTab.IsDocumentModifiedSinceLastSave;
            if (conflict)
            {
                await tabControlViewModel.SelectTabAsync(clickedTab);
            }
            else
            {
                SaveNeededStates status = CheckSavingRequirements(clickedTab);
                _removingTab = true;
                await CloseTabRequestedAsync(status, clickedTab, TabControlViewModel.OpenTabs == 1);
                _removingTab = false;
            }
        }

        private async void TabControlViewModel_VisibleTabChanged(TabControlViewModel tabControlViewModel)
        {
            if (_previousTab == TabControlViewModel.SelectedTab)
                return;

            CloseCurrentTool();
            AppbarInteraction();

            if (_previousTab != null)
            {
                await DeactivateTabAsync(_previousTab);
                SaveNeededStates saveState = CheckSavingRequirements(_previousTab);
                bool continueAsNormal = true;
                if (saveState == SaveNeededStates.saveNeeded)
                {
                    try
                    {
                        IsSecretlyModal = true;
                        TabControlViewModel.SelectedTab.PDFViewCtrl.IsHitTestVisible = false;
                        SaveResult result = await SaveToOriginalAsync(_previousTab, true, true);
                        if (result != SaveResult.e_normal && !_showingAutoSaveResumeMessage)
                        {
                            string message = string.Format(
                                _loader.GetString("ViewerPage_AutoSaveFailed_TabChanged_Message_AccessDenied"),
                                _previousTab.Title);
                            if (result == SaveResult.e_unknown_error)
                            {
                                message = string.Format(
                                    _loader.GetString("ViewerPage_AutoSaveFailed_TabChanged_Message_Unknown"),
                                    _previousTab.Title);
                            }

                            ContentDialog dialog = new ContentDialog
                            {
                                Title = _loader.GetString("ViewerPage_AutoSaveFailed_Title"),
                                Content = message,
                                PrimaryButtonText = _loader.GetString("ViewerPage_AutoSaveFailed_TabChanged_IgnoreAndDiscard"),
                                SecondaryButtonText = _loader.GetString("ViewerPage_AutoSaveFailed_TabChanged_ReturnToTab"),
                                CloseButtonText = _loader.GetString("pdftron_Tool_NoteDialog_Cancel")
                            };

                            dialog.SecondaryButtonCommand = new RelayCommand((_) =>
                            {
                                continueAsNormal = false;
                            });

                            var dialogResult = await DialogHelper.ShowContentDialogAsync(dialog);

                            if (dialogResult == ContentDialogResult.None || dialogResult == ContentDialogResult.Primary)
                                continueAsNormal = false;
                        }
                    }
                    finally
                    {
                        IsSecretlyModal = false;
                        TabControlViewModel.SelectedTab.PDFViewCtrl.IsHitTestVisible = true;
                    }
                }

                if (!continueAsNormal)
                {
                    PdfViewCtrlTabInfo reactivateTab = _previousTab;
                    _previousTab = null;
                    await TabControlViewModel.SelectTabAsync(reactivateTab);
                    return;
                }
            }

            if (TabControlViewModel.SelectedTab != null)
            {
                PdfViewCtrlTabInfo previousTab = _previousTab;
                _previousTab = TabControlViewModel.SelectedTab;
                if (!TabControlViewModel.SelectedTab.IsUpToDate && TabControlViewModel.SelectedTab.IsDocumentModifiedSinceLastSave)
                {
                    await ShowConflictingTabOptionsAsync(tabControlViewModel.SelectedTab, previousTab);
                }
                else
                {
                    (AnnotationToolbar as ControlBase.ICloseableControl).CloseControl();
                    await ActivateTabAsync(TabControlViewModel.SelectedTab);
                }
            }
            else
            {
                await NavigateAwayAsync();
            }

            if (!_removingTab)
            {
                if (IsInputMouse)
                {
                    await DelayThenCloseAppBarAsync(100);
                }
                else
                {
                    await DelayThenCloseAppBarAsync();
                }
            }

            RaisePropertyChanged(nameof(PageNumberIndicator));
        }

        private void TabRightTapHandler(object sender, TappedRoutedEventArgs args)
        {
            if (!IsUiPinned)
            {
                IsAppBarOpen = false;
                IsEntireAppBarOpen = !IsEntireAppBarOpen;
            }
        }

        private async Task ActivateTabAsync(PdfViewCtrlTabInfo tab)
        {
            if (tab == null || TabControlViewModel.SelectedTab != tab || _activeTabs.Contains(tab))
                return;

            if (tab.PDFViewCtrl == null)
            {
                await tab.ActivateAsync();
                _pdfIsReadyHandler = new EventHandler<RoutedEventArgs>(PDFIsReadyHander);
                tab.PdfIsReady += _pdfIsReadyHandler;
                return;
            }

            if (tab.FileLoadingError)
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("ViewerPage_TabControl_TabLoadError_FileMissing_Title"),
                    Content = string.Format(_loader.GetString("ViewerPage_TabControl_TabLoadError_FileMissing_Info"), tab.Title),
                    CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                await CloseTabAsync(tab);
                return;
            }

            if (tab.DocumentState == OpenedDocumentStates.Universal)
            {
                _pdfUpdatedHandler = new EventHandler<RoutedEventArgs>(PDFUpdatedHandler);
                tab.PdfUpdated += _pdfUpdatedHandler;
            }

            _activeTabs.Add(tab);

            _tabIsSubscribed = true;

            tab.Resume();
            // This is how the tools communicate that they have a modal dialog open
            tab.PDFViewCtrl.IsEnabledChanged += PDFViewCtrl_IsEnabledChanged;

            tab.PDFViewCtrl.OnPageNumberChanged += PDFViewCtrl_OnPageNumberChanged;
            if (tab.DocumentState == OpenedDocumentStates.Universal)
            {
                _onConversionChangedHandler = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
                tab.PDFViewCtrl.OnConversionChanged += _onConversionChangedHandler;
            }

            PageNumberIndicator.PDFViewCtrl = tab.PDFViewCtrl;
            PageNumberIndicator.NavigationStack = tab.NavigationStack;

            PDFDoc doc = tab.PDFViewCtrl.GetDoc();
            if (doc != null)
                PDFPrintManager.RegisterForPrintingContract(doc, tab.Title);

            tab.PDFViewCtrl.PointerPressed += PDFViewCtrl_PointerPressed;
            tab.PDFViewCtrl.PointerMoved += PDFViewCtrl_PointerMoved;

            if (tab.NavigationStack != null)
                tab.NavigationStack.NavigationStackChanged += NavigationStack_NavigationStackChanged;

            tab.IsCloseTabButtonVisible = IsCloseTabButtonEnabled;
            tab.ToolManager.ReadOnlyMode = IsReadOnlyModeEnabled;
            tab.ToolManager.InkSmoothingBehaviour = ViewerSettings.InkSmoothingOption;
            tab.ToolManager.StylusAsPen = ViewerSettings.StylusAsPen;
            tab.ToolManager.IsSignatureDialogFullScreen = ViewerSettings.SignatureFullScreenOption;
            tab.PDFViewCtrl.Focus(FocusState.Programmatic);

            AnnotationToolbar.ToolManager = tab.ToolManager;
            if (!IsOutlineDialogOpen && ViewerSettings.OutlineDialogOpen && UtilityFunctions.GetDeviceFormFactorType() != DeviceFormFactorType.Phone)
                IsOutlineDialogOpen = true;

            if (IsOutlineDialogOpen)
                OutlineDialogViewModel = tab.OutlineDialogViewModel;

            ResolveReflow();

            OptimizeFileViewModel = new OptimizeFileViewModel();
            OptimizeFileViewModel.PopupClosed += PopupClosed;
            OptimizeFileViewModel.OptimizeConfirmed += OptimizeFileViewModel_OptimizePopupOptimize;

            bool isPasswordProtected = !string.IsNullOrEmpty(tab.MetaData.Password);
            PasswordFileViewModel = new PasswordFileViewModel();
            PasswordFileViewModel.PopupClosed += PopupClosed;
            PasswordFileViewModel.PasswordConfirmed += PasswordFileViewModel_PasswordConfirmed;

            if (isPasswordProtected)
            {
                SaveAPasswordCopyText = _loader.GetString("ViewerPage_SecondaryOptions_ChangePassword");
            }
            else
            {
                SaveAPasswordCopyText = _loader.GetString("ViewerPage_SecondaryOptions_Password");
            }

            CropPopupViewModel = new CropPopupViewModel(tab.PDFViewCtrl);
            CropPopupViewModel.PopupClosed += PopupClosed;
            CropPopupViewModel.ManualCropRequested += CropPopupViewModel_ManualCropRequested;
            CropPopupViewModel.DocumentEdited += CropPopupViewModel_DocumentEdited;

            _toolManager_NewToolCreated = new NewToolCreatedDelegate(ToolManager_NewToolCreated);
            tab.ToolManager.NewToolCreated += _toolManager_NewToolCreated;
            _toolManager_SingleTap = new SingleTapDelegate(ToolManager_SingleTap);
            tab.ToolManager.SingleTap += _toolManager_SingleTap;

            if (!_tabRightTapHandlers.ContainsKey(tab))
            {
                tab.ReflowTapped += TabRightTapHandler;
                _tabRightTapHandlers[tab] = TabRightTapHandler;
            }

            _textToSpeechListener = new ToolManager.TextToSpeechDelegate(ToolManager_TextToSpeechActivated);
            tab.ToolManager.TextToSpeechActivated += _textToSpeechListener;

            _annotAddedListener = new ToolManager.AnnotationModificationHandler(ToolManager_AnnotationModified);
            _annotEditedListener = new ToolManager.AnnotationModificationHandler(ToolManager_AnnotationModified);
            _annotRemovedListener = new ToolManager.AnnotationModificationHandler(ToolManager_AnnotationModified);
            tab.ToolManager.AnnotationAdded += _annotAddedListener;
            tab.ToolManager.AnnotationEdited += _annotEditedListener;
            tab.ToolManager.AnnotationRemoved += _annotRemovedListener;

            _annotationGroupAddedListener = new ToolManager.AnnotationGroupModificationHandler(ToolManager_AnnotationGroupModified);
            _annotationGroupEditedListener = new ToolManager.AnnotationGroupModificationHandler(ToolManager_AnnotationGroupModified);
            _annotationGroupRemovedListener = new ToolManager.AnnotationGroupModificationHandler(ToolManager_AnnotationGroupModified);
            tab.ToolManager.AnnotationGroupAdded += _annotationGroupAddedListener;
            tab.ToolManager.AnnotationGroupEdited += _annotationGroupEditedListener;
            tab.ToolManager.AnnotationGroupRemoved += _annotationGroupRemovedListener;

            _soundAnnotCreateHandler = new ToolManager.SoundAnnotCreateHandler(ToolManager_OnSoundAnnotRequested);
            tab.ToolManager.OnSoundAnnotRequested += _soundAnnotCreateHandler;

            IsDocumentEditable = tab.DocumentState != OpenedDocumentStates.Uneditable;

            if (tab.OriginalFile == null)
            {
                OpenedDocumentStates oldstate = tab.DocumentState;

                if (oldstate == OpenedDocumentStates.Normal || oldstate == OpenedDocumentStates.NeedsFullSave)
                {
                    tab.DocumentState = OpenedDocumentStates.ReadOnly;

                    ContentDialog dialog = new ContentDialog
                    {
                        Title = _loader.GetString("ViewerPage_TabControl_TabLoadErro_OriginalMissing_Title"),
                        Content = string.Format(_loader.GetString("ViewerPage_TabControl_TabLoadErro_OriginalMissing_Info"), tab.Title),
                        PrimaryButtonText = _loader.GetString("SaveBeforeClosingDialog_SaveAs_Option"),
                        SecondaryButtonText = _loader.GetString("SaveBeforeClosingDialog_Discard_Option"),
                        CloseButtonText = _loader.GetString("pdftron_Tool_NoteDialog_Cancel")
                    };

                    dialog.PrimaryButtonCommand = new RelayCommand(async _ => 
                    {
                        await SaveAsAsync(tab, false, false, SaveAsOption.Save);
                    });

                    dialog.SecondaryButtonCommand = new RelayCommand(async _ =>
                    {
                        await CloseTabAsync(tab);
                    });

                    await DialogHelper.ShowContentDialogAsync(dialog);
                }
            }

            UpdateTabDependantProperties();

            if (!tab.FileLoadingError)
            {
                if (tab.PDFViewCtrl.GetPageCount() >= PAGES_NEEDED_FOR_THUMB_SLIDER)
                {
                    AddThumbnailSlider(tab);
                    PageNumberIndicator.Show();
                }
                else
                {
                    ThumbnailSlider = null;
                }
            }

            ActivateTabImpl(tab);
            IsFindTextDialogOpen = false;
            ResolveColorMode(ViewerSettings.ColorMode);
        }

        protected virtual void ActivateTabImpl(PdfViewCtrlTabInfo tab = null)
        {
        }

        private void AddThumbnailSlider(PdfViewCtrlTabInfo tab)
        {
            ThumbnailSlider = new ThumbnailSlider
            {
                PDFViewCtrl = tab.PDFViewCtrl,
                NavigationStack = tab.NavigationStack
            };

            ThumbnailSlider.ManipulationStarted += ThumbnailSlider_ManipulationStarted;
            ThumbnailSlider.ManipulationCompleted += ThumbnailSlider_ManipulationCompleted;
            ThumbnailSlider.AddHandler(UIElement.PointerPressedEvent, new PointerEventHandler(ThumbnailSlider_PointerPressed), true);
            ThumbnailSlider.PointerReleased += ThumbnailSlider_PointerReleased;
        }

        private void UpdateTabDependantProperties()
        {
            RaisePropertyChanged(string.Empty);
            RaisePropertyChanged(nameof(IsDocumentModifiedSinceLastSave));
            RaisePropertyChanged(nameof(CanSaveDocument));
            RaisePropertyChanged(nameof(HasUserBeenWarnedAboutSaving));
            RaisePropertyChanged(nameof(HasDocumentBeenModifiedSinceOpening));
            RaisePropertyChanged(nameof(CurrentPagePresentationModeSymbol));
            RaisePropertyChanged(nameof(CurrentPagePresentationModeSymbolMargin));
            RaisePropertyChanged(nameof(CurrentPagePresentationModeFontSize));
            RaisePropertyChanged(nameof(IsPreparingForConversion));
            RaisePropertyChanged(nameof(IsDocumentSearchable));
            RaisePropertyChanged(nameof(IsConverting));

            ResolveAutoSaverAndSaveButton();
            if (ViewerPageSettingsViewModel != null)
                ResolveColorMode(ViewerPageSettingsViewModel.ColorMode);
        }

        private async Task DeactivateTabAsync(PdfViewCtrlTabInfo tab)
        {
            try
            {
                await DeactivateTabImplAsync(tab);
            }
            catch (Exception ex)
            {
                PDFNetException pdfEx = new PDFNetException(ex.HResult);
                string error;

                if (pdfEx.IsPDFNetException)
                {
                    error = pdfEx.ToString();
                }
                else
                {
                    error = ex.ToString();
                }

                Debug.WriteLine(error);
            }
        }

        private async Task DeactivateTabImplAsync(PdfViewCtrlTabInfo tab)
        {
            if (tab == null || !_activeTabs.Contains(tab))
                return;

            if (_pdfIsReadyHandler != null)
            {
                tab.PdfIsReady -= _pdfIsReadyHandler;
                _pdfIsReadyHandler = null;
            }

            if (_pdfUpdatedHandler != null)
            {
                tab.PdfUpdated -= _pdfUpdatedHandler;
                _pdfUpdatedHandler = null;
            }

            if (tab.PDFViewCtrl == null)
                return;

            if (_onConversionChangedHandler != null)
            {
                tab.PDFViewCtrl.OnConversionChanged -= _onConversionChangedHandler;
                _onConversionChangedHandler = null;
            }

            if (!_tabIsSubscribed)
                return;

            if (_tabRightTapHandlers.ContainsKey(tab))
            {
                tab.ReflowTapped -= _tabRightTapHandlers[tab];
                _tabRightTapHandlers.Remove(tab);
                _tabRightTapHandlers[tab] = null;
            }

            _activeTabs.Remove(tab);

            _tabIsSubscribed = false;

            // unsubscribe to events
            tab.Pause();
            tab.PDFViewCtrl.IsEnabledChanged -= PDFViewCtrl_IsEnabledChanged;
            tab.PDFViewCtrl.OnPageNumberChanged -= PDFViewCtrl_OnPageNumberChanged;
            _pageNumberChanged?.Invoke(tab);

            if (PageNumberIndicator != null)
            {
                PageNumberIndicator.GotFocus -= PageNumberIndicator_GotFocus;
                PageNumberIndicator.PDFViewCtrl = null;
                PageNumberIndicator.NavigationStack = null;
                PageNumberIndicator.ReflowViewModel = null;
            }

            tab.PDFViewCtrl.PointerPressed -= PDFViewCtrl_PointerPressed;
            tab.PDFViewCtrl.PointerMoved -= PDFViewCtrl_PointerMoved;

            PDFPrintManager.UnRegisterForPrintingContract();

            IsFindTextDialogOpen = false;

            if (tab.NavigationStack != null)
                tab.NavigationStack.NavigationStackChanged -= NavigationStack_NavigationStackChanged;

            if (ViewerPageSettingsViewModel != null)
            {
                ViewerPageSettingsViewModel.PropertyChanged -= ViewerPageSettingsViewModel_PropertyChanged;
                ViewerPageSettingsViewModel.ViewRequestedHandler -= ViewerPageSettingsViewModel_ViewRequestedHandler;
                ViewerPageSettingsViewModel = null;
            }

            if (CustomColorViewModel != null)
            {
                CustomColorViewModel.CustomColorRequested -= CustomColorViewModel_CustomColorRequested;  
                CustomColorViewModel.PopupClosed -= PopupClosed;              
            }

            if (OptimizeFileViewModel != null)
            {
                OptimizeFileViewModel.PopupClosed -= PopupClosed;
                OptimizeFileViewModel.OptimizeConfirmed -= OptimizeFileViewModel_OptimizePopupOptimize;
            }

            if (PasswordFileViewModel != null)
            {
                PasswordFileViewModel.PopupClosed -= PopupClosed;
                PasswordFileViewModel.PasswordConfirmed -= PasswordFileViewModel_PasswordConfirmed;
            }

            if (CropPopupViewModel != null)
            {
                CropPopupViewModel.PopupClosed -= PopupClosed;
                CropPopupViewModel.ManualCropRequested -= CropPopupViewModel_ManualCropRequested;
                CropPopupViewModel.DocumentEdited -= CropPopupViewModel_DocumentEdited;
                CropPopupViewModel = null;
            }

            tab.ToolManager.NewToolCreated -= _toolManager_NewToolCreated;
            tab.ToolManager.SingleTap -= _toolManager_SingleTap;
            _toolManager_NewToolCreated = null;
            _toolManager_SingleTap = null;

            tab.ToolManager.TextToSpeechActivated -= _textToSpeechListener;

            if (MediaElement != null)
            {
                MediaElement.Stop();
                AreAudioButtonsVisible = false;
            }

            tab.ToolManager.AnnotationAdded -= _annotAddedListener;
            tab.ToolManager.AnnotationEdited -= _annotEditedListener;
            tab.ToolManager.AnnotationRemoved -= _annotRemovedListener;
            _annotAddedListener = null;
            _annotEditedListener = null;
            _annotRemovedListener = null;

            tab.ToolManager.AnnotationGroupAdded -= _annotationGroupAddedListener;
            tab.ToolManager.AnnotationGroupEdited -= _annotationGroupEditedListener;
            tab.ToolManager.AnnotationGroupRemoved -= _annotationGroupRemovedListener;
            _annotationGroupAddedListener = null;
            _annotationGroupEditedListener = null;
            _annotationGroupRemovedListener = null;

            if (ThumbnailSlider != null)
                ThumbnailSlider.PDFViewCtrl = null;

            if (FindTextViewModel == null)
            {
                FindTextViewModel.FindTextClosed -= FindTextViewModel_FindTextClosed;
                FindTextViewModel.FindTextResultFound -= FindTextViewModel_FindTextResultFound;
            }

            SaveNeededStates saveState = CheckSavingRequirements(tab);
            try
            {
                if (saveState == SaveNeededStates.saveAsNeeded)
                {
                    await tab.SaveDocumentImplAsync();
                    Debug.WriteLine("Saving on deactivate done.");
                }
                await tab.CloseOutlineAsync();
                if (tab.DisposeWhenSaved)
                {
                    Debug.WriteLine("Disposing tab " + tab.Title);
                    tab.Doc.Dispose();
                }
            }
            catch
            {
                //Swallow exception
            }

            await tab.CloseOutlineAsync();
        }

        private async Task ShowConflictingTabOptionsAsync(PdfViewCtrlTabInfo conflictedTab, PdfViewCtrlTabInfo previousTab)
        {
            ContentDialog dialog = new ContentDialog
            {
                Title = _loader.GetString("ViewerPage_TabConflictDialog_Title"),
                Content = string.Format(_loader.GetString("ViewerPage_TabConflictDialog_Info"), ViewerSettings.DisplayName, conflictedTab.Title),       
                SecondaryButtonText = _loader.GetString("ViewerPage_TabConflictDialog_SaveAs"),
                CloseButtonText = _loader.GetString("pdftron_Tool_NoteDialog_Cancel")
            };

            if (_fileToActivate == null)
            {
                dialog.PrimaryButtonText = _loader.GetString("ViewerPage_TabConflictDialog_UseFile");

                dialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                {
                    await ReopenTabWithNewOriginalFile(conflictedTab);
                });
            }
            else
            {
                dialog.PrimaryButtonText = _loader.GetString("SaveBeforeClosingDialog_Discard_Option");

                dialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                {
                    await DiscardTabAsync(conflictedTab);
                });
            }

            dialog.SecondaryButtonCommand = new RelayCommand(async _ =>
            {
                await SaveTabToNewFileAsync(conflictedTab, previousTab); 
            });

            TabConflictChoiceCanceller canceller = new TabConflictChoiceCanceller();
            if (_currentConflictCanceller != null)
                _currentConflictCanceller.Cancel = true;

            _currentConflictCanceller = canceller;
            if (!canceller.Cancel)
            {
                var dialogResult = await DialogHelper.ShowContentDialogAsync(dialog);

                if (dialogResult == ContentDialogResult.None)
                {
                    if (!canceller.Cancel)
                    {
                        await SwitchAwayFromCurrentTabAsync(conflictedTab, previousTab);
                    }
                    else
                    {
                        Debug.WriteLine("Choice cancelled");
                    }
                }
            }
        }

        private async Task ReopenTabWithNewOriginalFile(PdfViewCtrlTabInfo tab)
        {
            try
            {
                await DeactivateTabImplAsync(tab);
                if (TabControlViewModel.SelectedTab == tab)
                {
                    await tab.UpdateFileAsync();
                    await ActivateTabAsync(tab);
                }
                else
                {
                    tab.IsDocumentModifiedSinceLastSave = false;
                    await TabControlViewModel.SelectTabAsync(tab);
                }
            }
            catch
            {
                //Swallow exception
            }
        }

        private async Task DiscardTabAsync(PdfViewCtrlTabInfo tab) => await CloseTabRequestedAsync(SaveNeededStates.none, tab, true);

        private async Task SaveTabToNewFileAsync(PdfViewCtrlTabInfo conflictedTab, PdfViewCtrlTabInfo previousTab)
        {
            //await Task.Delay(100);
            SaveAsDone += SaveTabToNewFile_SaveDone;
            _tabThatIsMovingFile = conflictedTab;
            _tabToFallBackTo = previousTab;
            await SaveAsAsync(conflictedTab, false, false, SaveAsOption.Save);
        }

        private async void SaveTabToNewFile_SaveDone(bool cancelled, bool failed, Deferral deferral)
        {
            try
            {
                if (cancelled && _tabThatIsMovingFile != null)
                {
                    if (_tabToFallBackTo == null)
                    {
                        await ShowConflictingTabOptionsAsync(_tabThatIsMovingFile, null);
                    }
                    else
                    {
                        _fileToActivate = null; // cancel it all
                        await SwitchAwayFromCurrentTabAsync(_tabThatIsMovingFile, _tabToFallBackTo);
                    }
                }

                if (!cancelled && !failed && _fileToActivate != null)
                {
                    SaveNeededStates status = SaveNeededStates.none;
                    await CloseTabRequestedAsync(status, _tabThatIsMovingFile, true);
                }

                if (_tabToLoadAfterResolvingConflict != null)
                {
                    PdfViewCtrlTabInfo tab = _tabToLoadAfterResolvingConflict;
                    _tabToLoadAfterResolvingConflict = null;
                    await TabControlViewModel.SelectTabAsync(tab);
                }

                _tabThatIsMovingFile = null;
            }
            finally
            {
                deferral.Complete();
                SaveAsDone -= SaveTabToNewFile_SaveDone;
            }
        }

        private async Task SwitchAwayFromCurrentTabAsync(PdfViewCtrlTabInfo conflictedTab, PdfViewCtrlTabInfo previousTab)
        {
            if (previousTab != null)
            {
                await TabControlViewModel.SelectTabAsync(previousTab);
            }
            else
            {
                Frame frame = Window.Current.Content as Frame;
                if (frame.CanGoBack)
                {
                    await NavigateAwayAsync();
                }
                else
                {
                    await ExitAppAsync();
                }
            }
        }

        private async void PDFIsReadyHander(object sender, RoutedEventArgs e)
        {
            if (sender is PdfViewCtrlTabInfo tab)
                await ActivateTabAsync(tab);
        }

        private void PDFUpdatedHandler(object sender, RoutedEventArgs e)
        {
            PdfViewCtrlTabInfo tab = TabControlViewModel.SelectedTab;
            if (tab != null)
            {
                IsDocumentEditable = tab.DocumentState != OpenedDocumentStates.Uneditable && tab.DocumentState != OpenedDocumentStates.Universal;
                RaisePropertyChanged(nameof(IsDocumentSearchable));
                ViewerPageSettingsViewModel?.UpdateSettings();
            }
        }

        private async void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            RaisePropertyChanged(nameof(IsPreparingForConversion));
            RaisePropertyChanged(nameof(IsDocumentSearchable));
            RaisePropertyChanged(nameof(IsConverting));
            if (type == PDFViewCtrlConversionType.e_conversion_progress)
            {
                PageNumberIndicator.UpdatePageNumbers();
                ResolveReflow();
            }
            else
            {
                ResolveThumbnailSlider();
            }
        }

        /// <summary>
        /// The users pressed a button without any modifier keys down
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private async void HotKeyHandler_KeyPressedEvent(Windows.UI.Core.CoreWindow sender, Windows.UI.Core.KeyEventArgs args)
        {
            if (IsSecretlyModal)
                return;

            // Handle escape for closing dialogs
            if (args.VirtualKey == Windows.System.VirtualKey.Escape)
            {
                OutlineDialogViewModel?.InteractionOutsideDialog();

                // If something was closed, don't process escape for fullscreen
                if (ToolManager.CloseOpenDialog(true))
                    return;

                if (IsFindTextDialogOpen)
                {
                    IsFindTextDialogOpen = false;

#if SUBSCRIPTIONS
                    if (ViewerSettings.IsSubscriptionActivated)
                    {
                        IsPdfCompareVisible = true;
                    }
                    else
                    {
                        IsPdfCompareSubVisible = true;
                        IsDiamondSubVisible = true;
                    }
#endif

                    return;
                }

                AnnotationToolbarViewModel.PerformEscapeFromTool();
            }

            // Exit full screen mode via Escape/F11 if in full screen mode
            if ((args.VirtualKey == Windows.System.VirtualKey.Escape || args.VirtualKey == Windows.System.VirtualKey.F11) && IsFullScreen)
            {
                CloseDismissableDialogs();
                CloseDialogsOnAppBarOpen();
                UtilityFunctions.SetFullScreenModeDontWait(false);
            }            
            else if (args.VirtualKey == Windows.System.VirtualKey.F11 && !IsFullScreen)
            {
                // Enter full screen mode
                UtilityFunctions.SetFullScreenModeDontWait(true);
            }
            else if (args.VirtualKey == VirtualKey.PageUp)
            {
                // Scroll document up one page
                if (PDFViewCtrl != null)
                    PDFViewCtrl.GotoPreviousPage();
            }            
            else if (args.VirtualKey == VirtualKey.PageDown)
            {
                // Scroll document down one page
                if (PDFViewCtrl != null)
                    PDFViewCtrl.GotoNextPage();
            }
            else if (args.VirtualKey == VirtualKey.C)
                UpdateCurrentToolTo(ToolType.e_sticky_note_create);
            else if (args.VirtualKey == VirtualKey.H)
                UpdateCurrentToolTo(ToolType.e_text_highlight);
            else if (args.VirtualKey == VirtualKey.U)
                UpdateCurrentToolTo(ToolType.e_text_underline);
            else if (args.VirtualKey == VirtualKey.G)
                UpdateCurrentToolTo(ToolType.e_text_squiggly);
            else if (args.VirtualKey == VirtualKey.K)
                UpdateCurrentToolTo(ToolType.e_stamper);
            else if (args.VirtualKey == VirtualKey.S)
                UpdateCurrentToolTo(ToolType.e_signature);
            else if (args.VirtualKey == VirtualKey.I)
                UpdateCurrentToolTo(ToolType.e_ink_create);
            else if (args.VirtualKey == VirtualKey.E)
                UpdateCurrentToolTo(ToolType.e_ink_eraser);
            else if (args.VirtualKey == VirtualKey.M)
                UpdateCurrentToolTo(ToolType.e_ink_marker);
            else if (args.VirtualKey == VirtualKey.T)
                UpdateCurrentToolTo(ToolType.e_text_annot_create);
            else if (args.VirtualKey == VirtualKey.R)
                UpdateCurrentToolTo(ToolType.e_rect_create);
            else if (args.VirtualKey == VirtualKey.O)
                UpdateCurrentToolTo(ToolType.e_oval_create);
            else if (args.VirtualKey == VirtualKey.L)
                UpdateCurrentToolTo(ToolType.e_line_create);
            else if (args.VirtualKey == VirtualKey.A)
                UpdateCurrentToolTo(ToolType.e_arrow_create);
            else if (args.VirtualKey == VirtualKey.P)
                UpdateCurrentToolTo(ToolType.e_pan);

            if (IsFindTextDialogOpen && FindTextViewModel != null)
            {
                await FindTextViewModel.HandleKeyboardEventAsync(args.VirtualKey, false, false, _hotKeyHandler.IsShiftDown);
            }
        }

        private void UpdateCurrentToolTo(ToolType newCurrentTool)
        {
            if (IsFindTextDialogOpen)
                return;

            if (AnnotationToolbarViewModel == null || ToolManager == null)
                return;

            if (AnnotationToolbarViewModel.ShowInkToolbar 
                || AnnotationToolbarViewModel.ShowMeasureToolbar 
                || AnnotationToolbarViewModel.ShowPolygonSaveToolbar
                || AnnotationToolbarViewModel.ShowRedactionToolbar)
            {
                AnnotationToolbarViewModel.CommitCurrentToolChanges(AnnotationToolbarViewModel.CurrentTool);
                AnnotationToolbarViewModel.UpdateCurrentToolTo(newCurrentTool);
            }

            if (ToolManager.CurrentTool?.IsLoaded == false)
            {
                AnnotationToolbarViewModel.UpdateCurrentToolTo(newCurrentTool);
            }            
        }

        public async Task OpenTextSearchAsync()
        {
            if (OpenSearch != null)
            {
                var tcs = new TaskCompletionSource<bool>();
                using (var deferral = new Deferral(() => tcs.SetResult(true)))
                {
                    OpenSearch.Invoke(this, deferral);
                    await tcs.Task;
                }
            }
        }

        /// <summary>
        /// Shortcut events
        /// Note: Ctrl + something was pressed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private async void HotKeyHandler_HotKeyPressedEvent(Windows.UI.Core.CoreWindow sender, Windows.UI.Core.KeyEventArgs args)
        {
            if (IsModal)
                return;

            switch (args.VirtualKey)
            {
                case Windows.System.VirtualKey.S:
                    if (IsSaveButtonEnabled && CanSaveDocument)
                    {
                        await SaveAsync();
                    }
                    break;

                case Windows.System.VirtualKey.F:
                    if (!IsTextSearchButtonEnabled)
                        break;
                    await OpenTextSearchAsync();

                    IsFindTextDialogOpen = true;
                    break;

                case Windows.System.VirtualKey.P:
                    try
                    {
                        if (Windows.UI.Xaml.Window.Current.Bounds.Width >= 500)
                        {
                            await TryPrintAsync();
                        }
                    }
                    catch (Exception) { }
                    break;

                case Windows.System.VirtualKey.W:

                    if (TabControlViewModel.Tabs.Count > 0)
                    {
                        SaveNeededStates status = CheckSavingRequirements(TabControlViewModel.SelectedTab);
                        await CloseTabRequestedAsync(status, TabControlViewModel.SelectedTab, false);
                        if (TabControlViewModel.Tabs.Count == 0)
                        {
                            await NavigateAwayAsync();
                        }
                    }

                    break;

                case Windows.System.VirtualKey.Z:

                    if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
                    {
                        var annotToolBar = AnnotationToolbar.DataContext as AnnotationToolbarViewModel;
                        annotToolBar.PerformInkUndo();
                        return;
                    }
                    else if (ToolManager.UndoRedoAction.CanUndo)
                    {
                        // Ensure to clear any UI elements left over
                        PDFViewCtrl.GetAnnotationCanvas().Children.Clear();

                        ToolManager.UndoRedoAction.DoUndo();
                    }
                    break;

                case VirtualKey.X:
                    if (ToolManager.SelectedAnnotation != null)
                    {
                        Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Cut Annotation - Ctrl + X" });

                        AnnotationClipboardHelper.CopyAnnotToClipboard(PDFViewCtrl, ToolManager.SelectedAnnotation);
                        ToolManager.DeleteAnnotation(PDFViewCtrl, ToolManager.SelectedAnnotation, ToolManager.CurrentTool.AnnotationPageNumber);
                        CloseCurrentTool();
                    }
                    break;

                case Windows.System.VirtualKey.Y:

                    if (ToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
                    {
                        var annotToolBar = AnnotationToolbar.DataContext as AnnotationToolbarViewModel;
                        annotToolBar.PerformInkRedo();
                        return;
                    }
                    else if (ToolManager.UndoRedoAction.CanRedo)
                    {
                        // Ensure to clear any UI elements left over
                        PDFViewCtrl.GetAnnotationCanvas().Children.Clear();

                        ToolManager.UndoRedoAction.DoRedo();
                    }
                    break;

                case Windows.System.VirtualKey.Number1:
                case Windows.System.VirtualKey.NumberPad1:
                    await HotKeyTabHandlerAsync(0);
                    break;
                case Windows.System.VirtualKey.Number2:
                case Windows.System.VirtualKey.NumberPad2:
                    await HotKeyTabHandlerAsync(1);
                    break;
                case Windows.System.VirtualKey.Number3:
                case Windows.System.VirtualKey.NumberPad3:
                    await HotKeyTabHandlerAsync(2);
                    break;
                case Windows.System.VirtualKey.Number4:
                case Windows.System.VirtualKey.NumberPad4:
                    await HotKeyTabHandlerAsync(3);
                    break;
                case Windows.System.VirtualKey.Number5:
                case Windows.System.VirtualKey.NumberPad5:
                    await HotKeyTabHandlerAsync(4);
                    break;
                case Windows.System.VirtualKey.Number6:
                case Windows.System.VirtualKey.NumberPad6:
                    await HotKeyTabHandlerAsync(5);
                    break;
                case Windows.System.VirtualKey.Number7:
                case Windows.System.VirtualKey.NumberPad7:
                    await HotKeyTabHandlerAsync(6);
                    break;
                case Windows.System.VirtualKey.Number8:
                case Windows.System.VirtualKey.NumberPad8:
                    await HotKeyTabHandlerAsync(7);
                    break;
                case Windows.System.VirtualKey.Number9:
                case Windows.System.VirtualKey.NumberPad9:
                    await HotKeyTabHandlerAsync(9);
                    break;
                case Windows.System.VirtualKey.Number0:
                    if (PDFViewCtrl != null)
                    {
                        PDFViewCtrl.SetPageViewMode(PDFViewCtrlPageViewMode.e_fit_width, PDFViewCtrl.ActualWidth / 2, PDFViewCtrl.ActualHeight / 2, true);
                    }
                    break;

                case Windows.System.VirtualKey.Tab:
                    if (_hotKeyHandler.IsShiftDown)
                    {
                        await HotKeyNextTabHandlerAsync(true);
                    }
                    else
                    {
                        await HotKeyNextTabHandlerAsync(false);
                    }
                    args.Handled = true;
                    break;

                case VirtualKey.C:
                    if (ToolManager.SelectedAnnotation != null)
                    {
                        Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Copy Annotation - Ctrl + C" });
                        AnnotationClipboardHelper.CopyAnnotToClipboard(PDFViewCtrl, ToolManager.SelectedAnnotation);
                    }
                    break;

                case VirtualKey.V:
                    if (AnnotationClipboardHelper.IsAnnotToPaste)
                    {
                        Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Paste Annotation - Ctrl + V" });

                        var currentPage = PDFViewCtrl.GetCurrentPage();

                        bool hasPasted = await AnnotationClipboardHelper.PasteAnnotFromClipboardAsync(PDFViewCtrl, currentPage);
                        if (hasPasted == false)
                            return;

                        try
                        {
                            var doc = PDFViewCtrl.GetDoc();
                            if (doc.GetPageCount() >= currentPage)
                            {
                                var page = doc.GetPage(currentPage);
                                var annot = page.GetAnnot(page.GetNumAnnots() - 1);
                                ToolManager.RaiseAnnotationAddedEvent(annot, currentPage);
                                ToolManager.SelectAnnotation(annot, currentPage);
                            }
                        }
                        catch
                        {
                            //Swallow exception
                        }
                    }
                    break;
                case VirtualKey.M:
                    UpdateCurrentToolTo(ToolType.e_annot_edit_group_create);
                    break;
                case VirtualKey.L:
                    UpdateCurrentToolTo(ToolType.e_lasso_create);
                    break;
            }
        }

        private async Task HotKeyTabHandlerAsync(int index)
        {
            if (index == 9)
                index = TabControlViewModel.OpenTabs - 1;

            if (index >= 0 && TabControlViewModel.OpenTabs > index)
                await TabControlViewModel.SelectTabImplAsync(index);
        }

        private async Task HotKeyNextTabHandlerAsync(bool goBackwards)
        {
            int index = TabControlViewModel.VisibleTabIndex;
            if (goBackwards)
            {
                index--;
                if (index < 0)
                    index = TabControlViewModel.OpenTabs - 1;
            }
            else
            {
                index++;
                if (index >= TabControlViewModel.OpenTabs)
                    index = 0;
            }

            await HotKeyTabHandlerAsync(index);
        }

        /// <summary>
        /// ctrl + alt + something was pressed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private async void HotKeyHandler_CtrlAltHotKeyPressedEvent(CoreDispatcher sender, AcceleratorKeyEventArgs args)
        {
            if (IsModal)
                return;

            switch (args.VirtualKey)
            {
                case VirtualKey.S:
                    if (IsSaveAsButtonVisible)
                        await SaveAsAsync(TabControlViewModel.SelectedTab, false, false, SaveAsOption.Save);

                    break;
            }
        }

        private async void BackButtonHandler_BackPressed(object sender, BackRequestedEventArgs e)
        {
            var frame = Window.Current.Content as Frame;
            if (frame == null)
                return;

            if (IsAppBarOpen)
            {
                IsAppBarOpen = false;
                e.Handled = true;
                return;
            }

            if (ToolManager?.CloseOpenDialog() == true)
            {
                e.Handled = true;
                return;
            }

            if (IsAnnotationToolbarOpen)
            {
                if (AnnotationToolbar?.GoBack() != true)
                    IsAnnotationToolbarOpen = false;

                ToolManager.CreateDefaultTool();
                e.Handled = true;
                return;
            }

            if (IsThumbnailsViewOpen && ThumbnailViewer != null)
            {
                if (!ThumbnailViewer.GoBack())
                    IsThumbnailsViewOpen = false;

                e.Handled = true;
                return;
            }

            if (IsOutlineDialogOpen && OutlineDialogViewModel != null)
            {
                if (!OutlineDialogViewModel.GoBack())
                    IsOutlineDialogOpen = false; // we only hide it if the outline dialog didn't take care of itself.

                e.Handled = true;
                return;
            }

            if (IsFindTextDialogOpen && FindTextViewModel != null)
            {
                if (!FindTextViewModel.GoBack())
                    IsFindTextDialogOpen = false;

                e.Handled = true;
                return;
            }

            if (IsCropViewOpen)
            {
                if (!CropView.GoBack())
                    IsFindTextDialogOpen = false;

                e.Handled = true;
                return;
            }

            if (PageNumberIndicator.GoBack())
            {
                e.Handled = true;
                return;
            }

            if (IsFullScreen && pdftron.PDF.Tools.UtilityFunctions.GetDeviceFormFactorType() != DeviceFormFactorType.Phone)
            {
                pdftron.PDF.Tools.UtilityFunctions.SetFullScreenModeDontWait(false);
                e.Handled = true;
                return;
            }

            SaveNeededStates status = SaveNeededStates.none;
            if (TabControlViewModel.SelectedTab != null)
                status = CheckSavingRequirements(TabControlViewModel.SelectedTab);

            if (status == SaveNeededStates.none)
            {
                if (_launchedThroughFileActivation)
                {
                    e.Handled = true;
                    PDFViewCtrl?.CloseDoc();

                    await NavigateAwayAsync();
                }
                else
                {
                    e.Handled = true;
                    await NavigateAwayAsync();
                }
            }
            else
            {
                e.Handled = true;
                if (!frame.CanGoBack)
                    _closeAppAfterSaving = true;

                if (status == SaveNeededStates.saveNeeded)
                {
                    await CloseTabRequestedAsync(status, TabControlViewModel.SelectedTab, true);
                }
                else
                {
                    await NavigateAwayAsync();
                }
            }

            // NOTE: if ViewerControl is used as a standalone control it needs proper navigation deactivation handling
            if (IsDeactivationOnBackNavigationEnabled)
            {
                Deactivate(null);
            }

            if (IsForceGoBackAllowed)
            {
                if (frame.CanGoBack)
                {
                    frame.GoBack();
                }
            }
        }

        ////////////////////////////////////////////////////////////////////////////////////////////////////////////
        // These listeners lets us know when not to close the app bar automatically.

        private void InitAppBarManagement()
        {
            _appBarClosingTimer = new DispatcherTimer
            {
                Interval = TimeSpan.FromSeconds(5)
            };

            _appBarClosingTimer.Tick += AppBarClosingTimer_Tick;
        }

        // Auto closing the app bar
        private void AppbarInteraction()
        {
            OutlineDialogViewModel?.InteractionOutsideDialog();
            if (_useTimedAppBar)
            {
                _appBarClosingTimer?.Stop();
                _appBarClosingTimer?.Start();
            }
        }

        private void CloseCurrentTool(bool creationToolsOnly = false)
        {
            if (ToolManager != null)
            {
                if (!creationToolsOnly)
                {
                    ToolManager.CreateDefaultTool();
                }
                else
                {
                    ToolType currentToolType = ToolManager.CurrentTool.ToolMode;

                    if (currentToolType != ToolType.e_text_select
                        && currentToolType != ToolType.e_pan
                        && currentToolType != ToolType.e_signature)
                    {
                        ToolManager.CreateDefaultTool();
                    }
                }
            }
        }

        private void AppBarClosingTimer_Tick(object sender, object e)
        {
            DispatcherTimer timer = sender as DispatcherTimer;

            if (IsFindTextDialogOpen || IsAnnotationToolbarOpen)
            {
                timer.Stop();
                return;
            }

            IsAppBarOpen = false;
            if (!ViewerSettings.PinCommandBar)
                IsEntireAppBarOpen = false;

            timer.Stop();
        }

        private void ThumbnailSlider_ManipulationCompleted(object sender, ManipulationCompletedRoutedEventArgs e)
        {
            if (_useTimedAppBar)
                _appBarClosingTimer?.Start();
        }

        private void ThumbnailSlider_ManipulationStarted(object sender, ManipulationStartedRoutedEventArgs e)
        {
            if (_useTimedAppBar)
                _appBarClosingTimer?.Stop();
        }

        private void ThumbnailSlider_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            if (_useTimedAppBar)
                _appBarClosingTimer?.Stop();
        }

        private void ThumbnailSlider_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            if (_useTimedAppBar)
                _appBarClosingTimer?.Start();
        }

        public DocumentOpener DocumentOpener
        {
            get => _documentOpener;
            set => Set(ref _documentOpener, value);
        }

        /// <summary>
        /// Use this to activate the current view with a file.
        /// </summary>
        /// <remarks>
        /// Supported file types:
        /// .pdf, .doc, .docx, .ppt, .pptx, .xls, .xlsx, .oxps, .xps, .jpeg, .jpg, .png, .bmp, .cbz    
        ///  
        /// </remarks>
        /// <param name="file">The file to open</param>
        public async Task ActivateWithFileAsync(StorageFile file, NewDocumentProperties properties)
        {
            await ResumeSemaphore.WaitAsync();
            try
            {
                if (_viewerActivateWithFile != null)
                {
                    var tcs = new TaskCompletionSource<bool>();
                    using (var deferral = new Deferral(() => tcs.SetResult(true)))
                    {
                        _viewerActivateWithFile.Invoke(this, deferral);
                        await tcs.Task;
                    }
                }

                if (TabControlViewModel.IsReady)
                {
                    await TabControlViewModel.Instance.UpdateFilePropertiesAsync();
                }
                else
                {
                    IsModal = true;
                    await TabControlViewModel.Instance.InitializeInstanceAsync();
                    IsModal = false;
                }

                if (DocumentManager.IsReady == false)
                {
                    IsModal = true;
                    await DocumentManager.Instance.InitializeInstanceAsync();
                    IsModal = false;
                }

                await ActivateWithFileImplAsync(file, properties);
            }
            finally
            {
                ResumeSemaphore.Release();
            }

            // Ensure it is a supported file type
            string fileType = file.FileType.ToLower();
            if (ViewerSettings.AssociatedFileTypes.Contains(fileType) == false)
            {
                ContentDialog dialog = new ContentDialog
                {
                    Title = _loader.GetString("DocumentsPage_FileOpeningError_Title"),
                    Content = _loader.GetString("pdftron_Controls_fileFormatNotSupported"),
                    CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };

                await DialogHelper.ShowContentDialogAsync(dialog);
                return;
            }

            _fileToActivate = file;
            _launchedThroughFileActivation = true;
            _fileToActivateDocumentProperties = properties;

            if (_currentConflictCanceller != null)
                _currentConflictCanceller.Cancel = true;

            if (TabControlViewModel.ContainsFile(file))
            {
                _fileToActivate = null;
                foreach (PdfViewCtrlTabInfo tab in TabControlViewModel.Tabs)
                {
                    if (UtilityFunctions.AreFilesEqual(tab.OriginalFile, file))
                    {
                        if (!tab.IsUpToDate && tab.IsDocumentModifiedSinceLastSave)
                        {
                            _tabToLoadAfterResolvingConflict = tab;
                            await ShowConflictingTabOptionsAsync(tab, null);
                        }
                        else
                        {
                            if (tab.MetaData.IsOpenedThroughDrop
                                && (file.Attributes & FileAttributes.Temporary) != FileAttributes.Temporary
                                && (file.Attributes & FileAttributes.ReadOnly) != FileAttributes.ReadOnly)
                            {
                                tab.OriginalFile = file;
                                tab.DocumentState = OpenedDocumentStates.Normal;
                                tab.MetaData.IsOpenedThroughDrop = false;
                                UpdateTabDependantProperties();
                            }

                            await TabControlViewModel.SelectTabAsync(tab);
                            ActivateTabImpl(tab);
                        }

                        return;
                    }
                }
            }
            else
            {
                if (TabControlViewModel.OpenTabs >= TabControlViewModel.MaximumItems)
                {
                    bool conflict = !TabControlViewModel.OldestViewedTab.IsUpToDate && TabControlViewModel.OldestViewedTab.IsDocumentModifiedSinceLastSave;
                    if (conflict)
                    {
                        await ShowConflictingTabOptionsAsync(TabControlViewModel.OldestViewedTab, null);
                    }
                    else
                    {
                        SaveNeededStates status = CheckSavingRequirements(TabControlViewModel.OldestViewedTab);
                        await CloseTabRequestedAsync(status, TabControlViewModel.OldestViewedTab, true);
                    }
                }
                else
                {
                    await ContinueFileActivationAsync();
                }
            }
        }

        protected virtual Task ActivateWithFileImplAsync(StorageFile file, NewDocumentProperties properties) => Task.CompletedTask;

        private async Task ContinueFileActivationAsync()
        {
            StorageFile file = _fileToActivate;
            _fileToActivate = null;
            _suppressRecentListUpdate = false;

            if (_fileToActivateDocumentProperties?.OpenedThroughDrop == false)
            {
                DocumentOpener_DocumentReady(_fileToActivateDocumentProperties);
                _fileToActivateDocumentProperties = null;
            }
            else
            {
                IsModal = true;
                DocumentOpener = new DocumentOpener();
                DocumentOpener.DocumentReady += DocumentOpener_DocumentReady;
                DocumentOpener.ModalityChanged += DocumentOpener_ModalityChanged;
                DocumentOpener.DocumentLoadingFailed += DocumentOpener_DocumentLoadingFailed;
                NewDocumentProperties props = _fileToActivateDocumentProperties;
                _fileToActivateDocumentProperties = null;
                await DocumentOpener.OpenAsync(file, props);
            }
        }

        private void DocumentOpener_ModalityChanged(bool isModal) => IsModal = isModal;

        private async void DocumentOpener_DocumentReady(NewDocumentProperties selectedFileProperties)
        {
            //DocumentOpener = null;
            await InitPdfViewAsync(selectedFileProperties);
            ActivateTabImpl();
            IsModal = false;
        }

        private async void DocumentOpener_DocumentLoadingFailed(object sender, RoutedEventArgs e)
        {
            IsModal = false;
            //DocumentOpener = null;
            if (TabControlViewModel.OpenTabs > 0)
            {
                await TabControlViewModel.ShowLastViewedTabAsync();
            }
            else
            {
                await NavigateAwayAsync();
            }
        }

        private async Task ExitAppAsync()
        {
            try
            {
                //if (TabControlViewModel.SelectedTab?.IsFixedItem == false)
                await TabControlViewModel.SelectedTab?.SaveDocumentImplAsync();
                TabControlViewModel.SaveTabState();
            }
            catch
            {
                //Swallow exception
            }

            Application.Current.Exit();
        }

        private void CloseDismissableDialogs()
        {
            IsFindTextDialogOpen = false;
            IsDismissableDialogOpen = false;
        }

        private void CloseDialogsOnAppBarOpen()
        {
            // Don't close Find Text dialog when expanding app bar
        }

        private async Task DelayThenCloseAppBarAsync(int milleseconds = 300)
        {
            await Task.Delay(milleseconds);
            IsAppBarOpen = false;
        }

        private SaveNeededStates CheckSavingRequirements(PdfViewCtrlTabInfo tab)
        {
            if (tab == null)
                return SaveNeededStates.none;

            if (ViewerSettings.AutoSaveOn && tab.IsDocumentModifiedSinceLastSave &&
                (tab.DocumentState == OpenedDocumentStates.Normal
                || tab.DocumentState == OpenedDocumentStates.NeedsFullSave))
            {
                return SaveNeededStates.saveNeeded;
            }
            else if (tab.DocumentState == OpenedDocumentStates.CorruptedAndModified
                || tab.DocumentState == OpenedDocumentStates.Created
                || (tab.DocumentState == OpenedDocumentStates.ReadOnly && tab.HasDocumentBeenModifiedSinceOpening)
                || (tab.DocumentState == OpenedDocumentStates.Normal && tab.IsDocumentModifiedSinceLastSave)
                || (tab.DocumentState == OpenedDocumentStates.NeedsFullSave && tab.IsDocumentModifiedSinceLastSave)
                || (tab.DocumentState == OpenedDocumentStates.NonePDF && tab.IsDocumentModifiedSinceLastSave)
                )
            {
                return SaveNeededStates.saveAsNeeded;
            }

            return SaveNeededStates.none;
        }

        public async Task CloseTabRequestedAsync(SaveNeededStates saveNeededStatus, PdfViewCtrlTabInfo tabToClose, bool navigateWhenDone)
        {
            CloseDismissableDialogs();
            CloseDialogsOnAppBarOpen();

            if (saveNeededStatus == SaveNeededStates.saveNeeded)
            {
                if (tabToClose.PDFViewCtrl != null)
                {
                    tabToClose.PDFViewCtrl.CancelRendering();
                    tabToClose.PDFViewCtrl.CancelAllThumbRequests();
                }

                try
                {
                    IsSecretlyModal = true;
                    if (tabToClose.PDFViewCtrl != null)
                        tabToClose.PDFViewCtrl.IsHitTestVisible = false;

                    SaveResult result = await SaveToOriginalAsync(tabToClose, true, true);
                    if (result != SaveResult.e_normal)
                    {
                        string message = string.Format(
                            _loader.GetString("ViewerPage_AutoSaveFailed_Close_Message_AccessDenied"),
                            tabToClose.Title,
                            _loader.GetString("Generic_Cancel_Text"));
                        if (result == SaveResult.e_unknown_error)
                        {
                            message = string.Format(
                                _loader.GetString("ViewerPage_AutoSaveFailed_Close_Message_Unknown"),
                                tabToClose.Title);
                        }

                        ContentDialog dialog = new ContentDialog
                        {
                            Title = _loader.GetString("ViewerPage_AutoSaveFailed_Title"),
                            Content = message,
                            PrimaryButtonText = _loader.GetString("ViewerPage_AutoSaveFailed_Close_Discard"),
                            CloseButtonText = _loader.GetString("Generic_Cancel_Text")
                        };

                        dialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                        {
                            tabToClose.DocumentState = OpenedDocumentStates.None;
                            await CloseTabAsync(tabToClose);

                        // TODO : Phone devices may want to leave the app at this point if the app was activated through app picker
                        if (false && _launchedThroughFileActivation && _fileToActivate == null)
                            {
                                await ExitAppAsync();
                            }
                            else if (navigateWhenDone)
                            {
                                await NavigateAwayAsync();
                            }
                        });

                        dialog.CloseButtonCommand = new RelayCommand((_) =>
                        {
                            _fileToActivate = null;
                        });

                        var dialogResult = await DialogHelper.ShowContentDialogAsync(dialog);
                        if (dialogResult == ContentDialogResult.None)
                            _fileToActivate = null;
                    }
                    else
                    {
                        await CloseTabAsync(tabToClose);
                        if (navigateWhenDone)
                            await NavigateAwayAsync();
                    }
                }
                finally
                {
                    IsSecretlyModal = false;
                    if (tabToClose.PDFViewCtrl != null)
                        tabToClose.PDFViewCtrl.IsHitTestVisible = true;
                }
            }
            else if (saveNeededStatus == SaveNeededStates.saveAsNeeded)
            {
                string message = string.Format(_loader.GetString("SaveBeforeClosingDialog_Info"), tabToClose.Title);

                ContentDialog messageDialog = new ContentDialog
                {
                    Title = _loader.GetString("SaveBeforeClosingDialog_Title"),
                    Content = message,
					CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
                };              
          

                if (tabToClose.DocumentState == OpenedDocumentStates.Normal || tabToClose.DocumentState == OpenedDocumentStates.NeedsFullSave)
                {
                    messageDialog.PrimaryButtonText = _loader.GetString("SaveBeforeClosingDialog_Save_Option");

                    messageDialog.PrimaryButtonCommand = new RelayCommand(async _ => 
                    {
                        await SaveAndCloseTabAsync(tabToClose, navigateWhenDone);
                    });
                }
                else
                {
                    messageDialog.PrimaryButtonText = _loader.GetString("SaveBeforeClosingDialog_SaveAs_Option");

                    messageDialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                    {
                        // We can't call SaveAs(true) here, or we'll get an UnauthorizedAccessException when trying to open the file picker.
                        await DelaySaveAs(tabToClose, true, navigateWhenDone);
                    });
                }

                messageDialog.SecondaryButtonText = _loader.GetString("SaveBeforeClosingDialog_Discard_Option");

                messageDialog.SecondaryButtonCommand = new RelayCommand(async _ =>
                {
                    tabToClose.DocumentState = OpenedDocumentStates.None;
                    await CloseTabAsync(tabToClose);
                    // TODO : Phone devices may want to leave the app at this point if the app was activated through app picker
                    if (false && _launchedThroughFileActivation && _fileToActivate == null)
                    {
                        await ExitAppAsync();
                    }
                    else if (navigateWhenDone)
                    {
                        await NavigateAwayAsync();
                    }
                });

                if (pdftron.PDF.Tools.UtilityFunctions.GetDeviceFormFactorType() != DeviceFormFactorType.Phone)
                {
                    messageDialog.CloseButtonText = _loader.GetString("Generic_Cancel_Text");

                    messageDialog.CloseButtonCommand = new RelayCommand(async =>
                    {
                        _fileToActivate = null;
                    });
                }

                var command = await DialogHelper.ShowContentDialogAsync(messageDialog);
                if (command == ContentDialogResult.None)
                    _fileToActivate = null;
            }
            else
            {
                if (_tabToLoadAfterResolvingConflict != null)
                {
                    PdfViewCtrlTabInfo tab = _tabToLoadAfterResolvingConflict;
                    _tabToLoadAfterResolvingConflict = null;
                    await TabControlViewModel.SelectTabAsync(tab);
                }
                else
                {
                    await CloseTabAsync(tabToClose);
                    if (navigateWhenDone)
                        await NavigateAwayAsync();
                }
            }
        }

        private async Task SaveAndCloseTabAsync(PdfViewCtrlTabInfo tab, bool navigateWhenDone)
        {
            await SaveHelperAsync(tab, true, true);
            await CloseTabAsync(tab);
            if (navigateWhenDone)
                await NavigateAwayAsync();
        }

        private async Task DelaySaveAs(PdfViewCtrlTabInfo tab, bool closeWhenDone, bool quitWhenDone)
        {
            // We wait a little, and the save picker can safely be open.
            //await Task.Delay(100);
            if (_launchedThroughFileActivation)
                quitWhenDone = false;

            await SaveAsAsync(tab, closeWhenDone, quitWhenDone, SaveAsOption.Save);
        }

        private async Task CloseTabAsync(PdfViewCtrlTabInfo tab)
        {
            DeactivateReflow();
            tab?.NavigationStack?.Unsubscribe();
            await DeactivateTabAsync(tab);
            if (TabControlViewModel != null)
            {
                if (TabControlViewModel.SelectedTab == tab)
                    _previousTab = null;

                await TabControlViewModel.CloseTabImplAsync(tab, _fileToActivate == null);
            }
        }

        /// <summary>
        /// Closes the document and turns off all auto-saving, etc. The document will be disposed, so ensure it is saved before calling this function.
        /// </summary>
        private async Task NavigateAwayAsync()
        {
            _appBarClosingTimer?.Stop();
            TurnOffAutoSave();
            if (_fileToActivate != null)
            {
                await ContinueFileActivationAsync();
            }
            else
            {
                Reset();
                _newINavigableAvailable?.Invoke(this, NavigateBack);
            }
        }

        private void Reset()
        {
            IsSaveButtonVisible = false;
            IsSaveAsButtonVisible = false;
            IsActionsButtonVisible = false;
            IsAnnotationToolbarOpen = false;

            RaiseViewModelPropertyChanged(nameof(IsConverting));
            RaiseViewModelPropertyChanged(nameof(IsFindTextDialogOpen));
            RaiseViewModelPropertyChanged(nameof(IsDocumentEditable));
            RaiseViewModelPropertyChanged(nameof(IsDocumentSearchable));
            RaiseViewModelPropertyChanged(nameof(IsOutlineEnabled));
            RaiseViewModelPropertyChanged(nameof(IsQuickSettingsEnabled));
            RaiseViewModelPropertyChanged(nameof(ShowPageNumberIndicator));
            RaiseViewModelPropertyChanged(nameof(AreZoombuttonsVisible));
            RaiseViewModelPropertyChanged(nameof(IsThumbnailSliderVisible));
            RaiseViewModelPropertyChanged(nameof(IsPdfCompareActive));
            RaiseViewModelPropertyChanged(nameof(IsMultiWindowActive));
        }

        protected async Task BrowseAsync()
        {
            if (_filePickerOpen)
                return;

            FileOpenPicker fileOpenPicker = new FileOpenPicker
            {
                ViewMode = PickerViewMode.List
            };

            StorageFile file = null;
            foreach (string fileType in ViewerSettings.AssociatedFileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(fileType);
            }
            try
            {
                _filePickerOpen = true;
                // apparently, this sometimes throws a System.Exception "Element not found" for no apparent reason. We want to catch that.
                file = await fileOpenPicker.PickSingleFileAsync();
            }
            finally
            {
                _filePickerOpen = false;
            }

            if (file != null)
            {
                if (DocumentOpener == null)
                {
                    DocumentOpener = new DocumentOpener();
                    DocumentOpener.DocumentReady += DocumentOpener_DocumentReady;
                    DocumentOpener.ModalityChanged += DocumentOpener_ModalityChanged;
                    DocumentOpener.DocumentLoadingFailed += DocumentOpener_DocumentLoadingFailed;
                }
                else
                {
                    DocumentOpener.CancelDocumentOpening();
                }

                await DocumentOpener.OpenAsync(file);

                TabControlViewModel.RaisePropertyChangedImpl("IsAddTabButtonVisible");
            }
        }

        private void ResolveColorMode(CustomColorModeType mode)
        {
            //Color viewerBackground = Color.FromArgb(255, 241, 243, 245);
            Color viewerBackground = Color.FromArgb(255, 255, 255, 255);
            if (ResourceHandler.ViewerResources != null)
                viewerBackground = (Color)ResourceHandler.ViewerResources["MainViewerBackgroundColor"];

            if (PDFViewCtrl != null)
            {
                PDFRasterizerColorPostProcessMode oldMode = PDFViewCtrl.GetColorPostProcessMode();
                PDFViewCtrl.SetBackgroundColor(viewerBackground);
                if (ViewerSettings.ColorMode == CustomColorModeType.Night)
                {
                    // reflow inject css
                    PDFViewCtrl.SetColorPostProcessMode(PDFRasterizerColorPostProcessMode.e_postprocess_night_mode);
                    viewerBackground = (Color)ResourceHandler.ViewerResources["MainViewerNightModeBackgroundColor"];
                    PDFViewCtrl.SetBackgroundColor(viewerBackground);
                    PDFViewCtrl.Background = new SolidColorBrush(viewerBackground);
                    PDFViewCtrl.Update(true);
                }
                else if (ViewerSettings.ColorMode == CustomColorModeType.Sepia)
                {
                    Color paperColor = (Color)ResourceHandler.ViewerResources["ViewerSepiaPaperColor"];
                    Color textColor = (Color)ResourceHandler.ViewerResources["ViewerSepiaTextColor"];
                    viewerBackground = UtilityFunctions.GetViewerBackgroundColor(paperColor);
                    if (!UtilityFunctions.MatchesColorPostProcessMode(PDFViewCtrl, PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map, paperColor, textColor))
                    {
                        PDFViewCtrl.SetColorPostProcessMode(PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map);
                        PDFViewCtrl.Background = new SolidColorBrush(viewerBackground);
                        PDFViewCtrl.SetColorPostProcessColors(paperColor, textColor);
                    }

                    PDFViewCtrl.SetBackgroundColor(viewerBackground);
                    PDFViewCtrl.Update(true);
                }
                else if (ViewerSettings.ColorMode == CustomColorModeType.none && oldMode != PDFRasterizerColorPostProcessMode.e_postprocess_none)
                {
                    PDFViewCtrl.SetColorPostProcessMode(PDFRasterizerColorPostProcessMode.e_postprocess_none);
                    viewerBackground = (Color)ResourceHandler.ViewerResources["MainViewerBackgroundColor"];
                    PDFViewCtrl.SetBackgroundColor(viewerBackground);
                    PDFViewCtrl.Background = new SolidColorBrush(viewerBackground);
                    PDFViewCtrl.Update(true);
                }
                else if (ViewerSettings.ColorMode == CustomColorModeType.Custom)
                {
                    int iconIndex = ViewerSettings.CurrentCustomColorIcon;
                    Tuple<Color, Color> colors = ViewerSettings.GetCustomColors()[iconIndex];
                    viewerBackground = UtilityFunctions.GetViewerBackgroundColor(colors.Item1);
                    if (!UtilityFunctions.MatchesColorPostProcessMode(PDFViewCtrl, PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map, colors.Item1, colors.Item2))
                    {
                        PDFViewCtrl.SetColorPostProcessMode(PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map);
                        PDFViewCtrl.SetColorPostProcessColors(colors.Item1, colors.Item2);
                        PDFViewCtrl.Background = new SolidColorBrush(viewerBackground);
                    }

                    PDFViewCtrl.SetBackgroundColor(viewerBackground);
                    PDFViewCtrl.Update(true);
                }

                ResolveReflowAndOutlineColors();
            }
        }

        private void ResolveReflowAndOutlineColors()
        {
            if (PDFViewCtrl == null)
                return;

            bool changeReflow = false;
            bool changeOutline = false;

            if (TabControlViewModel != null
                && TabControlViewModel.SelectedTab != null
                && TabControlViewModel.SelectedTab.ReflowView != null)
            {
                if (TabControlViewModel.SelectedTab.IsReflow)
                    changeReflow = true;
            }

            if (OutlineDialogViewModel != null
                && OutlineDialogViewModel.Thumbnails != null)
            {
                if (IsOutlineDialogOpen)
                    changeOutline = true;
            }

            if (changeReflow || changeOutline)
            {
                PDFRasterizerColorPostProcessMode mode = PDFViewCtrl.GetColorPostProcessMode();
                Color newWhite = Colors.White;
                Color newBlack = Colors.Black;
                if (mode != PDFRasterizerColorPostProcessMode.e_postprocess_none)
                {
                    newWhite = UtilityFunctions.GetPostProcessedColor(newWhite, PDFViewCtrl);
                    newBlack = UtilityFunctions.GetPostProcessedColor(newBlack, PDFViewCtrl);
                }

                if (changeReflow)
                    TabControlViewModel.SelectedTab.ReflowView.SetColorPostProcessingMode(mode, newWhite, newBlack);

                if (changeOutline)
                    OutlineDialogViewModel.Thumbnails.BlankPageDefaultColor = newWhite;
            }
        }

        private void ResolveReflow()
        {
            if (TabControlViewModel != null)
            {
                PdfViewCtrlTabInfo tab = TabControlViewModel.SelectedTab;
                if (tab != null)
                {
                    if ((_currentReflow != null && tab != _currentReflow.Item1 && _currentReflow.Item2 != null)
                        || (_currentReflow?.Item2 != null && !_currentReflow.Item1.IsReflow))
                    {
                        DeactivateReflow();
                    }

                    if (tab.IsReflow && tab.ReflowView != null)
                    {
                        _currentReflow = new Tuple<PdfViewCtrlTabInfo, ReflowViewModel>(tab, tab.ReflowView.ReflowViewModel);
                        _currentReflow.Item2.PageChanged += CurrentreflowViewModel_PageChanged;
                        PageNumberIndicator.ReflowViewModel = tab.ReflowView.ReflowViewModel;

                        tab.NavigationStack.ReflowViewModel = tab.ReflowView.ReflowViewModel;
                        tab.NavigationStack.CurrentNavigator = NavigationStack.Navigators.ReflowView;
                    }

                    RaisePropertyChanged(nameof(IsReflow));
                    if (tab.IsReflow && tab.ReflowView != null && tab.IsReflow && tab.ReflowView != null)
                        tab.ReflowView.SetFocus(FocusState.Programmatic);

                    ResolveReflowAndOutlineColors();
                }
            }
        }

        private void DeactivateReflow()
        {
            if (_currentReflow != null)
            {
                _currentReflow.Item2.PageChanged -= CurrentreflowViewModel_PageChanged;
                PageNumberIndicator.ReflowViewModel = null;

                _currentReflow.Item1.NavigationStack.ReflowViewModel = null;
                _currentReflow.Item1.NavigationStack.CurrentNavigator = NavigationStack.Navigators.PDFViewCtrl;
            }

            _currentReflow = null;
        }

        private void CurrentreflowViewModel_PageChanged(int pageNumber, int totalPages)
        {
            if (PDFViewCtrl != null)
            {
                PDFViewCtrl.SetCurrentPage(pageNumber);
                PDFViewCtrl.CancelRendering();
            }
        }

        private async Task ShowCorruptedMessageAsync()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            ContentDialog dialog = new ContentDialog
            {
                Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                Content = string.Format(loader.GetString("DocumentsPage_FileOpeningError_CorruptFile_Info"), ViewerSettings.DisplayName),
                CloseButtonText = _loader.GetString("pdftron_Controls_Content_Dialog_OK")
            };

            await DialogHelper.ShowContentDialogAsync(dialog);
            await NavigateAwayAsync();
        }

        public async Task NotifyReadOnlyDocumentAsync()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            ContentDialog dialog = new ContentDialog
            {
                Title = loader.GetString("Viewerpage_DocumentModifiedDialog_ReadOnly_Title"),
                Content = string.Format(loader.GetString("Viewerpage_DocumentModifiedDialog_ReadOnly_Info"), loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option")),
                PrimaryButtonText = loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option"),
                SecondaryButtonText = loader.GetString("Viewerpage_DocumentModifiedDialog_ReadOnly_Dismiss"),
                CloseButtonText = _loader.GetString("pdftron_Tool_NoteDialog_Cancel")
            };

            dialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                    {
                        // We can't call SaveAs(false) here, or we'll get an UnauthorizedAccessException when trying to open the file picker.
                        await DelaySaveAs(TabControlViewModel.SelectedTab, false, false);
                    });

            await DialogHelper.ShowContentDialogAsync(dialog);
        }

        public async Task NotifyCreatedDocumentAsync()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");


            ContentDialog dialog = new ContentDialog
            {
                Title = loader.GetString("Viewerpage_DocumentModifiedDialog_Created_Title"),
                Content = string.Format(loader.GetString("Viewerpage_DocumentModifiedDialog_Created_Info"), loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option")),
                PrimaryButtonText = loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option"),
                CloseButtonText = loader.GetString("Viewerpage_DocumentModifiedDialog_Created_Dismiss")
            };

            dialog.PrimaryButtonCommand = new RelayCommand(async _ => 
            {
                // We can't call SaveAs(false) here, or we'll get an UnauthorizedAccessException when trying to open the file picker.
                await DelaySaveAs(TabControlViewModel.SelectedTab, false, false);
            });

            await DialogHelper.ShowContentDialogAsync(dialog);
        }

        public async Task NotyfyReparablePdfDocAsync()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            ContentDialog dialog = new ContentDialog
            {
                Title = loader.GetString("Viewerpage_DocumentModifiedDialog_Corrupted_Title"),
                Content = string.Format(loader.GetString("Viewerpage_DocumentModifiedDialog_Corrupted_Info"),
                ViewerSettings.DisplayName,
                loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option")),
                PrimaryButtonText = loader.GetString("ViewerPage_FileOpeningError_SaveACopy_Option"),
                CloseButtonText = loader.GetString("Viewerpage_DocumentModifiedDialog_Corrupted_Dismiss")
            };

            dialog.PrimaryButtonCommand = new RelayCommand(async _ =>
                    {
                        // We can't call SaveAs(false) here, or we'll get an UnauthorizedAccessException when trying to open the file picker.
                        await DelaySaveAs(TabControlViewModel.SelectedTab, false, false);
                    });


            await DialogHelper.ShowContentDialogAsync(dialog);
        }

        // If we could run into a state where _CurrentFile was null, we would return null and set the error
        private StorageFile SharingHelper_RetrieveSharingStorageFile(ref string errorMessage) => TemporaryFile;

        private string SharingHelper_RetrieveSharingString()
        {
            if (ToolManager.HasSelectedText)
                return ToolManager.GetSelectedText();

            return null;
        }

        private IAsyncAction CloseDocAsync(PDFDoc doc)
        {
            Task t = new Task(() => doc.Dispose());
            t.Start();

            return t.AsAsyncAction();
        }

        private void OpenThumbviewerInEditMode(int defaultSelectedPage)
        {
            IsThumbnailsViewOpen = true;
            ThumbnailViewer?.SelectPages(new List<int> { defaultSelectedPage });
        }

        [SuppressMessage("Redundancy", "RCS1163:Unused parameter.", Justification = "<Pending>")]
        [SuppressMessage("Style", "IDE0060:Remove unused parameter", Justification = "<Pending>")]
        [SuppressMessage("Performance", "CA1822:Mark members as static", Justification = "<Pending>")]
        public void DragOver(object sender, DragEventArgs e)
        {
            _viewerDragOver?.Invoke(sender, e);
        }

        [SuppressMessage("Redundancy", "RCS1163:Unused parameter.", Justification = "<Pending>")]
        [SuppressMessage("Style", "IDE0060:Remove unused parameter", Justification = "<Pending>")]
        [SuppressMessage("Performance", "CA1822:Mark members as static", Justification = "<Pending>")]
        public void Drop(object sender, DragEventArgs e)
        {
            _viewerDrop?.Invoke(sender, e);
        }

        public void RaiseViewModelPropertyChanged(string name)
        {
            RaisePropertyChanged(name);
        }
    }
}
