﻿using pdftron.PDF.Tools.Controls.Viewer.Dialogs;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Menu;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Search;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Resources;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.Storage;
using Windows.UI.Core;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238

namespace pdftron.PDF.Tools.Controls.Viewer
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public partial class ViewerPage : NavigablePage
    {
        private AppBar _topAppBar;
        private AppBar _bottomAppBar;
        private ViewerViewModel _viewModel;
        private bool _smallScreen;
        private double _commandBarCompactHeight = 48;
        private double _commandbarHeight = 48;
        private AnchorSides _oldOutlineAnchorSide;
        private double _oldWidth;
        private MenuFlyout _saveCopyMenuFlyout = new MenuFlyout
        {
            Placement = FlyoutPlacementMode.BottomEdgeAlignedLeft,
            MenuFlyoutPresenterStyle = (Style)Application.Current.Resources["MenuFlyoutPresenterDefaultStyle"]
        };
        private static IList<IMenuItem> _saveCopyMenuItems;
        private static string _licenseKey = string.Empty;

        private RoutedEventHandler _outlineResizer_PositionUpdated;
        private ViewerSettings.SettingUpdatedDelegate _settingsUpdatedHandler;

        private bool SmallScreen
        {
            get => _smallScreen;
            set
            {
                if (value != _smallScreen)
                {
                    _smallScreen = value;
                    if (_viewModel != null)
                        _viewModel.OutlineDialogNavigationCloses = _smallScreen;
                }
            }
        }

        public ViewerViewModel ViewModel => _viewModel;

        public ViewerPage()
        {
            InitializeComponent();

            if (InitializeOnCreation)
                Initialize();

            SizeChanged += ViewerPage_SizeChanged;
            RightTapped += ViewerPage_RightTapped;
            BackgroundGrid.Loaded += BackgroundGrid_Loaded;
            CompleteReaderTopAppBar.Opening += CompleteReaderTopAppBar_Opening;
            CompleteReaderTopAppBar.Closing += CompleteReaderTopAppBar_Closing;
            CompeteReaderTopCommandBar.Opening += CompeteReaderTopCommandBar_Opening;
            CompeteReaderTopCommandBar.Closing += CompeteReaderTopCommandBar_Closing;

            if (Resources.ContainsKey("AppBarThemeCompactHeight"))
            {
                _commandbarHeight = (double)Resources["AppBarThemeCompactHeight"];
                _commandBarCompactHeight = _commandbarHeight;
            }

            _topAppBar = CompleteReaderTopAppBar;
            _bottomAppBar = CompeteReaderBottomAppBar;
            PlaceOutlineDialog();

            string deviceFamily = Windows.System.Profile.AnalyticsInfo.VersionInfo.DeviceFamily;
            if (!string.IsNullOrWhiteSpace(deviceFamily) && deviceFamily.Equals("Windows.Mobile", StringComparison.OrdinalIgnoreCase))
            {
                AppBarButton_Files.Visibility = Visibility.Collapsed;
            }

            double tabButtonHeight = (double)Application.Current.Resources["OutlineTabButtonSize"];
        }

        private void Initialize()
        {
            PDFNet.AddResourceSearchPath(System.IO.Path.Combine(ApplicationData.Current.LocalFolder.Path, "fonts"));
            PDFNet.Initialize(_licenseKey);

            string res_path = ApplicationData.Current.LocalFolder.Path;

            PDFNet.SetResourcesPath(res_path);
            PDFNet.SetPersistentCachePath(res_path);
            string additionalResourcePath = System.IO.Path.Combine(Package.Current.InstalledLocation.Path, "Resources");
            PDFNet.AddResourceSearchPath(additionalResourcePath);

            PDFViewCtrl.RegisterDependencyProperties();

            DocumentPreviewCache.Initialize(100 * 1024 * 1024, 0.1);
            ReflowProcessor.Initialize();
            DocumentPreviewCache.ClearCache(); // attempt to prevent startup crash
        }

        private void InitializeDefaultSaveCopyMenuFlyout()
        {
            ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

            MenuFlyoutItem optimize = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)Application.Current.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("ViewerPage_SecondaryOptions_Optimize"),
                Command = ViewModel.SaveReducedFileSizeCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-compress.png"),
                    ShowAsMonochrome = false
                }
            };

            MenuFlyoutItem encrypt = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)Application.Current.Resources["CommandBarDefaultForeground"],
                Command = ViewModel.SavePasswordCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-encrypt.png"),
                    ShowAsMonochrome = false
                }
            };

            Binding encryptBinding = new Binding
            {
                Source = ViewModel,
                Path = new PropertyPath("SaveAPasswordCopyText"),
                Mode = BindingMode.TwoWay,
                UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
            };

            BindingOperations.SetBinding(encrypt, MenuFlyoutItem.TextProperty, encryptBinding);

            MenuFlyoutItem crop = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)Application.Current.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("ViewerPage_SecondaryOptions_PermCroppedCopy"),
                Command = ViewModel.SavePermCroppedCopyCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-crop-pdf.png"),
                    ShowAsMonochrome = false
                }
            };

            MenuFlyoutItem flatten = new MenuFlyoutItem
            {
                Foreground = (SolidColorBrush)Application.Current.Resources["CommandBarDefaultForeground"],
                Text = loader.GetString("MenuFlyoutItem_Flattened"),
                Command = ViewModel.SaveAndFlattenCommand,
                Icon = new BitmapIcon
                {
                    UriSource = new Uri("ms-appx:///pdftron.PDF.Tools/Resources/Images/icon-disp-flatten.png"),
                    ShowAsMonochrome = false
                }
            };

            _saveCopyMenuFlyout.Items.Add(optimize);
            _saveCopyMenuFlyout.Items.Add(encrypt);
            _saveCopyMenuFlyout.Items.Add(crop);
            _saveCopyMenuFlyout.Items.Add(flatten);
            AppBarButton_SaveCopy.Flyout = _saveCopyMenuFlyout;
        }

        private void InitializeCustomSaveCopyMenuFlyout()
        {
            foreach (var item in SaveCopyMenuItems)
            {
                MenuFlyoutItemBase menuItem = GetMenuItem(item);
                if (menuItem is MenuFlyoutSubItem subMenuItem)
                {
                    foreach (var subItem in item.SubMenuItems)
                    {
                        MenuFlyoutItemBase menuSubItem = GetMenuItem(subItem);
                        subMenuItem.Items.Add(menuSubItem);
                    }
                }

                _saveCopyMenuFlyout.Items.Add(menuItem);
            }

            AppBarButton_SaveCopy.Flyout = _saveCopyMenuFlyout;
        }

        private MenuFlyoutItemBase GetMenuItem(IMenuItem item)
        {
            MenuFlyoutItemBase menuItem;
            if (!item.IsSubMenuItem)
            {
                MenuFlyoutItem mItem = new MenuFlyoutItem
                {
                    Text = item.Text,
                    Command = item.Command,
                    CommandParameter = item.CommandParameter
                };

                if (!string.IsNullOrEmpty(item.Style))
                    mItem.Style = (Style)Application.Current.Resources[item.Style];

                if (item.IconType == MenuIconType.Font)
                {
                    mItem.Icon = new FontIcon
                    {
                        Glyph = item.Glyph
                    };
                }
                else
                {
                    mItem.Icon = new BitmapIcon
                    {
                        UriSource = new Uri(item.ImageUri),
                        ShowAsMonochrome = false
                    };
                }

                menuItem = mItem;
            }
            else
            {
                MenuFlyoutSubItem sItem = new MenuFlyoutSubItem
                {
                    Text = item.Text
                };

                if (item.IconType == MenuIconType.Font)
                {
                    sItem.Icon = new FontIcon
                    {
                        Glyph = item.Glyph
                    };
                }
                else
                {
                    sItem.Icon = new BitmapIcon
                    {
                        UriSource = new Uri(item.ImageUri),
                        ShowAsMonochrome = false
                    };
                }

                menuItem = sItem;
            }

            menuItem.Foreground = (SolidColorBrush)Application.Current.Resources["CommandBarDefaultForeground"];
            menuItem.FontSize = 13;

            if (item.IsCommandBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.CommandBindingPath)
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.CommandProperty, binding);
            }

            if (item.IsTextBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.TextBindingPath),
                    Mode = BindingMode.TwoWay,
                    UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.TextProperty, binding);
            }

            if (item.IsEnabledBinding)
            {
                Binding binding = new Binding
                {
                    Source = ViewModel,
                    Path = new PropertyPath(item.IsEnabledBindingPath),
                    Mode = BindingMode.TwoWay,
                    UpdateSourceTrigger = UpdateSourceTrigger.PropertyChanged
                };

                BindingOperations.SetBinding(menuItem, MenuFlyoutItem.IsEnabledProperty, binding);
            }

            return menuItem;
        }

        public static bool InitializeOnCreation { get; set; } = true;

        public static IList<IMenuItem> SaveCopyMenuItems
        {
            get => _saveCopyMenuItems;
            set => _saveCopyMenuItems = value;
        }

        private double CommandbarHeight
        {
            get => _commandbarHeight;
            set
            {
                if (_commandbarHeight != value)
                {
                    _commandbarHeight = value;

                    // NOTE: Once the Commandbar opens it adds margin to the tabbded control
                    if (this._viewModel.IsAnnotationToolbarOpen)
                        TabButtonControl.Margin = new Thickness(0, Math.Max(0, _commandbarHeight - _commandBarCompactHeight), 0, 0);
                }
            }
        }

        private void Settings_SettingUpdated(string settingName)
        {
            if (settingName == "OutlineDialogAnchorSide")
                PlaceOutlineDialog();

            if (settingName == "OutlineDialogWidth")
                PlaceOutlineDialog();
        }

        private void PlaceOutlineDialog()
        {
            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                OutlineDialog.AnchorSide = AnchorSides.None;
            }
            else
            {
                OutlineDialog.AnchorSide = ViewerSettings.OutlineDialogAnchorSide;
            }

            double newWidth = Math.Min(ActualWidth, 400);
            if (OutlineDialog.AnchorSide == _oldOutlineAnchorSide && newWidth == _oldWidth)
                return;

            _oldWidth = newWidth;
            if (OutlineDialog.AnchorSide == AnchorSides.None)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(0, 1, 0, 0);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Stretch;
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 1);
                SnapResizeHost.Visibility = Visibility.Collapsed;
                OutlineHostEntranceControl.EntranceAnimation = FadeInOutline;
                OutlineHostEntranceControl.ExitAnimation = FadeOutOutline;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostGridTranslation.X = 0;
                    OutlineHostEntranceControl.Opacity = 0;
                }

                UpdateOutlineWidthPosition();
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Left)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(0, 1, 1, 0);
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 0);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Left;
                SnapResizeHost.HorizontalAlignment = HorizontalAlignment.Left;
                OutlineHostEntranceControl.EntranceAnimation = SlideInOutline;
                OutlineHostEntranceControl.ExitAnimation = SlideOutOutlineToLeft;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostEntranceControl.Opacity = 1;
                    OutlineHostGridTranslation.X = -OutlineDialog.Width;
                }

                OutlineResizer.Offsets = GetOutlineOffsets();
                OutlineResizer.Position = ViewerSettings.OutlineDialogWidth;
                UpdateOutlineWidthPosition();
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Right)
            {
                NonModalDisplayGrid.BorderThickness = new Thickness(1, 1, 0, 0);
                OutlineSpaceHolderGrid.SetValue(Grid.ColumnProperty, 2);
                OutlineHostGrid.HorizontalAlignment = HorizontalAlignment.Right;
                SnapResizeHost.HorizontalAlignment = HorizontalAlignment.Right;
                OutlineHostEntranceControl.EntranceAnimation = SlideInOutline;
                OutlineHostEntranceControl.ExitAnimation = SlideOutOutlineToRight;
                if (!OutlineHostEntranceControl.IsOpen)
                {
                    OutlineHostEntranceControl.Opacity = 1;
                    OutlineHostGridTranslation.X = OutlineDialog.Width;
                }

                OutlineResizer.Offsets = GetOutlineOffsets();
                OutlineResizer.Position = 2 - ViewerSettings.OutlineDialogWidth;
                UpdateOutlineWidthPosition();
            }
        }

        private void OutlineResizer_PositionUpdated(object sender, RoutedEventArgs e)
        {
            int position = OutlineResizer.Position;
            if (OutlineDialog.AnchorSide == AnchorSides.Right)
                position = 2 - position;

            ViewerSettings.OutlineDialogWidth = position;
            UpdateOutlineWidthPosition();
        }

        private void UpdateOutlineWidthPosition()
        {
            int position = ViewerSettings.OutlineDialogWidth;
            if (position == 0)
            {
                OutlineDialog.Width = OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow];
            }
            else if (position == 1)
            {
                OutlineDialog.Width = OutlineDialog.LayoutMinWidths[WidthLayouts.Full];
            }
            else
            {
                OutlineDialog.Width = Math.Min(ActualWidth, 400.0);
            }
        }

        private List<double> GetOutlineOffsets()
        {
            List<double> offsets = new List<double>();
            if (OutlineDialog.AnchorSide == AnchorSides.Left)
            {
                offsets.Add(OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow]);
                offsets.Add(OutlineDialog.LayoutMinWidths[WidthLayouts.Full]);
                offsets.Add(Math.Min(ActualWidth, 400));
            }
            else if (OutlineDialog.AnchorSide == AnchorSides.Right)
            {
                offsets.Add(500 - Math.Min(ActualWidth, 400));
                offsets.Add(500 - OutlineDialog.LayoutMinWidths[WidthLayouts.Full]);
                offsets.Add(500 - OutlineDialog.LayoutMinWidths[WidthLayouts.Narrow]);
            }

            return offsets;
        }

        public void SearchFlyout_OnOpened(object sender, object e)
        {
            FindTextDialog.SetFocus();
            InputPane.GetForCurrentView().TryShow();
        }

        private void HandleSearchUi(double width)
        {
            if (width < 450)
            {
                AppBarButton_FindClose.Width = 40;
                AppBarButton_FindClose.Visibility = Visibility.Collapsed;
                FullTextSearchListView.ItemTemplate = (DataTemplate)Resources["SmallFullTextSearchListViewDataTemplate"];
            }
            else
            {
                AppBarButton_FindClose.Width = 68;
                AppBarButton_FindClose.Visibility = Visibility.Visible;
                FullTextSearchListView.ItemTemplate = (DataTemplate)Resources["FullTextSearchListViewDataTemplate"];
            }
        }

        private void ViewerPage_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            _viewModel?.ToolManager?.CloseOpenDialog(true);

            SmallScreen = (e.NewSize.Width < 500 || (e.NewSize.Height < 500 && e.NewSize.Width < 700));
            PlaceOutlineDialog();
            HandleSearchUi(e.NewSize.Width);

            if (e.NewSize.Width < 700)
            {
                Binding binding = AppBarButton_Save.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_Save2.SetBinding(VisibilityProperty, binding);
                    AppBarButton_Save.Visibility = Visibility.Collapsed;
                }

                Binding binding2 = AppBarButton_SaveAs.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding2 != null)
                {
                    AppBarButton_SaveAs2.SetBinding(VisibilityProperty, binding2);
                    AppBarButton_SaveAs.Visibility = Visibility.Collapsed;
                }
            }
            else
            {
                Binding binding = AppBarButton_Save2.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_Save.SetBinding(VisibilityProperty, binding);
                    AppBarButton_Save2.Visibility = Visibility.Collapsed;
                }

                Binding binding2 = AppBarButton_SaveAs2.GetBindingExpression(VisibilityProperty)?.ParentBinding;
                if (binding != null)
                {
                    AppBarButton_SaveAs.SetBinding(VisibilityProperty, binding2);
                    AppBarButton_SaveAs2.Visibility = Visibility.Collapsed;
                }
            }
        }

        protected override async void OnNavigatedTo(NavigationEventArgs e)
        {
            if (InitializeOnCreation)
                await TabControlViewModel.Instance.InitializeInstanceAsync();

            if (_settingsUpdatedHandler == null)
            {
                _settingsUpdatedHandler = new ViewerSettings.SettingUpdatedDelegate(Settings_SettingUpdated);
                ViewerSettings.SettingUpdated += _settingsUpdatedHandler;
            }

            ViewerViewModel viewModel = null;
            if (e.Parameter is ViewerViewModel)
            {
                viewModel = e.Parameter as ViewerViewModel;
                DataContext = viewModel;
                _viewModel = viewModel;
            }

            base.OnNavigatedTo(e);

            if (viewModel == null)
            {
                viewModel = new ViewerViewModel();
                _viewModel = viewModel;
                DataContext = _viewModel;
                ActivateViewModel(e.Parameter);
            }

            if (InitializeOnCreation)
            {
                InitializeDefaultSaveCopyMenuFlyout();
            }
            else
            {
                InitializeCustomSaveCopyMenuFlyout();
            }

            _viewModel.OpenSearch += ViewModel_OpenSearch;

            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                _viewModel.IsUiPinned = false;
                AppBarButton_PinUI.Visibility = Visibility.Collapsed;
                AppBarButton_UnpinUI.Visibility = Visibility.Collapsed;
            }

            viewModel.OutlineDialogNavigationCloses = SmallScreen;
            viewModel.PropertyChanged += ViewModel_PropertyChanged;
            viewModel.AnnotationToolbar.AppBarChanged += ViewerPage_AppBarChanged;
            viewModel.FindTextResultFound += ViewModel_FindTextResultFound;
            _outlineResizer_PositionUpdated = new RoutedEventHandler(OutlineResizer_PositionUpdated);
            OutlineResizer.PositionUpdated += _outlineResizer_PositionUpdated;
        }

        /// <summary>
        /// Initialize the viewer control
        /// </summary>
        /// <param name="viewModel">Custom ViewModel</param>
        /// <returns></returns>
        public async Task InitializePageAsync(ViewerViewModel viewModel = null)
        {
            if (InitializeOnCreation)
                await TabControlViewModel.Instance.InitializeInstanceAsync();

            if (_settingsUpdatedHandler == null)
            {
                _settingsUpdatedHandler = new ViewerSettings.SettingUpdatedDelegate(Settings_SettingUpdated);
                ViewerSettings.SettingUpdated += _settingsUpdatedHandler;
            }

            _viewModel = viewModel ?? new ViewerViewModel();
            DataContext = _viewModel;
            ActivateViewModel(viewModel);

            if (InitializeOnCreation)
            {
                InitializeDefaultSaveCopyMenuFlyout();
            }
            else
            {
                InitializeCustomSaveCopyMenuFlyout();
            }

            _viewModel.OpenSearch += ViewModel_OpenSearch;

            if (UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                _viewModel.IsUiPinned = false;
                AppBarButton_PinUI.Visibility = Visibility.Collapsed;
                AppBarButton_UnpinUI.Visibility = Visibility.Collapsed;
            }

            _viewModel.OutlineDialogNavigationCloses = SmallScreen;
            _viewModel.PropertyChanged += ViewModel_PropertyChanged;
            _viewModel.AnnotationToolbar.AppBarChanged += ViewerPage_AppBarChanged;
            _viewModel.FindTextResultFound += ViewModel_FindTextResultFound;
            _outlineResizer_PositionUpdated = new RoutedEventHandler(OutlineResizer_PositionUpdated);
            OutlineResizer.PositionUpdated += _outlineResizer_PositionUpdated;
        }

        protected override void OnNavigatingFrom(NavigatingCancelEventArgs e)
        {
            if (_settingsUpdatedHandler != null)
            {
                ViewerSettings.SettingUpdated -= _settingsUpdatedHandler;
                _settingsUpdatedHandler = null;
            }

            if (_outlineResizer_PositionUpdated != null)
            {
                OutlineResizer.PositionUpdated -= _outlineResizer_PositionUpdated;
                _outlineResizer_PositionUpdated = null;
            }

            base.OnNavigatingFrom(e);

            DataContext = null;
        }

        protected async override void NavigableViewModel_NewINavigableAvailable(INavigable sender, INavigable newNavigable)
        {
            if (newNavigable != null && newNavigable != sender)
            {
                if (newNavigable is ViewerViewModel)
                {
                    if (_viewModel != null)
                    {
                        DeactivateViewModel(_viewModel);
                        _viewModel.PropertyChanged -= ViewModel_PropertyChanged;
                    }

                    _viewModel = (newNavigable as ViewerViewModel);
                    DataContext = _viewModel;
                    _viewModel.PropertyChanged += ViewModel_PropertyChanged;
                    _viewModel.OpenSearch += ViewModel_OpenSearch;
                    ResolveEditButtonFlyout();
                    ActivateViewModel(newNavigable);
                }
                else
                {
                    //If there are no more opened documents, then navigate to the Recent documents page
                    // since the Opened Docs will be empty and removed from the navigation
                    //if (newNavigable is DocumentViewModel viewModel
                    //    && viewModel.GetSubView() == SubViewState.OpenedDocs
                    //    && !viewModel.OpenedTabs)
                    //{
                    //    await viewModel.SetSubViewAsync(SubViewState.RecentPage);
                    //    Frame.Navigate(typeof(DocumentBasePage), newNavigable);
                    //    return;
                    //}

                    if (Frame.CanGoBack)
                    {
                        Frame.GoBack();
                    }
                    else if (newNavigable.NavigablePage != null)
                    {
                        Frame.Navigate(newNavigable.NavigablePage, newNavigable);
                    }
                }
            }
            else
            {
                _viewModel.IsSaveButtonVisible = false;
                _viewModel.IsSaveAsButtonVisible = false;
                _viewModel.IsOutlineDialogOpen = false;
                _viewModel.IsAnnotationToolbarOpen = false;
                _viewModel.RaiseViewModelPropertyChanged("IsConverting");
                _viewModel.RaiseViewModelPropertyChanged("IsFindTextDialogOpen");
                _viewModel.RaiseViewModelPropertyChanged("IsDocumentEditable");
                _viewModel.RaiseViewModelPropertyChanged("IsDocumentSearchable");
                _viewModel.RaiseViewModelPropertyChanged("ShowPageNumberIndicator");
                _viewModel.RaiseViewModelPropertyChanged("AreZoombuttonsVisible");
                _viewModel.RaiseViewModelPropertyChanged("IsThumbnailSliderVisible");
            }
        }

        private void ViewModel_OpenSearch(object sender, Deferral e)
        {
            try
            {
                AppBarButton_Search.Flyout.ShowAt(AppBarButton_Search);
            }
            finally
            {
                e.Complete();
            }
        }

        private async void ViewModel_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            // hide the ViewerPage app bar in thumbnails/crop view
            if (e.PropertyName.Equals("IsThumbnailsViewOpen") || e.PropertyName.Equals("IsCropViewOpen"))
            {
                if (_viewModel.IsThumbnailsViewOpen || _viewModel.IsCropViewOpen)
                {
                    TopAppBar = null;
                    BottomAppBar = null;
                    AppBarButton_ViewMode.Flyout.Hide();
                    SetViewerPageUp();
                }
                else
                {
                    TopAppBar = _topAppBar;
                    BottomAppBar = _bottomAppBar;
                    if (ViewerSettings.PinCommandBar)
                        SetViewerPageDown();
                }
            }

            if (e.PropertyName.Equals("IsFullScreen"))
                AppBarButton_ViewMode.Flyout.Hide();

            if (e.PropertyName.Equals("MediaElement"))
            {
                // remove any previous text to speech elements placed
                foreach (var child in BackgroundGrid.Children)
                {
                    if (child is MediaElement element && element.Name == "TextToSpeechMedia")
                        BackgroundGrid.Children.Remove(child);
                }

                MediaElement media = _viewModel.MediaElement;
                media.Name = "TextToSpeechMedia";

                if (media != null)
                {
                    // need to add MediaElement to UI tree in order to access its events
                    BackgroundGrid.Children.Add(media);
                    media.MediaEnded += (_, __) => _viewModel.AreAudioButtonsVisible = false;
                }
            }

            if (e.PropertyName.Equals("IsQuickSettingsOpen") && !_viewModel.IsQuickSettingsOpen)
                await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => AppBarButton_ViewMode.Flyout.Hide());

            if (e.PropertyName.Equals("IsAnnotationToolbarOpen"))
            {
                if (_viewModel.IsAnnotationToolbarOpen)
                {
                    SetViewerPageSlightlyDown();
                    TabStoryBoardClose.Begin();
                }
                else
                {
                    if (!ViewerSettings.PinCommandBar)
                    {
                        SetViewerPageUp();
                    }
                    else
                    {
                        SetViewerPageDown();
                    }

                    TabStoryBoardCollapse.Begin();
                }
            }

            if (e.PropertyName.Equals("IsUiPinned"))
            {
                if (_viewModel.IsUiPinned)
                {
                    SetViewerPageDown();
                }
                else
                {
                    SetViewerPageUp();
                }
            }

            if (e.PropertyName.Equals("IsConverting"))
                ResolveEditButtonFlyout();
        }

        private void ResolveEditButtonFlyout()
        {
            if (_viewModel?.IsConverting == true)
            {
                if (AppBarButton_Edit.Flyout == null)
                {
                    AppBarButton_Edit.Flyout = AppBarButton_Edit.Resources["EditButtonFlyout"] as Flyout;
                    AppBarButton_Edit.Foreground = Resources["SystemControlDisabledBaseMediumLowBrush"] as SolidColorBrush;
                }
            }
            else
            {
                AppBarButton_Edit.Flyout = null;
                AppBarButton_Edit.Foreground = Resources["SystemControlForegroundBaseHighBrush"] as SolidColorBrush;
            }
        }

        public void FindText_ContentIsAvailable(EntranceAnimationContentControlEx control) => FindTextDialog.SetFocus();

        public async Task ActivateWithFileAsync(StorageFile file, NewDocumentProperties properties = null)
        {
            if (DataContext is ViewerViewModel viewModel)
            {
                await viewModel.ActivateWithFileAsync(file, properties);
                viewModel.IsThumbnailsViewOpen = false;
            }
        }

        private void BackgroundGrid_Loaded(object sender, RoutedEventArgs e)
        {
            if (ViewerSettings.PinCommandBar)
                SetViewerPageDown();
        }

        private void ViewerPage_RightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType == PointerDeviceType.Mouse)
            {
                HandleCommandBarInteraction();
                e.Handled = true;
            }
        }

        private void HandleCommandBarInteraction()
        {
            var annotationCommandBar = (AnnotationBorder.Child as AnnotationCommandBar);
            if (!ViewerSettings.PinCommandBar)
            {
                if (!_viewModel.IsEntireAppBarOpen)
                {
                    _viewModel.IsEntireAppBarOpen = true;
                }
                else if (!_viewModel.IsAnnotationToolbarOpen)
                {
                    _viewModel.IsAppBarOpen = false;
                    _viewModel.IsEntireAppBarOpen = !_viewModel.IsEntireAppBarOpen;
                    annotationCommandBar.IsAppBarOpen = false;
                }
                else
                {
                    (AnnotationBorder.Child as AnnotationCommandBar).IsAppBarOpen = !(AnnotationBorder.Child as AnnotationCommandBar).IsAppBarOpen;
                    _viewModel.UpdateUIVisiblity();
                }
            }
            else
            {
                if (!_viewModel.IsAnnotationToolbarOpen)
                {
                    _viewModel.IsAppBarOpen = !_viewModel.IsAppBarOpen;
                }
                else
                {
                    if (!annotationCommandBar.IsInkOpen && !annotationCommandBar.IsPolygonSaveOpen && !annotationCommandBar.IsMeasureOpen)
                    {
                        (AnnotationBorder.Child as AnnotationCommandBar).IsAppBarOpen = !(AnnotationBorder.Child as AnnotationCommandBar).IsAppBarOpen;
                        _viewModel.UpdateUIVisiblity();
                    }
                }
            }
        }

        private void ViewerPage_AppBarChanged(bool isOpen)
        {
            if (isOpen)
            {
                HandleOpeningCommandBar();
            }
            else
            {
                HandleClosingCommandBar();
            }

            _viewModel.UpdateUIVisiblity();
        }

        private void CompleteReaderTopAppBar_Opening(object sender, object e)
        {
        }

        private void CompleteReaderTopAppBar_Closing(object sender, object e)
        {
            if (ViewerSettings.PinCommandBar)
            {
                CompleteReaderTopAppBar.IsOpen = true;
                _viewModel.UpdateUIVisiblity();
            }
        }

        private void CompeteReaderTopCommandBar_Opening(object sender, object e) => HandleOpeningCommandBar();

        private void CompeteReaderTopCommandBar_Closing(object sender, object e) => HandleClosingCommandBar();

        private void HandleOpeningCommandBar()
        {
            if (_viewModel?.IsAnnotationToolbarOpen == true)
            {
                TabBorder.Margin = new Thickness(0, 48, 0, 0); // Note: Update margin to the TabBorder once the Annotation Toolbar is opened

                CommandbarHeight = 60; // Magic number, height AppBar expands to if all buttons have only 1 line of text
            }
            else
            {
                CommandbarHeight = 76; // Magic number, height AppBar expands to if at least one button has 2 lines of text
            }

            if (pdftron.PDF.Tools.UtilityFunctions.GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                TabStoryBoardCollapse.Begin();
            }
            else
            {
                TabStoryBoardExpand.Begin();
            }
        }

        private void HandleClosingCommandBar()
        {
            CommandbarHeight = _commandBarCompactHeight;

            if (_viewModel.IsAnnotationToolbarOpen)
            {
                TabBorder.Margin = new Thickness(0, 40, 0, 0); // Note: Update margin to the TabBorder once the Annotation Toolbar is closed

                TabStoryBoardClose.Begin();
            }
            else
            {
                TabStoryBoardCollapse.Begin();
            }
        }

        private void SetViewerPageSlightlyDown() => BackgroundGrid.Margin = new Thickness(0, 40, 0, 0);

        private void SetViewerPageDown() => BackgroundGrid.Margin = new Thickness(0, 80, 0, 0);

        private void SetViewerPageUp() => BackgroundGrid.Margin = new Thickness(0, 0, 0, 0);

        private void ViewModel_FindTextResultFound(Deferral deferral)
        {
            try
            {
                if (FindTextDialog.DoesSearchBoxHaveFocus)
                {
                    //AppBarButton_SearchOpen.Focus(FocusState.Keyboard);
                }
            }
            finally
            {
                deferral.Complete();
            }
        }

        /// <summary>
        /// Waits until the data context containing the appropriate full search item is ready and
        /// then positions the highlight rectangle according to the information
        /// </summary>
        public void Highlight_Text(FrameworkElement sender, DataContextChangedEventArgs args)
        {
            Rectangle rect = sender as Rectangle;

            // Only care when data context becomes non-null
            if (!(rect.DataContext is FullSearchItem dataContext))
                return;

            // Use a textblock to hold text to measure where highlight should be positioned
            TextBlock textBlock = new TextBlock();
            string contextText = dataContext.ContextText;

            string leftText = contextText.Substring(0, contextText.IndexOf(dataContext.SearchText));
            textBlock.Text = leftText;
            // Trailing empty spaces won't count for width, so use a placeholder character
            if (textBlock.Text.EndsWith(" "))
                textBlock.Text = textBlock.Text.Substring(0, textBlock.Text.Length - 1) + "i";

            textBlock.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            textBlock.Arrange(new Windows.Foundation.Rect(new Windows.Foundation.Point(0, 0), textBlock.DesiredSize));
            double leftOffset = textBlock.DesiredSize.Width;

            string rightText = leftText + dataContext.SearchText;
            if (textBlock.Text.EndsWith(" "))
                textBlock.Text = textBlock.Text.Substring(0, textBlock.Text.Length - 1) + "i";

            textBlock.Text = rightText;
            textBlock.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            textBlock.Arrange(new Windows.Foundation.Rect(new Windows.Foundation.Point(0, 0), textBlock.DesiredSize));
            double rightOffset = textBlock.DesiredSize.Width;

            // Apply calculations to position highlight for text
            rect.Margin = new Thickness(leftOffset, 0, 0, 0);
            rect.Height = 17;
            rect.Width = rightOffset - leftOffset;
        }

        public static void SetLicenseKey(string licenseKey) => _licenseKey = licenseKey;
    }
}
