﻿using System;
using System.Diagnostics.CodeAnalysis;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;

namespace Actions.Crypto
{
    /// <summary>
    /// DataProtection
    /// </summary>
    public sealed class DataProtection
    {
        private const string FALLBACK_ID = "bcc22927-4566-4160-94da-3eddb6877a0f";

        private static readonly DataProtection _instance = new DataProtection();
        private string _uniqueId = FALLBACK_ID;

        private byte[] _entropy;

        static DataProtection()
        {
        }

        private DataProtection()
        {
        }

        //private async Task InitializeAsync()
        //{
        //    try
        //    {
        //        DefaultSettings settings = await Desktop.GetDefaultSettingsAsync();
        //        if (settings != null)
        //        {
        //            _uniqueId = settings.Id;
        //        }
        //        else
        //        {
        //            _uniqueId = FALLBACK_ID;
        //        }
        //    }
        //    catch
        //    {
        //        _uniqueId = FALLBACK_ID;
        //    }
        //}

        /// <summary>
        /// Singleton Instance
        /// </summary>
        [SuppressMessage("Simplification", "RCS1085:Use auto-implemented property.", Justification = "<Pending>")]
        public static DataProtection Instance => _instance;

        private byte[] GetEntropy()
        {
            return _entropy ?? (_entropy = Encoding.Unicode.GetBytes(_uniqueId));
        }

        /// <summary>
        /// Encrypt string
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public string Encrypt(string text)
        {
            //if (string.IsNullOrEmpty(_uniqueId))
            //    await InitializeAsync();

            try
            {
                if (string.IsNullOrEmpty(text))
                    return string.Empty;

                byte[] originalText = Encoding.Unicode.GetBytes(text);
                byte[] encryptedText = ProtectedData.Protect(originalText, GetEntropy(), DataProtectionScope.LocalMachine);

                return Convert.ToBase64String(encryptedText);
            }
            catch (Exception)
            {
                return string.Empty;
            }
        }

        /// <summary>
        /// Decrypt string
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public string Decrypt(string text)
        {
            //if (string.IsNullOrEmpty(_uniqueId))
            //    await InitializeAsync();

            try
            {
                if (string.IsNullOrEmpty(text))
                    return string.Empty;

                byte[] encryptedText = Convert.FromBase64String(text);
                byte[] originalText = ProtectedData.Unprotect(encryptedText, GetEntropy(), DataProtectionScope.LocalMachine);

                return Encoding.Unicode.GetString(originalText);
            }
            catch (Exception)
            {
                return string.Empty;
            }
        }
    }
}
