﻿using System.Collections.Generic;
using System.Linq;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;

namespace pdftron.PDF.Tools.Extensions
{
    public static class DependencyObjectExtensions
    {
        private static List<FrameworkElement> AllChildren(this DependencyObject parent)
        {
            var list = new List<FrameworkElement>();
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(parent); i++)
            {
                var child = VisualTreeHelper.GetChild(parent, i);
                if (child is FrameworkElement)
                    list.Add(child as FrameworkElement);

                list.AddRange(AllChildren(child));
            }

            return list;
        }

        public static IEnumerable<T> FindControls<T>(this DependencyObject parent)
        {
            var children = AllChildren(parent);
            return children.OfType<T>();
        }

        public static T FindVisualChild<T>(this DependencyObject depObj) where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child is T t)
                        return t;

                    T childItem = FindVisualChild<T>(child);
                    if (childItem != null)
                        return childItem;
                }
            }

            return null;
        }

        public static T FindControl<T>(this UIElement parent, string controlName) where T : FrameworkElement
        {
            if (parent == null)
                return null;

            if (parent.GetType() == typeof(T) && ((T)parent).Name == controlName)
                return (T)parent;

            T result = null;
            int count = VisualTreeHelper.GetChildrenCount(parent);
            for (int i = 0; i < count; i++)
            {
                UIElement child = (UIElement)VisualTreeHelper.GetChild(parent, i);

                if (FindControl<T>(child, controlName) != null)
                {
                    result = FindControl<T>(child, controlName);
                    break;
                }
            }

            return result;
        }
    }
}
