﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UISize = Windows.Foundation.Size;
using UIRect = Windows.Foundation.Rect;
using PDFRect = pdftron.PDF.Rect;
using Windows.UI.Input;
using Windows.UI.Xaml.Controls.Primitives;

namespace pdftron.PDF.Tools
{
    public class MagnifyingTool : Tool
    {
        public MagnifyingTool(PDFViewCtrl ctrl, ToolManager tMan)
        : base(ctrl, tMan)
        {
            mToolMode = ToolType.e_magnifying_rectangle;
            mNextToolMode = ToolType.e_magnifying_rectangle;
        }

        internal override void OnCreate()
        {
            mWidgetPosition = new PDFRect();
            mWidgetBoundsOnScreen = new PDFRect();
            Size = WidgetMinSize;
            base.OnCreate();
        }

        internal override void OnClose()
        {
            if (mWidgetPopup != null)
            {
                mWidgetPopup.IsOpen = false;
            }
            base.OnClose();
        }

        public event Windows.UI.Xaml.RoutedEventHandler PositionChanged;

        #region Appearance Customization

        private const double _WidgetThickness = 1;
        private const double _EllipseRadius = 7;
        private const double[] _WidgetStrokeDash = null;
        private readonly Brush _WidgetStrokeBrush = new SolidColorBrush(Colors.Blue);
        private readonly Brush _EllipseFillBrush = new SolidColorBrush(Colors.Blue);

        private Path GetWidgetPath()
        {
            Path path = new Path();
            path.StrokeThickness = _WidgetThickness;
            path.Stroke = _WidgetStrokeBrush;
            path.StrokeLineJoin = PenLineJoin.Miter;
            path.StrokeEndLineCap = PenLineCap.Flat;
            path.StrokeStartLineCap = PenLineCap.Flat;
            PathGeometry geom = new PathGeometry();
            PathFigure widgetFigure = new PathFigure();
            widgetFigure.Segments.Add(new LineSegment());
            widgetFigure.Segments.Add(new LineSegment());
            widgetFigure.Segments.Add(new LineSegment());
            widgetFigure.Segments.Add(new LineSegment());
            geom.Figures.Add(widgetFigure);
            path.Data = geom;
            return path;
        }

        private Ellipse GetEllipse()
        {
            Ellipse ellipse = new Ellipse();
            ellipse.StrokeThickness = 0;
            ellipse.Fill = _EllipseFillBrush;
            ellipse.Width = _EllipseRadius * 2;
            ellipse.Height = _EllipseRadius * 2;
            return ellipse;
        }


        #endregion Appearance Customization


        #region Methods

        public void SnapWidget()
        {
            UpdateWidgetBoundsOnScreen();
            if (mWidgetPosition.x2 > mWidgetBoundsOnScreen.x2)
            {
                mWidgetPosition.x1 = mWidgetBoundsOnScreen.x2 - mWidgetPosition.Width();
                mWidgetPosition.x2 = mWidgetBoundsOnScreen.x2;
            }
            if (mWidgetPosition.y2 > mWidgetBoundsOnScreen.y2)
            {
                mWidgetPosition.y1 = mWidgetBoundsOnScreen.y2 - mWidgetPosition.Height();
                mWidgetPosition.y2 = mWidgetBoundsOnScreen.y2;
            }
            if (mWidgetPosition.x1 < mWidgetBoundsOnScreen.x1)
            {
                mWidgetPosition.x2 = mWidgetBoundsOnScreen.x1 + mWidgetPosition.Width();
                mWidgetPosition.x1 = mWidgetBoundsOnScreen.x1;
            }
            if (mWidgetPosition.y1 < mWidgetBoundsOnScreen.y1)
            {
                mWidgetPosition.y2 = mWidgetBoundsOnScreen.y1 + mWidgetPosition.Height();
                mWidgetPosition.y1 = mWidgetBoundsOnScreen.y1;
            }
            PositionWidget();
            PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
        }

        public enum NavigationDirections
        {
            Left,
            Up,
            Right,
            Down,
        }

        public void MoveWidget(NavigationDirections direction)
        {
            double availableSpace = GetAvailableSpace(direction);
            switch (direction)
            {
                case NavigationDirections.Left:
                    Position = new UIPoint(Math.Max(Position.X - Size.Width, Position.X - availableSpace), Position.Y);
                    break;
                case NavigationDirections.Up:
                    Position = new UIPoint(Position.X, Math.Max(Position.Y - Size.Height, Position.Y - availableSpace));
                    break;
                case NavigationDirections.Right:
                    Position = new UIPoint(Math.Min(Position.X + Size.Width, Position.X + availableSpace), Position.Y);
                    break;
                case NavigationDirections.Down:
                    Position = new UIPoint(Position.X, Math.Min(Position.Y + Size.Height, Position.Y + availableSpace));
                    break;
            }
        }

        #endregion Methods


        #region Public Properties

        private bool mScrollingEnabled = true;
        /// <summary>
        /// Gets or sets whether or not scrolling is enabled.
        /// </summary>
        public bool ScrollingEnabled
        {
            get { return mScrollingEnabled; }
            set { mScrollingEnabled = value; }
        }

        /// <summary>
        /// How touch scrolling is handled.
        /// Either use the systems default panning for ScrollViewers
        /// or use a more mechanical drag like hand tool for mouse
        /// </summary>
        public enum TouchScrollingModes
        {
            system,
            drag,
        }

        private TouchScrollingModes mTouchScrollingmode = TouchScrollingModes.system;
        /// <summary>
        /// Gets or sets whether system scrolling en enabled.
        /// </summary>
        public TouchScrollingModes TouchScrollingmode
        {
            get { return mTouchScrollingmode; }
            set { mTouchScrollingmode = value; }
        }

        private bool mEnablePinchZooming = true;
        /// <summary>
        /// Gets or sets whether or not the user can pinch to zoom
        /// Can only be enabled if TouchScrollingmode is system.
        /// </summary>
        public bool EnablePinchZooming
        {
            get { return mEnablePinchZooming; }
            set { mEnablePinchZooming = value; }
        }

        // Gets or sets the top left corner of the magnifying widget relative to the PDFViewCtrl
        public UIPoint Position
        {
            get { return new UIPoint(mWidgetPosition.x1, mWidgetPosition.y1); }
            set
            {
                if (Position != value)
                {
                    mWidgetPosition.x2 = value.X + Size.Width;
                    mWidgetPosition.y2 = value.Y + Size.Height;
                    mWidgetPosition.x1 = value.X;
                    mWidgetPosition.y1 = value.Y;
                    if (Size.Width > 0 && Size.Height > 0)
                    {
                        SnapWidget();
                        PositionWidget();
                    }
                    InteractionDrivers driver = InteractionDriver;
                    try
                    {
                        if (driver == InteractionDrivers.None)
                        {
                            InteractionDriver = InteractionDrivers.Tool;
                        }
                        PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        InteractionDriver = driver;
                    }
                }
            }
        }

        /// <summary>
        /// Gets or sets the Size of the magnifying widget
        /// </summary>
        public UISize Size
        {
            get { return new UISize(mWidgetPosition.Width(), mWidgetPosition.Height()); }
            set
            {
                if (Size != value)
                {
                    double newWidth = Math.Max(value.Width, WidgetMinSize.Width);
                    double newHeight = Math.Max(value.Height, WidgetMinSize.Height);
                    UISize newSize = new UISize(newWidth, newHeight);
                    if (Size != newSize)
                    {
                        mWidgetPosition.x2 = mWidgetPosition.x1 + newSize.Width;
                        mWidgetPosition.y2 = mWidgetPosition.y1 + newSize.Height;
                        if (newSize.Width > 0 && newSize.Height > 0)
                        {
                            EnsureAttached();
                            PositionWidget();
                        }
                        InteractionDrivers driver = InteractionDriver;
                        try
                        {
                            if (driver == InteractionDrivers.None)
                            {
                                InteractionDriver = InteractionDrivers.Tool;
                            }
                            PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
                        }
                        catch (Exception ex)
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                        }
                        finally
                        {
                            InteractionDriver = driver;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Gets the closest page that the top left corner of the rectangle is on.
        /// </summary>
        public int PageOfTopLeftCorner
        {
            get { return 0; } // TODO
        }
        
        /// <summary>
        /// Returns the top left corner in the page coordinates of page PageOfTopLeftCorner
        /// </summary>
        public UIPoint TopLeftCornerInPageSpace
        {
            get { return new UIPoint(); } // TODO
        }

        private UISize mWidgetMinSize = new UISize(30, 30);
        /// <summary>
        /// Gets or sets the minimum size of the widget.
        /// </summary>
        public UISize WidgetMinSize
        {
            get { return mWidgetMinSize; }
            set
            {
                if (mWidgetMinSize != value)
                {
                    mWidgetMinSize = value;
                    double newWidth = Math.Max(Size.Width, mWidgetMinSize.Width);
                    double newHeight = Math.Max(Size.Height, mWidgetMinSize.Height);
                    Size = new UISize(newWidth, newHeight);
                }
            }
        }

        private double mEllipseHitRadius = 15;
        public double EllipseHitRadius
        {
            get { return mEllipseHitRadius; }
            set { mEllipseHitRadius = value; }
        }

        #endregion Public Properties


        #region Interaction

        private enum ControlPoints
        {
            None,
            Pan,
            Rectangle,
            Ellipse,
        }

        private ControlPoints mCurrentControlPoint = ControlPoints.None;
        private uint mCurrentPointerID;
        private UIPoint mViewerDownPoint;
        private PDFRect mWidgetRectOnManipulationStarted;
        private PDFRect mWidgetBoundsOnScreen;

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerPressedHandler(sender, e);
            if (mToolManager.ContactPoints.Count == 1)
            {
                PointerPoint pp = e.GetCurrentPoint(mPDFView);
                mCurrentControlPoint = GetControlPoint(pp.Position);
                if (mCurrentControlPoint == ControlPoints.Pan && ScrollingEnabled 
                    && TouchScrollingmode == TouchScrollingModes.system 
                    && e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
                {
                    mPDFView.SetZoomEnabled(EnablePinchZooming);
                    mPDFView.SetScrollEnabled(true);
                    mCurrentControlPoint = ControlPoints.None;
                }
                else
                {
                    mPDFView.SetZoomEnabled(false);
                    mPDFView.SetScrollEnabled(false);
                    mPDFView.CapturePointer(e.Pointer);
                    mCurrentPointerID = e.Pointer.PointerId;
                    mViewerDownPoint = pp.Position;
                    mWidgetRectOnManipulationStarted = new PDFRect(mWidgetPosition.x1,
                        mWidgetPosition.y1, mWidgetPosition.x2,
                        mWidgetPosition.y2);
                }
                if (mCurrentControlPoint != ControlPoints.Pan)
                {
                    InteractionDriver = InteractionDrivers.Tool;
                }
            }
            else
            {
                mCurrentControlPoint = ControlPoints.None;
                mWidgetRectOnManipulationStarted = null;
                mPDFView.ReleasePointerCaptures();
            }
            return false;
        }

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mCurrentControlPoint != ControlPoints.None && e.Pointer.PointerId == mCurrentPointerID)
            {
                UIPoint dragPoint = e.GetCurrentPoint(mPDFView).Position;
                if (mCurrentControlPoint == ControlPoints.Rectangle)
                {
                    MoveRectangle(dragPoint);
                    PositionWidget();
                    PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
                }
                else if (mCurrentControlPoint == ControlPoints.Ellipse)
                {
                    MoveEllipse(dragPoint);
                    PositionWidget();
                }
                else if (mCurrentControlPoint == ControlPoints.Pan)
                {
                    mPDFView.SetHScrollPos(mPDFView.GetHScrollPos() + mViewerDownPoint.X - dragPoint.X);
                    mPDFView.SetVScrollPos(mPDFView.GetVScrollPos() + mViewerDownPoint.Y - dragPoint.Y);
                    mViewerDownPoint = dragPoint;
                }
                return true;
            }
            return false;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerReleasedHandler(sender, e);
            mPDFView.SetZoomEnabled(true);
            mPDFView.SetScrollEnabled(true);
            if (mCurrentControlPoint == ControlPoints.Ellipse)
            {
                PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
            }
            mCurrentControlPoint = ControlPoints.None;
            InteractionDriver = InteractionDrivers.None;
            mPDFView.ReleasePointerCaptures();
            return false;
        }

        internal override bool PointerCanceledHandler(object sender, PointerRoutedEventArgs e)
        {
            mPDFView.SetZoomEnabled(true);
            mPDFView.SetScrollEnabled(true);
            if (mCurrentControlPoint == ControlPoints.Ellipse)
            {
                PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
            };
            mCurrentControlPoint = ControlPoints.None;
            InteractionDriver = InteractionDrivers.None;
            base.PointerCanceledHandler(sender, e);
            mPDFView.ReleasePointerCaptures();
            return false;
        }

        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            mPDFView.SetZoomEnabled(true);
            mPDFView.SetScrollEnabled(true);
            if (mCurrentControlPoint == ControlPoints.Ellipse)
            {
                PositionChanged?.Invoke(this, new Windows.UI.Xaml.RoutedEventArgs());
            };
            mCurrentControlPoint = ControlPoints.None;
            InteractionDriver = InteractionDrivers.None;
            base.PointerCaptureLostHandler(sender, e);
            return false;
        }

        private ControlPoints GetControlPoint(UIPoint positionInPDFviewCtrl)
        {
            double ellipseXDist = mWidgetPosition.x2 - positionInPDFviewCtrl.X;
            double ellipseYDist = mWidgetPosition.y2 - positionInPDFviewCtrl.Y;
            if ((ellipseXDist * ellipseXDist) + (ellipseYDist * ellipseYDist) <= mEllipseHitRadius * mEllipseHitRadius)
            {
                return ControlPoints.Ellipse;
            }
            else if (positionInPDFviewCtrl.X >= mWidgetPosition.x1 - _WidgetThickness
                && positionInPDFviewCtrl.X <= mWidgetPosition.x2 + _WidgetThickness
                && positionInPDFviewCtrl.Y >= mWidgetPosition.y1 - _WidgetThickness
                && positionInPDFviewCtrl.Y <= mWidgetPosition.y2 + _WidgetThickness)
            {
                return ControlPoints.Rectangle;
            }

            return ControlPoints.Pan;
        }

        private void MoveRectangle(UIPoint pointerPosition)
        {
            double newLeft = mWidgetRectOnManipulationStarted.x1 + pointerPosition.X - mViewerDownPoint.X;
            double newTop = mWidgetRectOnManipulationStarted.y1 + pointerPosition.Y - mViewerDownPoint.Y;
            double newRight = mWidgetRectOnManipulationStarted.x2 + pointerPosition.X - mViewerDownPoint.X;
            double newBottom = mWidgetRectOnManipulationStarted.y2 + pointerPosition.Y - mViewerDownPoint.Y;
            UpdateWidgetBoundsOnScreen();
            if (newRight > mWidgetBoundsOnScreen.x2)
            {
                newLeft += mWidgetBoundsOnScreen.x2 - newRight;
                newRight = mWidgetBoundsOnScreen.x2;
            }
            if (newBottom > mWidgetBoundsOnScreen.y2)
            {
                newTop += mWidgetBoundsOnScreen.y2 - newBottom;
                newBottom = mWidgetBoundsOnScreen.y2;
            }
            if (newLeft < mWidgetBoundsOnScreen.x1)
            {
                newRight += mWidgetBoundsOnScreen.x1 - newLeft;
                newLeft = mWidgetBoundsOnScreen.x1;
            }
            if (newTop < mWidgetBoundsOnScreen.y1)
            {
                newBottom += mWidgetBoundsOnScreen.y1 - newTop;
                newTop = mWidgetBoundsOnScreen.y1;
            }

            mWidgetPosition.x1 = newLeft;
            mWidgetPosition.y1 = newTop;
            mWidgetPosition.x2 = newRight;
            mWidgetPosition.y2 = newBottom;
        }

        private void MoveEllipse(UIPoint pointerPosition)
        {
            UpdateWidgetBoundsOnScreen();
            double wToHRatio = mWidgetPosition.Width() / mWidgetPosition.Height();

            // determine which dimensions would prompt the largest move
            double hDist = Math.Abs(pointerPosition.X - mViewerDownPoint.X);
            double vDist = Math.Abs(pointerPosition.Y - mViewerDownPoint.Y);
            bool pickByHorizontal = true;

            double newRight = mWidgetRectOnManipulationStarted.x2 + pointerPosition.X - mViewerDownPoint.X;
            double newBottom = mWidgetRectOnManipulationStarted.y2 + pointerPosition.Y - mViewerDownPoint.Y;
            if (pickByHorizontal)
            {
                newBottom = mWidgetRectOnManipulationStarted.y2 +
                    (((pointerPosition.X - mViewerDownPoint.X) / mWidgetRectOnManipulationStarted.Width()) * mWidgetRectOnManipulationStarted.Height());
            }
            else
            {
                newRight = mWidgetRectOnManipulationStarted.x2 +
                    (((pointerPosition.Y - mViewerDownPoint.Y) / mWidgetRectOnManipulationStarted.Height()) * mWidgetRectOnManipulationStarted.Width());
            }

            if (newRight < mWidgetPosition.x1 + WidgetMinSize.Width)
            {
                newRight = mWidgetPosition.x1 + WidgetMinSize.Width;
                newBottom = mWidgetPosition.y1 + (WidgetMinSize.Width * (1.0 / wToHRatio));
            }
            if (newBottom < mWidgetPosition.y1 + WidgetMinSize.Height)
            {
                newBottom = mWidgetPosition.y1 + WidgetMinSize.Height;
                newRight = mWidgetPosition.x1 + (WidgetMinSize.Height * wToHRatio);
            }

            if (newRight > mWidgetBoundsOnScreen.x2)
            {
                newRight = mWidgetBoundsOnScreen.x2;
                newBottom = mWidgetPosition.y1 + ((newRight - mWidgetPosition.x1) * (1.0 / wToHRatio));
            }
            if (newBottom > mWidgetBoundsOnScreen.y2)
            {
                newBottom = mWidgetBoundsOnScreen.y2;
                newRight = mWidgetPosition.x1 + ((newBottom - mWidgetPosition.y1) * wToHRatio);
            }

            mWidgetPosition.x2 = newRight;
            mWidgetPosition.y2 = newBottom;
            PositionWidget();
        }


        private void UpdateWidgetBoundsOnScreen()
        {
            PDFRect rect = mPDFView.GetContentBoundingBox();

            mWidgetBoundsOnScreen.x1 = Math.Max(rect.x1, 0);
            mWidgetBoundsOnScreen.y1 = Math.Max(rect.y1, 0);
            mWidgetBoundsOnScreen.x2 = Math.Min(rect.x2, mPDFView.ActualWidth) - 1;
            mWidgetBoundsOnScreen.y2 = Math.Min(rect.y2, mPDFView.ActualHeight) - 1;

            UIRect pdfViewRect = UtilityFunctions.GetElementRect(mPDFView, null);
            mWidgetPopup.HorizontalOffset = pdfViewRect.X;
            mWidgetPopup.VerticalOffset = pdfViewRect.Y;

            RectangleGeometry clipGeom = new RectangleGeometry();
            clipGeom.Rect = new UIRect(0, 0, pdfViewRect.Width, pdfViewRect.Height);
            this.Clip = clipGeom;
        }

        #endregion Interaction


        #region Interaction Widget

        private PDFRect mWidgetPosition;
        private Popup mWidgetPopup;

        private void EnsureAttached()
        {
            if (mWidgetPopup == null)
            {
                mWidgetPopup = new Popup();
                mWidgetPopup.IsHitTestVisible = false;
                mWidgetPopup.IsOpen = true;
                Canvas c = new Canvas();
                mWidgetPopup.Child = c;
                c.Children.Add(this);
                if (mMagnifyingRectangle == null)
                {
                    CreateWidget();
                }
            }
        }

        private Path mMagnifyingRectangle;

        private Ellipse mResizeEllipse;

        private void CreateWidget()
        {
            if (mMagnifyingRectangle == null)
            {
                mMagnifyingRectangle = GetWidgetPath();
                this.Children.Add(mMagnifyingRectangle);
            }
            
            if (mResizeEllipse == null)
            {
                mResizeEllipse = GetEllipse();
                this.Children.Add(mResizeEllipse);
            }
        }

        /// <summary>
        /// This assumes the Widget is already added, and should be used by the manipulation events
        /// </summary>
        private void PositionWidget()
        {
            if (mWidgetPosition != null)
            {
                UIPoint start = new UIPoint(mWidgetPosition.x2 + (_WidgetThickness / 2),
                    mWidgetPosition.y2 + (_WidgetThickness / 2));
                UIPoint tr = new UIPoint(start.X, start.Y - mWidgetPosition.Height() - _WidgetThickness);
                UIPoint tl = new UIPoint(tr.X - mWidgetPosition.Width() - _WidgetThickness, tr.Y);
                UIPoint bl = new UIPoint(tl.X, tl.Y + mWidgetPosition.Height() + _WidgetThickness);
                UIPoint end = new UIPoint(bl.X + mWidgetPosition.Width() + _WidgetThickness, bl.Y);

                PathFigure widgetFigure = (mMagnifyingRectangle.Data as PathGeometry).Figures[0];
                widgetFigure.StartPoint = start;
                (widgetFigure.Segments[0] as LineSegment).Point = tr;
                (widgetFigure.Segments[1] as LineSegment).Point = tl;
                (widgetFigure.Segments[2] as LineSegment).Point = bl;
                (widgetFigure.Segments[3] as LineSegment).Point = end;

                mResizeEllipse.SetValue(Canvas.LeftProperty, mWidgetPosition.x2 - _EllipseRadius);
                mResizeEllipse.SetValue(Canvas.TopProperty, mWidgetPosition.y2 - _EllipseRadius);
            }
        }

        private double GetAvailableSpace(NavigationDirections direction)
        {
            UpdateWidgetBoundsOnScreen();
            switch (direction)
            {
                case NavigationDirections.Left:
                    return Math.Max(Position.X - mWidgetBoundsOnScreen.x1, 0);
                case NavigationDirections.Up:
                    return Math.Max(Position.Y - mWidgetBoundsOnScreen.y1, 0);
                case NavigationDirections.Right:
                    return Math.Max(mWidgetBoundsOnScreen.x2 - Position.X - Size.Width, 0);
                case NavigationDirections.Down:
                    return Math.Max(mWidgetBoundsOnScreen.y2 - Position.Y - Size.Height, 0);
            }
            return 0;
        }

        #endregion InteractionWidget


        #region Synchronization
        
        public enum InteractionDrivers
        {
            None,
            Tool,
            User
        }

        private InteractionDrivers _InteractionDriver = InteractionDrivers.None;
        /// <summary>
        /// Gets or sets which widget or control is driving interaction. If the MagnifyingTool's wigets is being manipulated,
        /// the driver is tool, but if the view that shows the zoomed in rectangle is being manipulated, it can set the 
        /// InteractionDriver to user.
        /// 
        /// <remark> Note, never set InteractionDriver to user unless it currently is none, and always return it to
        /// none when finished.
        /// </remark>
        /// </summary>
        public InteractionDrivers InteractionDriver
        {
            get { return _InteractionDriver; }
            set { _InteractionDriver = value; }
        }


        #endregion Synchronization

        #region Utility Functions


        #endregion Utility Functions

        // TODO REMOVE!!!!
        internal override bool RightTappedHandler(object sender, RightTappedRoutedEventArgs e)
        {
            UIPoint pt = e.GetPosition(mPDFView);
            mViewerDownPoint = new UIPoint(mWidgetPosition.x2, mWidgetPosition.y2);
            MoveEllipse(pt);
            PositionWidget();
            return true;
        }
    }
}
