﻿using System;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.Popups;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Utilities
{
    /// <summary>
    /// This class makes sure that there's only ever 1 message dialog open at any given time.
    /// </summary>
    public static class DialogHelper
    {
        private static IAsyncOperation<IUICommand> _lastMessageDialog;
        private static IAsyncOperation<ContentDialogResult> _lastContentDialog;

        [Obsolete("ShowMessageDialogAsync() is deprecated, please use ShowContentDialogAsync() instead.")]
        /// <summary>
        /// Shows the message dialog passed and returns the IUICommand executed after the dialog is dismissed.
        /// </summary>
        /// <param name="md">a message dialog</param>
        /// <returns>the IUICommand executed</returns>
        public static async Task<IUICommand> ShowMessageDialogAsync(MessageDialog md)
        {
            _lastMessageDialog?.Cancel();

            IUICommand clickedCommand = null;
            try
            {
                _lastMessageDialog = md.ShowAsync();
                clickedCommand = await _lastMessageDialog;
            }
            catch
            {
                //Swallow exception
            }

            return clickedCommand;
        }

        /// <summary>
        /// Shows the content dialog passed and returns the ContentDialogResult after the dialog is dismissed.
        /// </summary>
        /// <param name="cd">a content dialog</param>
        /// <returns>the ContentDialogResult</returns>
        public static async Task<ContentDialogResult> ShowContentDialogAsync(ContentDialog cd)
        {
            _lastContentDialog?.Cancel();
            try
            {
                _lastContentDialog = cd.ShowAsync(ContentDialogPlacement.Popup);
                return await _lastContentDialog;
            }
            catch
            {
                //Swallow exception
            }

            return ContentDialogResult.None;
        }

        /// <summary>
        /// Close any started message dialog by selecting Cancel option
        /// </summary>
        public static void CancelLastMessageDialog()
        {
            if (_lastMessageDialog?.Status == AsyncStatus.Started)
            {
                _lastMessageDialog.Cancel();
            }
        }

        /// <summary>
        /// Close any started content dialog by selecting Cancel option
        /// </summary>
        public static void CancelLastContentDialog()
        {
            if (_lastContentDialog?.Status == AsyncStatus.Started)
            {
                _lastContentDialog.Cancel();
            }
        }
    }
}
